@@ -33,23 +33,17 @@ abstract_token = core.abstract_token
 canonicalize_shape = core.canonicalize_shape
 raise_to_shaped = core.raise_to_shaped
 
-def zeros_like_array(x):
-  dtype, weak_type = dtypes._lattice_result_type(x)
-  dtype = dtypes.canonicalize_dtype(dtype)
-  aval = ShapedArray(np.shape(x), dtype, weak_type=weak_type)
-  return ad_util.zeros_like_aval(aval)
-
 numpy_scalar_types: set[type] = {  # pylint: disable=g-bare-generic
-    np.int8, np.int16, np.int32, np.int64,
-    np.uint8, np.uint16, np.uint32, np.uint64,
+    dtypes.int4, np.int8, np.int16, np.int32, np.int64,
+    dtypes.uint4, np.uint8, np.uint16, np.uint32, np.uint64,
     np.complex64, np.complex128,
     np.bool_, np.longlong, np.intc,
 } | {np.dtype(dt).type for dt in dtypes._float_types}
 
-if dtypes.int4 is not None:
-  numpy_scalar_types.add(dtypes.int4)
-if dtypes.uint4 is not None:
-  numpy_scalar_types.add(dtypes.uint4)
+if dtypes.int2 is not None:
+  assert dtypes.uint2 is not None
+  numpy_scalar_types.add(dtypes.int2)
+  numpy_scalar_types.add(dtypes.uint2)
 
 array_types: set[type] = {np.ndarray} | numpy_scalar_types  # pylint: disable=g-bare-generic
 
@@ -62,11 +56,9 @@ def masked_array_error(*args, **kwargs):
                    "Use arr.filled() to convert the value to a standard numpy array.")
 
 core.pytype_aval_mappings[np.ma.MaskedArray] = masked_array_error
-ad_util.jaxval_zeros_likers[np.ma.MaskedArray] = masked_array_error
 
 for t in array_types:
   core.pytype_aval_mappings[t] = canonical_concrete_aval
-  ad_util.jaxval_zeros_likers[t] = zeros_like_array
 
 core.literalable_types.update(array_types)
 
@@ -82,6 +74,5 @@ def _make_concrete_python_scalar(t, x):
 
 for t in dtypes.python_scalar_dtypes:
   core.pytype_aval_mappings[t] = partial(_make_concrete_python_scalar, t)
-  ad_util.jaxval_zeros_likers[t] = partial(_zeros_like_python_scalar, t)
 
-core.literalable_types.update(dtypes.python_scalar_dtypes.keys())  # type: ignore
+core.literalable_types.update(dtypes.python_scalar_dtypes.keys())
