@@ -12,26 +12,28 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections.abc import Sequence
+from __future__ import annotations
+
+from collections.abc import Callable, Sequence
 import functools
 from functools import partial
 import logging
-from typing import Any, Callable, Optional, Union
+from typing import Any
 import types
 
 import numpy as np
 
 from jax._src import ad_util
 from jax._src import api
+from jax._src import config
 from jax._src import core
 from jax._src import dispatch
+from jax._src import dtypes
 from jax._src import linear_util as lu
 from jax._src import effects
 from jax._src import source_info_util
 from jax._src import traceback_util
-from jax._src import util
 from jax._src.api_util import flatten_fun, shaped_abstractify
-from jax._src.config import config
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
@@ -76,6 +78,17 @@ def dot_with_no_batch_dims_saveable(prim, *_, **params) -> bool:
       return True
   return False
 
+def offload_dot_with_no_batch_dims(offload_src, offload_dst):
+  def policy(prim, *_, **params):
+    # This is a useful heuristic for transformers.
+    if prim is lax_internal.dot_general_p:
+      (_, _), (lhs_b, rhs_b) = params['dimension_numbers']
+      if not lhs_b and not rhs_b:
+        return pe.Offloadable(src=offload_src, dst=offload_dst)
+    return pe.Recompute
+  return policy
+
+
 name_p = core.Primitive('name')
 
 def save_anything_except_these_names(*names_not_to_save):
@@ -105,6 +118,27 @@ def save_only_these_names(*names_which_can_be_saved):
     return False  # not saveable unless it's in the allow-list
   return policy
 
+def save_and_offload_only_these_names(
+    *, names_which_can_be_saved, names_which_can_be_offloaded,
+    offload_src, offload_dst):
+  names_which_can_be_saved = set(names_which_can_be_saved)
+  names_which_can_be_offloaded = set(names_which_can_be_offloaded)
+  intersection = names_which_can_be_saved.intersection(names_which_can_be_offloaded)
+  if intersection:
+    raise ValueError(
+        "The names should be exclusive and should not intersect in"
+        " `names_which_can_be_saved` and `names_which_can_be_offloaded`. Got"
+        f" names_which_can_be_saved={names_which_can_be_saved},"
+        f" names_which_can_be_offloaded={names_which_can_be_offloaded} and the"
+        f" intersection={intersection}")
+  def policy(prim, *_, **params):
+    if prim is name_p and params['name'] in names_which_can_be_saved:
+      return pe.Saveable
+    if prim is name_p and params['name'] in names_which_can_be_offloaded:
+      return pe.Offloadable(src=offload_src, dst=offload_dst)
+    return pe.Recompute  # not saveable unless it's in the allow-list
+  return policy
+
 
 def save_from_both_policies(policy_1, policy_2):
 
@@ -121,18 +155,20 @@ checkpoint_policies = types.SimpleNamespace(
     checkpoint_dots=dots_saveable,
     dots_with_no_batch_dims_saveable=dot_with_no_batch_dims_saveable,
     checkpoint_dots_with_no_batch_dims=dot_with_no_batch_dims_saveable,
+    offload_dot_with_no_batch_dims=offload_dot_with_no_batch_dims,
     save_anything_except_these_names=save_anything_except_these_names,
     save_any_names_but_these=save_any_names_but_these,
     save_only_these_names=save_only_these_names,
-    save_from_both_policies=save_from_both_policies)
+    save_from_both_policies=save_from_both_policies,
+    save_and_offload_only_these_names=save_and_offload_only_these_names)
 
 
 ### Main API
 
 @api_boundary
 def checkpoint(fun: Callable, *, prevent_cse: bool = True,
-               policy: Optional[Callable[..., bool]] = None,
-               static_argnums: Union[int, tuple[int, ...]] = (),
+               policy: Callable[..., bool] | None = None,
+               static_argnums: int | tuple[int, ...] = (),
                ) -> Callable:
   """Make ``fun`` recompute internal linearization points when differentiated.
 
@@ -370,7 +406,7 @@ def _trace_to_jaxpr(fun, in_tree, in_avals):
   flat_fun, out_tree = flatten_fun(lu.wrap_init(fun), in_tree)
   debug = pe.debug_info(fun, in_tree, out_tree, True, "checkpoint")
   try:
-    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals, debug)
+    jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals, debug)
   except core.ConcretizationTypeError as e:
     msg, = e.args
     if 'for checkpoint' not in msg:
@@ -429,11 +465,14 @@ def _saved_residuals(jaxpr, arg_info) -> list[tuple[core.AbstractValue, str]]:
         src = 'from the argument at flattened index {i}'
       results.append((v.aval, src))
 
+  named_vars = {v: e for e in jaxpr.eqns if e.primitive is name_p
+                for v in e.invars}
+
   for eqn in jaxpr.eqns:
     src = source_info_util.summarize(eqn.source_info)
     for v in eqn.outvars:
       if v in res_vars:
-        if eqn.primitive is name_p:
+        if eqn.primitive is name_p or v in named_vars and (eqn := named_vars[v]):
           results.append((v.aval, f"named '{eqn.params['name']}' from {src}"))
         elif str(eqn.primitive) == 'xla_call':
           results.append((v.aval,
@@ -506,10 +545,15 @@ def remat_partial_eval(trace, *tracers, jaxpr, **params):
   jaxpr_known, in_used_known = pe.dce_jaxpr(jaxpr_known, out_used_known)
   num_res = sum(used_res)
 
+  # To avoid precision mismatches in fwd and bwd passes due to XLA excess
+  # precision, insert explicit x = reduce_precision(x, **finfo(x.dtype)) calls
+  # on producers of any residuals. See https://github.com/google/jax/pull/22244.
+  jaxpr_known_ = _insert_reduce_precision(jaxpr_known, num_res)
+
   # compute known outputs and residuals (hoisted out of remat primitive)
   _, in_consts_ = unzip2(t.pval for t in tracers if t.pval.is_known())
   _, in_consts = partition_list(in_used_known, in_consts_)
-  out_consts = core.eval_jaxpr(jaxpr_known, (), *in_consts)
+  out_consts = core.eval_jaxpr(jaxpr_known_, (), *in_consts)
   out_knowns, residuals = split_list(out_consts, [len(out_consts)-num_res])
 
   # set up unknown outputs with a recipe to call remat
@@ -524,23 +568,24 @@ def remat_partial_eval(trace, *tracers, jaxpr, **params):
                              source_info_util.current())
 
   # log info about saved residuals
-  try:
-    _, staged_unk = partition_list(in_used_staged, in_unknowns)
-    res_invars, _ = partition_list(staged_unk, jaxpr_unknown.invars[num_res:])
-    res_outvars = jaxpr_known.outvars[len(jaxpr_known.outvars) - num_res:]
-    body_res = _saved_residuals(jaxpr_known.replace(outvars=res_outvars), None)
-    logger.log(logging.WARNING if config.jax_log_checkpoint_residuals
-               else logging.DEBUG,
-               'remat-decorated function ' +
-               'saving inputs with shapes:\n' * bool(res_invars) +
-               '  %s\n' * len(res_invars) +
-               'and ' * bool(res_invars) * bool(body_res) +
-               'saving these intermediates:\n' * bool(body_res) +
-               '  %s from %s\n' * len(body_res),
-               *[v.aval.str_short() for v in res_invars],
-               *[elt for (a, s) in body_res for elt in [a.str_short(), s]])
-  except:
-    pass  # just don't log anything on failure
+  log_level = logging.WARNING if config.log_checkpoint_residuals.value else logging.DEBUG
+  if logger.isEnabledFor(log_level):
+    try:
+      _, staged_unk = partition_list(in_used_staged, in_unknowns)
+      res_invars, _ = partition_list(staged_unk, jaxpr_unknown.invars[num_res:])
+      res_outvars = jaxpr_known.outvars[len(jaxpr_known.outvars) - num_res:]
+      body_res = _saved_residuals(jaxpr_known.replace(outvars=res_outvars), None)
+      logger.log(log_level,
+                'remat-decorated function ' +
+                'saving inputs with shapes:\n' * bool(res_invars) +
+                '  %s\n' * len(res_invars) +
+                'and ' * bool(res_invars) * bool(body_res) +
+                'saving these intermediates:\n' * bool(body_res) +
+                '  %s from %s\n' * len(body_res),
+                *[v.aval.str_short() for v in res_invars],
+                *[elt for (a, s) in body_res for elt in [a.str_short(), s]])
+    except:
+      pass  # just don't log anything on failure
 
   for t in out_jaxpr_tracers: t.recipe = recipe
 
@@ -548,6 +593,43 @@ def remat_partial_eval(trace, *tracers, jaxpr, **params):
   return merge_lists(out_unknowns, out_knowns, out_jaxpr_tracers)
 pe.custom_partial_eval_rules[remat_p] = remat_partial_eval
 
+@weakref_lru_cache
+def _insert_reduce_precision(jaxpr: core.Jaxpr, num_res: int) -> core.Jaxpr:
+  res_vars = jaxpr.outvars[len(jaxpr.outvars) - num_res:]
+  used_vars = {x for e in jaxpr.eqns for x in e.invars if isinstance(x, core.Var)}
+  invars, constvars, eqns = jaxpr.invars[:], jaxpr.constvars[:], jaxpr.eqns[:]
+  for v in res_vars:
+    if (not isinstance(v.aval, core.UnshapedArray) or
+        not dtypes.issubdtype(v.aval.dtype, np.inexact)):
+      continue
+    if v not in used_vars:
+      continue
+    assert isinstance(v, core.Var)
+    newvar = core.Var(v.suffix, v.aval)
+    finfo = dtypes.finfo(v.aval.dtype)
+    params = dict(exponent_bits=finfo.nexp, mantissa_bits=finfo.nmant)
+    if v in constvars or v in invars:
+      lst = constvars if v in constvars else invars
+      new_eqn = core.new_jaxpr_eqn(
+          [newvar], [v], lax_internal.reduce_precision_p, params, set())
+      lst[lst.index(v)] = newvar
+      eqns.insert(0, new_eqn)
+    else:
+      (eqn_idx, eqn), = ((i, e) for i, e in enumerate(eqns) if v in e.outvars)
+      if (eqn.primitive == lax_internal.reduce_precision_p and
+          eqn.params == params):
+        continue
+      replace_eqn = eqn.replace(outvars=[v_ if v_ != v else newvar
+                                         for v_ in eqn.outvars])
+      new_eqn = core.new_jaxpr_eqn(
+          [newvar], [v], lax_internal.reduce_precision_p, params, set(),
+          eqn.source_info, eqn.ctx)
+      eqns[eqn_idx] = replace_eqn
+      eqns.insert(eqn_idx+1, new_eqn)
+  new_jaxpr = jaxpr.replace(invars=invars, constvars=constvars, eqns=eqns)
+  config.enable_checks.value and core.check_jaxpr(new_jaxpr)
+  return new_jaxpr
+
 def remat_partial_eval_custom_params_updater(*args):
   *_, params_known, params_staged = args
   return params_known, dict(params_staged, differentiated=True)
@@ -555,12 +637,12 @@ pe.partial_eval_jaxpr_custom_rules[remat_p] = \
     partial(pe.call_partial_eval_custom_rule, 'jaxpr',
             remat_partial_eval_custom_params_updater)
 
-def remat_transpose(reduce_axes, out_cts, *in_primals, jaxpr, **params):
+def remat_transpose(out_cts, *in_primals, jaxpr, **params):
   assert not jaxpr.constvars
   in_linear = [ad.is_undefined_primal(x) for x in in_primals]
   out_zeros = [type(ct) is ad_util.Zero for ct in out_cts]
   transposed_jaxpr_, in_zeros = transpose_jaxpr(
-      pe.close_jaxpr(jaxpr), in_linear, out_zeros, reduce_axes)
+      pe.close_jaxpr(jaxpr), in_linear, out_zeros)
   transposed_jaxpr, consts = transposed_jaxpr_.jaxpr, transposed_jaxpr_.consts
   transposed_jaxpr = pe.convert_constvars_jaxpr(transposed_jaxpr)
   args, _ = tree_flatten((in_primals, out_cts))
@@ -574,19 +656,17 @@ def remat_transpose(reduce_axes, out_cts, *in_primals, jaxpr, **params):
 ad.reducing_transposes[remat_p] = remat_transpose
 
 # TODO(mattjj): move this to ad.py
-def transpose_jaxpr(jaxpr: core.ClosedJaxpr, in_linear: Union[bool, Sequence[bool]],
-                    out_zeros: Union[bool, Sequence[bool]],
-                    reduce_axes: Sequence[core.AxisName],
+def transpose_jaxpr(jaxpr: core.ClosedJaxpr, in_linear: bool | Sequence[bool],
+                    out_zeros: bool | Sequence[bool],
                     ) -> tuple[core.ClosedJaxpr, list[bool]]:
   if type(in_linear) is bool:
     in_linear = (in_linear,) * len(jaxpr.in_avals)
   if type(out_zeros) is bool:
     out_zeros = (out_zeros,) * len(jaxpr.out_avals)
-  return _transpose_jaxpr(jaxpr, tuple(in_linear), tuple(out_zeros),
-                          tuple(reduce_axes))
+  return _transpose_jaxpr(jaxpr, tuple(in_linear), tuple(out_zeros))
 
 @weakref_lru_cache
-def _transpose_jaxpr(jaxpr, in_lin, out_zeros, reduce_axes):
+def _transpose_jaxpr(jaxpr, in_lin, out_zeros):
   in_avals = ([a for a,  lin in zip(jaxpr.in_avals,  in_lin   ) if not lin] +
               [a for a, zero in zip(jaxpr.out_avals, out_zeros) if not zero])
   cell = lambda: None
@@ -601,8 +681,9 @@ def _transpose_jaxpr(jaxpr, in_lin, out_zeros, reduce_axes):
                 pe.PartialVal.known(next(ins_iter))
                 for aval, lin in zip(jaxpr.in_avals, in_lin)]
     assert next(ins_iter, None) is None
-    lin_jaxpr, _, consts = pe.trace_to_jaxpr_nounits(
-        lu.wrap_init(core.jaxpr_as_fun(jaxpr)), in_pvals, False)
+    with source_info_util.extend_name_stack('rematted_computation'):
+      lin_jaxpr, _, consts = pe.trace_to_jaxpr_nounits(
+          lu.wrap_init(core.jaxpr_as_fun(jaxpr)), in_pvals, False)
 
     # Transpose the linear jaxpr (which only has linear inputs).
     out_cts_iter = iter(out_cts_flat)
@@ -610,17 +691,16 @@ def _transpose_jaxpr(jaxpr, in_lin, out_zeros, reduce_axes):
                for aval, zero in zip(jaxpr.out_avals, out_zeros)]
     assert next(out_cts_iter, None) is None
     dummy_args = [ad.UndefinedPrimal(v.aval) for v in lin_jaxpr.invars]
-    in_cts = ad.backward_pass(lin_jaxpr, reduce_axes, False, consts, dummy_args,
-                              out_cts)
+    in_cts = ad.backward_pass(lin_jaxpr, False, consts, dummy_args, out_cts)
 
     # Identify symbolic zeros in the resulting cotangents, and return nonzeros.
     in_zeros = cell.in_cts_zero = [type(ct) is ad_util.Zero for ct in in_cts]
     in_cts_nz, _ = partition_list(in_zeros, in_cts)
     return in_cts_nz
 
-  transposed_jaxpr_, _, consts = pe.trace_to_jaxpr_dynamic(transposed, in_avals)
+  transposed_jaxpr_, _, consts, () = pe.trace_to_jaxpr_dynamic(transposed, in_avals)
   transposed_jaxpr = core.ClosedJaxpr(transposed_jaxpr_, consts)
-  return transposed_jaxpr, cell.in_cts_zero  # type: ignore
+  return transposed_jaxpr, cell.in_cts_zero  # pytype: disable=attribute-error
 
 def remat_vmap(spmd_axis_name, axis_size, axis_name, main_type, args, dims, *,
                jaxpr, **params):
@@ -639,27 +719,31 @@ batching.spmd_axis_primitive_batchers[remat_p] = remat_vmap
 
 # TODO(mattjj,sharadmv): de-duplicate with pe.dce_jaxpr_call_rule
 def remat_dce(used_outputs: list[bool], eqn: core.JaxprEqn
-              ) -> tuple[list[bool], Optional[core.JaxprEqn]]:
+              ) -> tuple[list[bool], core.JaxprEqn | None]:
   new_jaxpr, used_inputs = pe.dce_jaxpr(eqn.params['jaxpr'], used_outputs)
   new_params = dict(eqn.params, jaxpr=new_jaxpr)
-  if not any(used_inputs) and not any(used_outputs) and not new_jaxpr.effects:
+  if (not any(used_inputs) and not any(used_outputs) and
+      _has_effects(new_jaxpr.effects)):
     return used_inputs, None
   else:
     new_eqn = pe.new_jaxpr_eqn(
         [v for v, used in zip(eqn.invars, used_inputs) if used],
         [v for v, used in zip(eqn.outvars, used_outputs) if used],
-        eqn.primitive, new_params, new_jaxpr.effects, eqn.source_info)
+        eqn.primitive, new_params, new_jaxpr.effects, eqn.source_info, eqn.ctx)
     return used_inputs, new_eqn
 pe.dce_rules[remat_p] = remat_dce
 
+def _has_effects(effects) -> bool:
+  return bool({e for e in effects if not isinstance(e, core.NamedAxisEffect)})
 
-def remat_lowering(*args, jaxpr: core.Jaxpr, prevent_cse: bool,
-                   differentiated: bool, is_gpu_platform: bool = False,
-                   **_):
+
+def remat_expansion(*args, jaxpr: core.Jaxpr, prevent_cse: bool,
+                    differentiated: bool, is_gpu_platform: bool = False,
+                    **_):
   assert not jaxpr.constvars
 
   if differentiated and prevent_cse:
-    if config.jax_remat_opt_barrier:
+    if config.remat_opt_barrier.value:
       translation_rule = _remat_translation_using_opt_barrier
     elif is_gpu_platform:
       translation_rule = _remat_translation_using_while
@@ -668,7 +752,7 @@ def remat_lowering(*args, jaxpr: core.Jaxpr, prevent_cse: bool,
   else:
     translation_rule = lambda *args, jaxpr: core.eval_jaxpr(jaxpr, (), *args)
 
-  return api.named_call(translation_rule, name="remat")(*args, jaxpr=jaxpr)
+  return api.named_call(translation_rule, name="checkpoint")(*args, jaxpr=jaxpr)
 
 def _remat_translation_using_opt_barrier(*args, jaxpr: core.Jaxpr):
   args = _optimization_barrier(args)
@@ -724,22 +808,43 @@ def _remat_translation_using_cond(*args, jaxpr: core.Jaxpr):
   unif = lax_internal.rng_uniform(np.float32(0), np.float32(1), shape=())
   return lax_control_flow.cond(unif < np.float32(2), remat_comp, dummy_comp, *args)
 
-mlir.register_lowering(
-    remat_p, mlir.lower_fun(remat_lowering, multiple_results=True))
-mlir.register_lowering(
-    remat_p,
-    mlir.lower_fun(partial(remat_lowering, is_gpu_platform=True),
-                   multiple_results=True),
-    platform="gpu")
+def _remat_lowering(ctx, *args, jaxpr: core.Jaxpr, prevent_cse: bool,
+                   differentiated: bool, policy, is_gpu_platform=False):
+  jaxpr_args: Sequence[mlir.IrValues]
+  if differentiated and prevent_cse:
+    # If we're using the loop or cond lowerings, use the slower lower_fun
+    # based path.
+    if not config.remat_opt_barrier.value:
+      return mlir.lower_fun(remat_expansion, multiple_results=True)(
+          ctx, *args, jaxpr=jaxpr, prevent_cse=prevent_cse,
+          differentiated=differentiated, policy=policy,
+          is_gpu_platform=is_gpu_platform)
+
+    arg_types = map(mlir.aval_to_ir_type, ctx.avals_in)
+    flat_args = mlir.flatten_ir_values(args)
+    barrier_op = hlo.OptimizationBarrierOp(flat_args)
+    jaxpr_args = mlir.unflatten_ir_values_like_types(
+      barrier_op.results, arg_types)
+  else:
+    jaxpr_args = args
+  outs, tokens_out = mlir.jaxpr_subcomp(
+      ctx.module_context, jaxpr, ctx.name_stack.extend('checkpoint'),
+      ctx.tokens_in, (), *jaxpr_args, dim_var_values=ctx.dim_var_values)
+  ctx.set_tokens_out(tokens_out)
+  return outs
+
+mlir.register_lowering(remat_p, _remat_lowering)
+mlir.register_lowering(remat_p, partial(_remat_lowering, is_gpu_platform=True),
+                       platform="gpu")
 
 def _optimization_barrier_abstract_eval(*args):
   return args
 
 def _optimization_barrier_lowering_rule(ctx, *args):
-  barrier_types = map(mlir.aval_to_ir_types, ctx.avals_in)
-  flat_args = mlir.flatten_lowering_ir_args(args)
+  barrier_types = map(mlir.aval_to_ir_type, ctx.avals_in)
+  flat_args = mlir.flatten_ir_values(args)
   barrier_op = hlo.OptimizationBarrierOp(flat_args)
-  return util.unflatten(barrier_op.results, map(len, barrier_types))
+  return mlir.unflatten_ir_values_like_types(barrier_op.results, barrier_types)
 
 def _optimization_barrier(arg):
   flat_args, treedef = tree_flatten(arg)
@@ -779,8 +884,8 @@ def checkpoint_wrapper(
     *,
     concrete: bool = False,
     prevent_cse: bool = True,
-    static_argnums: Union[int, tuple[int, ...]] = (),
-    policy: Optional[Callable[..., bool]] = None,
+    static_argnums: int | tuple[int, ...] = (),
+    policy: Callable[..., bool] | None = None,
 ) -> Callable:
   if concrete:
     msg = ("The 'concrete' option to jax.checkpoint / jax.remat is deprecated; "
