@@ -13,13 +13,13 @@
 # limitations under the License.
 from __future__ import annotations
 
+from collections.abc import Callable
 import types
-from typing import Any, Callable, TypeVar, cast
+from typing import Any, TypeVar
 
 from jax._src import core
 from jax._src import traceback_util
-from jax._src.core import (lattice_join, Primitive, valid_jaxtype,
-                           raise_to_shaped, get_aval)
+from jax._src.core import Primitive, valid_jaxtype, raise_to_shaped, get_aval
 from jax._src.tree_util import register_pytree_node
 from jax._src.typing import Array, ArrayLike
 from jax._src.util import safe_map
@@ -30,44 +30,34 @@ T = TypeVar('T')
 
 map = safe_map
 
-jaxval_adders: dict[type, Callable[[ArrayLike, ArrayLike], Array]] = {}
-
 def add_jaxvals(x: ArrayLike, y: ArrayLike) -> Array:
+  dtype = core.get_aval(x).dtype
   return add_jaxvals_p.bind(x, y)
 
-add_jaxvals_p: Primitive = Primitive('add_any')
+add_jaxvals_p = Primitive('add_any')
 add_any_p = add_jaxvals_p
 
 @add_jaxvals_p.def_impl
-def add_impl(xs, ys):
-  return jaxval_adders[type(xs)](xs, ys)
+def add_impl(x, y):
+  return raw_jaxval_adders[type(x)](x, y)
+raw_jaxval_adders = {}  # type: ignore
 
 @add_jaxvals_p.def_abstract_eval
-def add_abstract(xs, ys):
-  return lattice_join(xs, ys)
-
-jaxval_zeros_likers: dict[type, Callable[[Any], Array]] = {}
-
-def instantiate(z: Zero | Array) -> Array:
-  if type(z) is Zero:
-    return zeros_like_aval(z.aval)
-  return cast(Array, z)
+def add_abstract(x, y):
+  return core.lattice_join(x, y)
 
 def zeros_like_aval(aval: core.AbstractValue) -> Array:
   return aval_zeros_likers[type(aval)](aval)
-
 aval_zeros_likers: dict[type, Callable[[Any], Array]] = {}
 
-def zeros_like_jaxval(val: ArrayLike) -> Array:
-  return zeros_like_p.bind(val)
+def zeros_like_jaxval(val):
+  return zeros_like_aval(core.raise_to_shaped(core.get_aval(val)))
 
-zeros_like_p: Primitive = Primitive('zeros_like')
+def instantiate(z: Zero | Array) -> Array:
+  if isinstance(z, Zero):
+    return zeros_like_aval(z.aval)
+  return z
 
-@zeros_like_p.def_impl
-def zeros_like_impl(example):
-  return jaxval_zeros_likers[type(example)](example)
-
-zeros_like_p.def_abstract_eval(lambda x: x)
 
 class Zero:
   __slots__ = ['aval']
@@ -128,3 +118,7 @@ def replace_internal_symbolic_zeros(
 def replace_rule_output_symbolic_zeros(
     x: JaxTypeOrTracer | SymbolicZero) -> JaxTypeOrTracer | Zero:
   return Zero(x.aval) if type(x) is SymbolicZero else x
+
+
+# TODO(mattjj): remove these after fixing downstream users relying on them
+zeros_like_p: Primitive = Primitive('zeros_like')
