@@ -14,12 +14,13 @@
 """Module for JAX callbacks."""
 from __future__ import annotations
 
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
+import dataclasses
 import functools
-from typing import Any, Callable
-
-import numpy as np
+import logging
+from typing import Any
 
+import jax
 from jax._src import core
 from jax._src import dispatch
 from jax._src import dtypes
@@ -30,26 +31,62 @@ from jax._src import util
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
-from jax._src.lib import xla_client as xc
 from jax._src.lax.control_flow.loops import map as lax_map
+from jax._src.lib import xla_client as xc
 from jax._src.sharding_impls import SingleDeviceSharding
+import numpy as np
+
+logger = logging.getLogger(__name__)
+
 
 # `pure_callback_p` is the main primitive for staging out Python pure callbacks.
 pure_callback_p = core.Primitive("pure_callback")
 pure_callback_p.multiple_results = True
+dispatch.prim_requires_devices_during_lowering.add(pure_callback_p)
 
 map, unsafe_map = util.safe_map, map
 
 
+@dataclasses.dataclass(frozen=True)
+class _FlatCallback:
+  """A Python function callable with flat arguments and results.
+
+  An instance of this class is used as a parameter for the callback primitives.
+  We prefer it to an anonymous flattened function because it produces
+  equal objects when we call the same Python function with the same argument
+  structure.
+  """
+  callback_func: Callable[..., Any]
+  in_tree: tree_util.PyTreeDef  # (args, kwargs) pytree for `callback_func`.
+
+  def __call__(self, *flat_args: jax.Array) -> Sequence[jax.Array]:
+    args, kwargs = tree_util.tree_unflatten(self.in_tree, flat_args)
+    return tree_util.tree_leaves(self.callback_func(*args, **kwargs))
+
+
 def pure_callback_impl(
     *args,
     result_avals,
-    callback: Callable[..., Any],
+    callback: _FlatCallback,
     sharding: SingleDeviceSharding | None,
     vectorized: bool,
 ):
   del sharding, vectorized, result_avals
-  return callback(*args)
+  try:
+    cpu_device, *_ = jax.local_devices(backend="cpu")
+  except RuntimeError as e:
+    raise RuntimeError(
+        "jax.pure_callback failed to find a local CPU device to place the"
+        " inputs on. Make sure \"cpu\" is listed in --jax_platforms or the"
+        " JAX_PLATFORMS environment variable."
+    ) from e
+  args = jax.device_put(args, cpu_device)
+  with jax.default_device(cpu_device):
+    try:
+      return tree_util.tree_map(np.asarray, callback(*args))
+    except BaseException:
+      logger.exception("jax.pure_callback failed")
+      raise
 
 
 pure_callback_p.def_impl(functools.partial(dispatch.apply_primitive,
@@ -59,7 +96,7 @@ pure_callback_p.def_impl(functools.partial(dispatch.apply_primitive,
 @pure_callback_p.def_abstract_eval
 def pure_callback_abstract_eval(
     *avals,
-    callback: Callable[..., Any],
+    callback: _FlatCallback,
     result_avals,
     sharding: SingleDeviceSharding | None,
     vectorized: bool,
@@ -87,16 +124,16 @@ def pure_callback_transpose_rule(*args, **kwargs):
 ad.primitive_transposes[pure_callback_p] = pure_callback_transpose_rule
 
 
-def pure_callback_batching_rule(
+def callback_batching_rule(
+    prim,
     args,
     dims,
     *,
-    callback,
-    sharding: SingleDeviceSharding | None,
     vectorized: bool,
     result_avals: Sequence[core.ShapedArray],
+    **kwargs: Any,
 ):
-  axis_size = next(a.shape[0] for a, d in zip(args, dims)
+  axis_size = next(a.shape[d] for a, d in zip(args, dims)
                    if d is not batching.not_mapped)
   new_args = [arg if dim is batching.not_mapped else
               batching.moveaxis(arg, dim, 0) for arg, dim in zip(args, dims)]
@@ -104,30 +141,30 @@ def pure_callback_batching_rule(
     result_avals = tuple(
         core.unmapped_aval(axis_size, core.no_axis_name, 0, aval)  # type: ignore
         for aval in result_avals)
-    outvals = pure_callback_p.bind(
+    outvals = prim.bind(
         *new_args,
-        callback=callback,
-        sharding=sharding,
         vectorized=vectorized,
         result_avals=result_avals,
+        **kwargs,
     )
   else:
     is_batched = [d is not batching.not_mapped for d in dims]
     unbatched_args, batched_args = util.partition_list(is_batched, new_args)
     def _batch_fun(batched_args):
       merged_args = util.merge_lists(is_batched, unbatched_args, batched_args)
-      return pure_callback_p.bind(
+      return prim.bind(
           *merged_args,
-          callback=callback,
-          sharding=sharding,
           result_avals=result_avals,
           vectorized=vectorized,
+          **kwargs,
       )
     outvals = lax_map(_batch_fun, batched_args)
   return tuple(outvals), (0,) * len(outvals)
 
 
-batching.primitive_batchers[pure_callback_p] = pure_callback_batching_rule
+batching.primitive_batchers[pure_callback_p] = functools.partial(
+    callback_batching_rule, pure_callback_p
+)
 
 
 def _callback_op_sharding(axis_context, sharding: SingleDeviceSharding | None):
@@ -156,14 +193,17 @@ def _callback_op_sharding(axis_context, sharding: SingleDeviceSharding | None):
             f" {type(sharding)}"
         )
       device = next(iter(sharding.device_set))
+      device_assignment = axis_context.device_assignment
+      if device_assignment is None:
+        raise AssertionError(
+            "Please file a bug at https://github.com/google/jax/issues")
       try:
-        device_index = axis_context.device_assignment.index(device)
+        device_index = device_assignment.index(device)
       except IndexError as e:
         raise ValueError(
             "Sharding provided to pure_callback specifies a device"
             f" {device} that is not in the device assignment"
-            f" ({axis_context.device_assignment})"
-        ) from e
+            f" ({device_assignment})") from e
     else:
       device_index = 0
 
@@ -181,7 +221,7 @@ def _callback_op_sharding(axis_context, sharding: SingleDeviceSharding | None):
 
 
 def pure_callback_lowering(
-    ctx, *args, callback, sharding: SingleDeviceSharding | None, **params
+    ctx, *args, callback: _FlatCallback, sharding: SingleDeviceSharding | None, **params
 ):
   def _callback(*flat_args):
     return tuple(
@@ -194,17 +234,16 @@ def pure_callback_lowering(
     )
 
   op_sharding = _callback_op_sharding(ctx.module_context.axis_context, sharding)
-  result, _, keepalive = mlir.emit_python_callback(
+  result, _, _ = mlir.emit_python_callback(
       ctx,
       _callback,
       None,
       list(args),
       ctx.avals_in,
       ctx.avals_out,
-      False,
+      has_side_effect=False,
       sharding=op_sharding,
   )
-  ctx.module_context.add_keepalive(keepalive)
   return result
 
 
@@ -214,7 +253,7 @@ def _check_shape_dtype(shape_dtype):
   dt = np.dtype(shape_dtype.dtype)
   if dtypes.canonicalize_dtype(dt) != dt:
     raise ValueError(
-        "Cannot return 64-bit values when `jax_enable_x64` is disabled")
+        "result_shape_dtypes cannot specify 64-bit types when `jax_enable_x64` is disabled")
 
 
 def pure_callback(
@@ -225,16 +264,42 @@ def pure_callback(
     vectorized: bool = False,
     **kwargs: Any,
 ):
-  """Calls a pure Python callback.
+  """Calls a pure Python callback. Works under :func:`jit`/:func:`~vmap`/etc.
 
   For more explanation, see `External Callbacks`_.
 
+  ``pure_callback`` enables calling a Python function in JIT-ed JAX functions.
+  The input ``callback`` will be passed JAX arrays placed on a local CPU, and
+  it should also return JAX arrays on CPU.
+
+  The callback is treated as functionally pure, meaning it has no side-effects
+  and its output value depends only on its argument values. As a consequence, it
+  is safe to be called multiple times (e.g. when transformed by :func:`~vmap` or
+  :func:`~pmap`), or not to be called at all when e.g. the output of a
+  `jit`-decorated function has no data dependence on its value. Pure callbacks
+  may also be reordered if data-dependence allows.
+
+  When `vmap`-ed the behavior will depend on the value of the
+  ``vectorized`` keyword argument. When ``vectorized`` is ``True``, the callback
+  is assumed to obey
+  ``jax.vmap(callback)(xs) == callback(xs) == jnp.stack([callback(x) for x in xs])``.
+  Therefore, the callback will be called directly on batched inputs (where the
+  batch axes are the leading dimensions). Additionally, the callbacks should
+  return outputs that have corresponding leading batch axes. If not vectorized
+  ``callback`` will be mapped sequentially across the batched axis.
+  For example, if ``callback = lambda x, y: np.matmul(x, y)``, then we are free
+  to set ``vectorized=True`` because the ``np.matmul`` function handles
+  arbitrary leading batch dimensions.
+
   Args:
     callback: function to execute on the host. The callback is assumed to be a pure
       function (i.e. one without side-effects): if an impure function is passed, it
-      may behave in unexpected ways, particularly under transformation.
+      may behave in unexpected ways, particularly under transformation. The callable
+      will be passed PyTrees of arrays as arguments, and should return a PyTree of
+      arrays that matches ``result_shape_dtypes``.
     result_shape_dtypes: pytree whose leaves have ``shape`` and ``dtype`` attributes,
       whose structure matches the expected output of the callback function at runtime.
+      :class:`jax.ShapeDtypeStruct` is often used to define leaf values.
     *args: arguments to be passed to the callback function
     sharding: optional sharding that specifies the device from which the callback should
       be invoked.
@@ -253,10 +318,6 @@ def pure_callback(
 
   .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
   """
-  def _flat_callback(*flat_args):
-    args, kwargs = tree_util.tree_unflatten(in_tree, flat_args)
-    return tree_util.tree_leaves(callback(*args, **kwargs))
-
   flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
   tree_util.tree_map(_check_shape_dtype, result_shape_dtypes)
   result_avals = tree_util.tree_map(
@@ -264,7 +325,7 @@ def pure_callback(
   flat_result_avals, out_tree = tree_util.tree_flatten(result_avals)
   out_flat = pure_callback_p.bind(
       *flat_args,
-      callback=_flat_callback,
+      callback=_FlatCallback(callback, in_tree),
       result_avals=tuple(flat_result_avals),
       sharding=sharding,
       vectorized=vectorized,
@@ -272,79 +333,11 @@ def pure_callback(
   return tree_util.tree_unflatten(out_tree, out_flat)
 
 
-def pure_callback_api(
-    callback: Callable[..., Any],
-    result_shape_dtypes: Any,
-    *args: Any,
-    sharding: SingleDeviceSharding | None = None,
-    vectorized: bool = False,
-    **kwargs: Any,
-):
-  """Applies a functionally pure Python callable. Works under :func:`jit`/:func:`~pmap`/etc.
-
-  ``pure_callback`` enables calling a Python function in JIT-ed JAX functions.
-  The input ``callback`` will be passed NumPy arrays in place of JAX arrays and
-  should also return NumPy arrays. Execution takes place on CPU, like any
-  Python+NumPy function.
-
-  The callback is treated as functionally pure, meaning it has no side-effects
-  and its output value depends only on its argument values. As a consequence, it
-  is safe to be called multiple times (e.g. when transformed by :func:`~vmap` or
-  :func:`~pmap`), or not to be called at all when e.g. the output of a
-  `jit`-decorated function has no data dependence on its value. Pure callbacks
-  may also be reordered if data-dependence allows.
-
-  When :func:`~pmap`-ed, the pure callback will be called several times (one on each
-  axis of the map). When `vmap`-ed the behavior will depend on the value of the
-  ``vectorized`` keyword argument. When ``vectorized`` is ``True``, the callback
-  is assumed to obey
-  ``jax.vmap(callback)(xs) == callback(xs) == jnp.stack([callback(x) for x in xs])``.
-  Therefore, the callback will be called directly on batched inputs (where the
-  batch axes are the leading dimensions). Additionally, the callbacks should
-  return outputs that have corresponding leading batch axes. If not vectorized
-  ``callback`` will be mapped sequentially across the batched axis.
-  For example, if ``callback = lambda x, y: np.matmul(x, y)``, then we are free
-  to set ``vectorized=True`` because the ``np.matmul`` function handles
-  arbitrary leading batch dimensions.
-
-  Args:
-    callback: A Python callable. The callable will be passed PyTrees of NumPy
-      arrays as arguments, and should return a PyTree of NumPy arrays that
-      matches ``result_shape_dtypes``.
-    result_shape_dtypes: A PyTree with leaves that are objects with ``shape``
-      and ``dtype`` attributes which represent to the shapes and dtypes of the
-      value of ``callback`` applied to ``args`` and ``kwargs``.
-    *args: The positional arguments to the callback. Must be PyTrees of JAX
-      types.
-    sharding: optional sharding that specifies the device from which the
-      callback should be invoked.
-    vectorized: A boolean that indicates whether or not ``callback`` is
-      vectorized, meaning it can handle arrays with additional leading
-      dimensions. If ``vectorized`` is `True`, when the callback is mapped
-      via `jax.vmap`, it will be called directly on inputs with leading batch
-      dimensions instead of executing ``callback`` on each mapped input
-      individually. The callback should also return outputs batched across the
-      leading axis. By default, ``vectorized`` is ``False``.
-    **kwargs: The keyword arguments to the callback. Must be PyTrees of JAX
-      types.
-
-  Returns:
-    The value of ``callback(*args, **kwargs)``.
-  """
-  return pure_callback(
-      callback,
-      result_shape_dtypes,
-      *args,
-      sharding=sharding,
-      vectorized=vectorized,
-      **kwargs,
-  )
-
-
 # IO Callback
 
 io_callback_p = core.Primitive("io_callback")
 io_callback_p.multiple_results = True
+dispatch.prim_requires_devices_during_lowering.add(io_callback_p)
 
 class IOEffect(effects.Effect):
   __str__ = lambda _: "IO"
@@ -365,12 +358,26 @@ effects.shardable_ordered_effects.add_type(OrderedIOEffect)
 def io_callback_impl(
     *args,
     result_avals,
-    callback: Callable[..., Any],
+    callback: _FlatCallback,
     sharding: SingleDeviceSharding | None,
     ordered: bool,
 ):
   del result_avals, sharding, ordered
-  return callback(*args)
+  try:
+    cpu_device, *_ = jax.local_devices(backend="cpu")
+  except RuntimeError as e:
+    raise RuntimeError(
+        "jax.io_callback failed to find a local CPU device to place the"
+        " inputs on. Make sure \"cpu\" is listed in --jax_platforms or the"
+        " JAX_PLATFORMS environment variable."
+    ) from e
+  args = jax.device_put(args, cpu_device)
+  with jax.default_device(cpu_device):
+    try:
+      return tree_util.tree_map(np.asarray, callback(*args))
+    except BaseException:
+      logger.exception("jax.io_callback failed")
+      raise
 
 
 io_callback_p.def_impl(functools.partial(dispatch.apply_primitive,
@@ -380,7 +387,7 @@ io_callback_p.def_impl(functools.partial(dispatch.apply_primitive,
 @io_callback_p.def_effectful_abstract_eval
 def io_callback_abstract_eval(
     *avals,
-    callback: Callable[..., Any],
+    callback: _FlatCallback,
     result_avals,
     sharding: SingleDeviceSharding | None,
     ordered: bool,
@@ -407,16 +414,16 @@ def io_callback_batching_rule(
 ):
   if ordered:
     raise ValueError("Cannot `vmap` ordered IO callback.")
-  return pure_callback_batching_rule(
-      args,
-      dims,
-      callback=callback,
-      sharding=sharding,
-      vectorized=False,
-      result_avals=result_avals,
-  )
-
-
+  is_batched = [d is not batching.not_mapped for d in dims]
+  new_args = [arg if dim is batching.not_mapped else
+              batching.moveaxis(arg, dim, 0) for arg, dim in zip(args, dims)]
+  unbatched_args, batched_args = util.partition_list(is_batched, new_args)
+  def _batch_fun(batched_args):
+    merged = util.merge_lists(is_batched, unbatched_args, batched_args)
+    return io_callback_p.bind(*merged, callback=callback, sharding=sharding,
+                              result_avals=result_avals, ordered=False)
+  out_vals = lax_map(_batch_fun, batched_args)
+  return out_vals, (0,) * len(out_vals)
 batching.primitive_batchers[io_callback_p] = io_callback_batching_rule
 
 
@@ -434,30 +441,29 @@ def io_callback_lowering(ctx, *args, callback, sharding, ordered, **params):
 
   op_sharding = _callback_op_sharding(ctx.module_context.axis_context, sharding)
   if ordered:
-    token = ctx.tokens_in.get(_OrderedIOEffect)[0]
-    result, token, keepalive = mlir.emit_python_callback(
+    token = ctx.tokens_in.get(_OrderedIOEffect)
+    result, token, _ = mlir.emit_python_callback(
         ctx,
         _callback,
         token,
         list(args),
         ctx.avals_in,
         ctx.avals_out,
-        True,
+        has_side_effect=True,
         sharding=op_sharding,
     )
-    ctx.set_tokens_out(mlir.TokenSet({_OrderedIOEffect: (token,)}))
+    ctx.set_tokens_out(mlir.TokenSet({_OrderedIOEffect: token}))
   else:
-    result, token, keepalive = mlir.emit_python_callback(
+    result, token, _ = mlir.emit_python_callback(
         ctx,
         _callback,
         None,
         list(args),
         ctx.avals_in,
         ctx.avals_out,
-        True,
+        has_side_effect=True,
         sharding=op_sharding,
     )
-  ctx.module_context.add_keepalive(keepalive)
   return result
 
 
@@ -477,11 +483,12 @@ def io_callback(
   For more explanation, see `External Callbacks`_.
 
   Args:
-    callback: function to execute on the host. It is assumet to be an impure function.
+    callback: function to execute on the host. It is assumed to be an impure function.
       If ``callback`` is pure, using :func:`jax.pure_callback` instead may lead to
       more efficient execution.
     result_shape_dtypes: pytree whose leaves have ``shape`` and ``dtype`` attributes,
       whose structure matches the expected output of the callback function at runtime.
+      :class:`jax.ShapeDtypeStruct` is often used to define leaf values.
     *args: arguments to be passed to the callback function
     sharding: optional sharding that specifies the device from which the callback should
       be invoked.
@@ -499,10 +506,6 @@ def io_callback(
 
   .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
   """
-  def _flat_callback(*flat_args):
-    args, kwargs = tree_util.tree_unflatten(in_tree, flat_args)
-    return tree_util.tree_leaves(callback(*args, **kwargs))
-
   flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
   tree_util.tree_map(_check_shape_dtype, result_shape_dtypes)
   flat_shape_dtypes, out_tree = tree_util.tree_flatten(result_shape_dtypes)
@@ -511,7 +514,7 @@ def io_callback(
   flat_args = map(core.raise_as_much_as_possible, flat_args)
   out_flat = io_callback_p.bind(
       *flat_args,
-      callback=_flat_callback,
+      callback=_FlatCallback(callback, in_tree),
       result_avals=tuple(flat_result_avals),
       sharding=sharding,
       ordered=ordered,
