@@ -13,7 +13,9 @@
 # limitations under the License.
 
 import os
-import warnings
+from jax import version
+from jax._src import config
+from jax._src import hardware_utils
 
 running_in_cloud_tpu_vm: bool = False
 
@@ -32,6 +34,18 @@ def maybe_import_libtpu():
     return libtpu
 
 
+def get_tpu_library_path() -> str | None:
+  path_from_env = os.getenv("TPU_LIBRARY_PATH")
+  if path_from_env is not None and os.path.isfile(path_from_env):
+    return path_from_env
+
+  libtpu_module = maybe_import_libtpu()
+  if libtpu_module is not None:
+    return libtpu_module.get_library_path()
+
+  return None
+
+
 def jax_force_tpu_init() -> bool:
   return 'JAX_FORCE_TPU_INIT' in os.environ
 
@@ -54,28 +68,33 @@ def cloud_tpu_init() -> None:
   """
   global running_in_cloud_tpu_vm
 
-  # We assume we are in a correctly-configured Cloud TPU environment
-  # if the following hold: a) libtpu is installed b) JAX_FORCE_TPU_INIT is set
-  # Exit early if we're not running on Cloud TPU.
-  libtpu_module = maybe_import_libtpu()
-  if libtpu_module is not None:
-    libtpu_module.configure_library_path()
-  elif not jax_force_tpu_init():
+  # Exit early if we're not running on a Cloud TPU VM or libtpu isn't installed.
+  libtpu_path = get_tpu_library_path()
+  num_tpu_chips = hardware_utils.num_available_tpu_chips_and_device_id()[0]
+  if (libtpu_path is None or num_tpu_chips == 0) and not jax_force_tpu_init():
     return
 
   running_in_cloud_tpu_vm = True
 
   os.environ.setdefault('GRPC_VERBOSITY', 'ERROR')
-  os.environ.setdefault('JAX_PLATFORMS', 'tpu,cpu')
   os.environ['TPU_ML_PLATFORM'] = 'JAX'
+  os.environ['TPU_ML_PLATFORM_VERSION'] = version.__version__
+  os.environ.setdefault('ENABLE_RUNTIME_UPTIME_TELEMETRY', '1')
+  if hardware_utils.tpu_enhanced_barrier_supported():
+    os.environ["LIBTPU_INIT_ARGS"] = os.environ.get("LIBTPU_INIT_ARGS","") + " --xla_tpu_use_enhanced_launch_barrier=true"
 
-  # TODO(skyewm): remove this warning at some point, say around Sept 2023.
-  use_pjrt_c_api = os.environ.get('JAX_USE_PJRT_C_API_ON_TPU', None)
-  if use_pjrt_c_api:
-    warnings.warn(
-        "JAX_USE_PJRT_C_API_ON_TPU no longer has an effect (the new TPU "
-        "runtime is always enabled now). Unset the environment variable "
-        "to disable this warning.")
+  # this makes tensorstore serialization work better on TPU
+  os.environ.setdefault('TENSORSTORE_CURL_LOW_SPEED_TIME_SECONDS', '60')
+  os.environ.setdefault('TENSORSTORE_CURL_LOW_SPEED_LIMIT_BYTES', '256')
 
-  # Remove when minimum jaxlib version is >= 0.4.15
-  os.environ['JAX_USE_PJRT_C_API_ON_TPU'] = "true"
+  # If the JAX_PLATFORMS env variable isn't set, config.jax_platforms defaults
+  # to None. In this case, we set it to 'tpu,cpu' to ensure that JAX uses the
+  # TPU backend.
+  if config.jax_platforms.value is None:
+    config.update('jax_platforms', 'tpu,cpu')
+
+  if config.jax_pjrt_client_create_options.value is None:
+    config.update(
+      'jax_pjrt_client_create_options',
+      f'ml_framework_name:JAX;ml_framework_version:{version.__version__}'
+      )
