@@ -22,4 +22,6 @@ from .cluster import ClusterEnv
 # available one from the list will be picked.
 from .ompi_cluster import OmpiCluster
 from .slurm_cluster import SlurmCluster
-from .cloud_tpu_cluster import TpuCluster
+from .mpi4py_cluster import Mpi4pyCluster
+from .cloud_tpu_cluster import GkeTpuCluster
+from .cloud_tpu_cluster import GceTpuCluster
