@@ -12,19 +12,20 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import annotations
+
+from collections.abc import Callable
 import functools
 import operator
-from typing import Callable, Optional
 
-import jax
-from jax import tree_util
-from jax.tree_util import (tree_flatten, tree_map, tree_structure,
-                           tree_unflatten, treedef_tuple)
+from jax import lax
+from jax._src import api
 from jax._src import core
 from jax._src import custom_api_util
 from jax._src import linear_util as lu
 from jax._src import source_info_util
 from jax._src import traceback_util
+from jax._src import tree_util
 from jax._src import util
 from jax._src.api_util import flatten_fun_nokwargs
 from jax._src.interpreters import ad
@@ -33,6 +34,8 @@ from jax._src.interpreters.batching import not_mapped
 from jax._src.interpreters import mlir
 from jax._src.interpreters import partial_eval as pe
 from jax._src.interpreters import xla
+from jax._src.tree_util import (tree_flatten, tree_map, tree_structure,
+                                tree_unflatten, treedef_tuple)
 
 
 source_info_util.register_exclusion(__file__)
@@ -46,11 +49,11 @@ zip, unsafe_zip = util.safe_zip, zip
 @custom_api_util.register_custom_decorator_type
 class custom_vmap:
   fun: Callable
-  vmap_rule: Optional[Callable]
+  vmap_rule: Callable | None
 
   def __init__(self, fun: Callable):
     functools.update_wrapper(self, fun)
-    self.fun = fun  # type: ignore[assignment]
+    self.fun = fun
     self.vmap_rule = None
 
   __getattr__ = custom_api_util.forward_attr
@@ -66,9 +69,10 @@ class custom_vmap:
     flat_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(self.fun), in_tree)
     in_avals = [core.raise_to_shaped(core.get_aval(x)) for x in args_flat]
     debug = pe.debug_info(self.fun, in_tree, out_tree, False, "custom_vmap")
-    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals, debug)
+    jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(flat_fun, in_avals, debug)
     closed_call = core.ClosedJaxpr(pe.convert_constvars_jaxpr(jaxpr), ())
     in_tree = treedef_tuple((tree_structure(consts), in_tree))
+    assert self.vmap_rule is not None
     out_flat = custom_vmap_p.bind(*consts, *args_flat,
                                   call=closed_call,
                                   rule=ClosedRule(self.vmap_rule),
@@ -194,7 +198,7 @@ def custom_vmap_jvp(primals, tangents, *, call, rule, in_tree, out_tree):
       return out
 
     def to_vmap_over_extra_batched_dims(primals, tangents):
-      return jax.jvp(to_jvp, primals, tangents)
+      return api.jvp(to_jvp, primals, tangents)
 
     to_vmap_over_extra_batched_dims_flat, out_tree2 = flatten_fun_nokwargs(
         lu.wrap_init(to_vmap_over_extra_batched_dims),
@@ -274,7 +278,7 @@ def sequential_vmap(f):
       return f(*args)
 
     mapped_args, bcast_args = tree_split(in_batched, list(args))
-    out = jax.lax.map(to_map, mapped_args)
+    out = lax.map(to_map, mapped_args)
     out_batched = tree_map(lambda _: True, out)
     return out, out_batched
 
