@@ -12,12 +12,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections.abc import Sequence
+from __future__ import annotations
+
+from collections.abc import Callable, Sequence
 import dataclasses
 from functools import update_wrapper, reduce, partial
 import inspect
-from typing import Any, Callable, Generic, Optional, TypeVar
+from typing import Any, Generic, TypeVar
 
+from jax._src import config
 from jax._src import core
 from jax._src import custom_api_util
 from jax._src.custom_transpose import custom_transpose
@@ -28,7 +31,6 @@ from jax._src import traceback_util
 from jax._src.ad_util import (
     stop_gradient_p, SymbolicZero, Zero, zeros_like_aval)
 from jax._src.api_util import argnums_partial, flatten_fun_nokwargs
-from jax._src.config import config
 from jax._src.core import raise_to_shaped
 from jax._src.errors import UnexpectedTracerError
 from jax._src.interpreters import ad
@@ -38,10 +40,12 @@ from jax._src.interpreters import partial_eval as pe
 from jax._src.interpreters import xla
 from jax._src.interpreters.batching import not_mapped
 from jax._src.lax import lax
-from jax._src.tree_util import (tree_flatten, tree_unflatten, tree_map,
-                                treedef_is_leaf, treedef_tuple,
-                                register_pytree_node_class, tree_leaves)
-from jax._src.util import cache, safe_zip, safe_map, split_list, Unhashable
+from jax._src.tree_util import (
+    tree_flatten, tree_unflatten, tree_map, treedef_is_leaf, treedef_tuple,
+    register_pytree_node_class, tree_leaves, tree_flatten_with_path, keystr,
+    treedef_children)
+from jax._src.util import (cache, safe_zip, safe_map, split_list, Unhashable,
+                           unzip2)
 
 
 traceback_util.register_exclusion(__file__)
@@ -64,11 +68,11 @@ def _resolve_kwargs(fun, args, kwargs):
     return ba.args
 
 def _initial_style_jaxpr(fun, in_avals):
-  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(fun, in_avals)
+  jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(fun, in_avals)
   return jaxpr, consts
 
 def _close_jaxpr(jaxpr):
-  return core.ClosedJaxpr(pe.convert_constvars_jaxpr(jaxpr), ())
+  return pe.close_jaxpr(pe.convert_constvars_jaxpr(jaxpr))
 
 def _sum_tangents(_, x, *xs):
   return reduce(ad.add_tangents, xs, x)
@@ -136,7 +140,7 @@ class custom_jvp(Generic[ReturnValue]):
   """
   fun: Callable[..., ReturnValue]
   nondiff_argnums: tuple[int, ...]
-  jvp: Optional[Callable[..., tuple[ReturnValue, ReturnValue]]] = None
+  jvp: Callable[..., tuple[ReturnValue, ReturnValue]] | None = None
   symbolic_zeros: bool = False
 
   def __init__(self,
@@ -176,7 +180,7 @@ class custom_jvp(Generic[ReturnValue]):
     Returns:
       None.
 
-    Example::
+    Examples:
 
       @jax.custom_jvp
       def f(x, y):
@@ -194,7 +198,7 @@ class custom_jvp(Generic[ReturnValue]):
     self.symbolic_zeros = symbolic_zeros
     return jvp
 
-  def defjvps(self, *jvps: Optional[Callable[..., ReturnValue]]):
+  def defjvps(self, *jvps: Callable[..., ReturnValue] | None):
     """Convenience wrapper for defining JVPs for each argument separately.
 
     This convenience wrapper cannot be used together with ``nondiff_argnums``.
@@ -208,7 +212,7 @@ class custom_jvp(Generic[ReturnValue]):
     Returns:
       None.
 
-    Example::
+    Examples:
 
       @jax.custom_jvp
       def f(x, y):
@@ -248,7 +252,7 @@ class custom_jvp(Generic[ReturnValue]):
       static_args = [args[i] for i in self.nondiff_argnums]
       jvp = _add_args(lu.wrap_init(self.jvp), static_args)
     else:
-      f_, dyn_args = lu.wrap_init(self.fun), args  # type: ignore
+      f_, dyn_args = lu.wrap_init(self.fun), args
       jvp = lu.wrap_init(self.jvp)
     args_flat, in_tree = tree_flatten(dyn_args)
     flat_fun, out_type1 = _flatten_fun_nokwargs(f_, in_tree)
@@ -323,13 +327,11 @@ def _flatten_jvp(primal_name, jvp_name, in_tree, maybe_out_type, *args):
            f"""    {str(ty_tree_).replace("'", "")}""")
       raise TypeError(m)
   # TODO(mattjj): compare primals' tangent types to tangent objects' types
-  primal_avals_out = [
-      raise_to_shaped(core.get_aval(x), weak_type=False).strip_named_shape()
-      for x in primals_out]
-  tangent_avals_out = [
-      raise_to_shaped(core.get_aval(t), weak_type=False).strip_named_shape()
-      if type(t) is not SymbolicZero else t.aval.strip_weak_type()
-      for t in tangents_out]
+  primal_avals_out = [raise_to_shaped(core.get_aval(x), weak_type=False)
+                      for x in primals_out]
+  tangent_avals_out = [raise_to_shaped(core.get_aval(t), weak_type=False)
+                       if type(t) is not SymbolicZero else t.aval.strip_weak_type()
+                       for t in tangents_out]
   if primal_avals_out != tangent_avals_out:
     if len(primal_avals_out) == 1:
       (av1,), (av2,) = primal_avals_out, tangent_avals_out
@@ -355,9 +357,9 @@ class CustomJVPCallPrimitive(core.Primitive):
         fun, self, top_trace and top_trace.level, False)
     jvp, env_trace_todo2 = process_env_traces(
         jvp, self, top_trace and top_trace.level, True)
-    tracers = map(top_trace.full_raise, args)  # type: ignore
-    outs = top_trace.process_custom_jvp_call(self, fun, jvp, tracers,  # type: ignore
-                                             symbolic_zeros=symbolic_zeros)  # type: ignore
+    tracers = map(top_trace.full_raise, args)
+    outs = top_trace.process_custom_jvp_call(self, fun, jvp, tracers,
+                                             symbolic_zeros=symbolic_zeros)
     _, env_trace_todo = lu.merge_linear_aux(env_trace_todo1, env_trace_todo2)
     return core.apply_todos(env_trace_todo, map(core.full_lower, outs))
 
@@ -432,11 +434,10 @@ core.custom_typechecks[custom_jvp_call_p] = _custom_jvp_call_typecheck
 def _custom_jvp_call_mlir_translation(ctx, *args, call_jaxpr, jvp_jaxpr_thunk,
                                       num_consts, symbolic_zeros):
   del jvp_jaxpr_thunk, num_consts, symbolic_zeros
-  args_ = map(mlir.wrap_singleton_ir_values, args)
   consts = mlir._ir_consts(call_jaxpr.consts)
   out, tokens = mlir.jaxpr_subcomp(ctx.module_context, call_jaxpr.jaxpr,
-                                   ctx.tokens_in, consts, *args_,
-                                   dim_var_values=ctx.dim_var_values)
+                                   ctx.name_stack, ctx.tokens_in, consts,
+                                   *args, dim_var_values=ctx.dim_var_values)
   ctx.set_tokens_out(tokens)
   return out
 mlir.register_lowering(custom_jvp_call_p, _custom_jvp_call_mlir_translation)
@@ -444,9 +445,9 @@ mlir.register_lowering(custom_jvp_call_p, _custom_jvp_call_mlir_translation)
 # If a (multi)linear function is defined with a custom jvp, then
 # custom_jvp_call_ can appear in jaxprs to be transposed. Since it's already
 # been linearized, we can drop the jvp rule.
-def _custom_jvp_call_transpose(params, jaxpr, args, ct, _, reduce_axes):
+def _custom_jvp_call_transpose(params, jaxpr, args, ct, _):
   del params
-  return ad.backward_pass(jaxpr.jaxpr, reduce_axes, None, jaxpr.consts, args, ct)
+  return ad.backward_pass(jaxpr.jaxpr, None, jaxpr.consts, args, ct)
 ad.primitive_transposes[custom_jvp_call_p] = _custom_jvp_call_transpose
 
 
@@ -493,8 +494,8 @@ class custom_vjp(Generic[ReturnValue]):
     update_wrapper(self, fun)
     self.fun = fun
     self.nondiff_argnums = nondiff_argnums
-    self.fwd: Optional[Callable[..., tuple[ReturnValue, Any]]] = None
-    self.bwd: Optional[Callable[..., tuple[Any, ...]]] = None
+    self.fwd: Callable[..., tuple[ReturnValue, Any]] | None = None
+    self.bwd: Callable[..., tuple[Any, ...]] | None = None
     self.symbolic_zeros = False
 
   __getattr__ = custom_api_util.forward_attr
@@ -563,7 +564,7 @@ class custom_vjp(Generic[ReturnValue]):
     Returns:
       None.
 
-    Example::
+    Examples:
 
       @jax.custom_vjp
       def f(x, y):
@@ -590,7 +591,7 @@ class custom_vjp(Generic[ReturnValue]):
       raise AttributeError(msg)
     fwd_name = getattr(self.fwd, '__name__', str(self.fwd))
     args = _resolve_kwargs(self.fun, args, kwargs)
-    if config.jax_enable_custom_vjp_by_custom_transpose:
+    if config.enable_custom_vjp_by_custom_transpose.value:
       if self.nondiff_argnums:
         raise NotImplementedError(
             'nondiff_argnums not implemented for new custom_vjp')
@@ -725,33 +726,74 @@ def _flatten_bwd(in_tree, in_avals, out_trees, *args):
   py_res = tree_unflatten(res_tree, res)
   py_cts_out = tree_unflatten(out_tree, cts_out)
   py_cts_in = yield (py_res, py_cts_out), {}
+  if isinstance(py_cts_in, list) and len(py_cts_in) == len(treedef_children(in_tree)):
+    py_cts_in = tuple(py_cts_in)
   # For each None in py_cts_in, indicating an argument for which the rule
   # produces no cotangent, we replace it with a pytree with the structure of the
   # corresponding subtree of in_tree and with leaves of a non-pytree sentinel
   # object, to be replaced with Nones in the final returned result.
   zero = object()  # non-pytree sentinel to replace Nones in py_cts_in
   dummy = tree_unflatten(in_tree, [object()] * in_tree.num_leaves)
+  keypaths, _ = unzip2(tree_flatten_with_path(dummy)[0])
   cts_in_flat = []
-  append = lambda x, d: cts_in_flat.extend([x] * len(tree_flatten(d)[0])) or x
+  def append(x, d):
+    num_leaves = len(tree_flatten(d)[0])
+    if x is None and d is not None:
+      cts_in_flat.extend([zero] * num_leaves)
+    elif x is not None:
+      cts_in_flat.extend([x] * num_leaves)
+    return x
   try:
     if not isinstance(py_cts_in, tuple):
       raise ValueError
-    tree_map(append,
-             tuple(zero if ct is None else ct for ct in py_cts_in), dummy)
+    tree_map(append, py_cts_in, dummy, is_leaf=lambda x: x is None)
   except ValueError:
     _, in_tree2 = tree_flatten(py_cts_in)
-    msg = ("Custom VJP rule must produce an output with the same container "
+    msg = ("Custom VJP bwd rule must produce an output with the same container "
            "(pytree) structure as the args tuple of the primal function, "
            "and in particular must produce a tuple of length equal to the "
-           "number of arguments to the primal function, but got VJP output "
+           "number of arguments to the primal function, but got bwd output "
            "structure {} for primal input structure {}.")
     raise TypeError(msg.format(in_tree2, in_tree)) from None
-  # Ignore any None cotangents, and any corresponding to inputs for which the
-  # type doesn't equal the tangent type (i.e. float0s)
-  # TODO(mattjj): change this to check if tangent type represents 0dim vspace
-  yield [Zero(a.at_least_vspace()) if ct is zero or a != a.at_least_vspace()
-         else ct for a, ct in zip(in_avals, cts_in_flat)]
+  results = []
+  for kp, a, ct in zip(keypaths, in_avals, cts_in_flat):
+    if ct is zero or a != a.at_least_vspace():
+      results.append(Zero(a.at_least_vspace()))
+    elif type(ct) is SymbolicZero:
+      if not core.typecompat(a.at_least_vspace(), a_ := ct.aval):
+        msg = ("Custom VJP bwd rule produced a SymbolicZero with a shape/dtype "
+               "that does not match the corresponding input tangent shape/dtype: "
+               f"at output{keystr(kp)} the SymbolicZero had shape/dtype "
+               f"{a_.str_short()} while the "
+               f"corresponding input had shape/dtype {a.str_short()}. "
+               "Consider just returning a None here instead of a SymbolicZero "
+               "object.")
+        raise ValueError(msg)
+      results.append(Zero(ct.aval))
+    else:
+      if (not core.typecompat(a.at_least_vspace(), a_ := core.get_aval(ct))
+          and not (_temporary_dtype_exception(a, a_) or
+                   _temporary_shape_exception(a, a_))):
+        msg = ("Custom VJP bwd rule must produce an output with the same "
+               "shape/dtypes as the args tuple of the primal function, but at "
+               f"output{keystr(kp)} the bwd rule produced an output of "
+               f"shape/dtype {raise_to_shaped(a_).str_short()} corresponding "
+               f"to an input of shape/dtype {a.str_short()}.")
+        raise ValueError(msg)
+      results.append(ct)
+  yield results
 
+# TODO(mattjj): remove both these exceptions to cotangent compatibility check
+def _temporary_dtype_exception(a, a_) -> bool:
+  if isinstance(a, core.ShapedArray) and isinstance(a_, core.ShapedArray):
+    return (a.shape == a_.shape and
+            (dtypes.issubdtype(a_.dtype, dtypes.extended) or
+             dtypes.issubdtype(a.dtype, dtypes.np.inexact)))
+  return False
+
+# TODO(mattjj): remove both these exceptions to cotangent compatibility check
+def _temporary_shape_exception(a, a_) -> bool:
+  return config.custom_vjp_disable_shape_check.value
 
 class CustomVJPCallPrimitive(core.CallPrimitive):
   initial_style: core.Primitive
@@ -763,7 +805,7 @@ class CustomVJPCallPrimitive(core.CallPrimitive):
         fun, self, top_trace and top_trace.level, False)
     fwd, env_trace_todo2 = process_env_traces_fwd(
       fwd, top_trace and top_trace.level, out_trees)
-    tracers = map(top_trace.full_raise, args)  # type: ignore
+    tracers = map(top_trace.full_raise, args)
     bwd_ = lambda *args: bwd(*args)
     outs = top_trace.process_custom_vjp_call(self, fun, fwd, bwd_, tracers,
                                              out_trees=out_trees,
@@ -975,7 +1017,7 @@ def custom_gradient(fun):
     ans_flat, out_tree = tree_flatten((ans,))
     rule, in_tree = flatten_fun_nokwargs(lu.wrap_init(rule), out_tree)
     ans_avals = [core.get_aval(x).at_least_vspace() for x in ans_flat]
-    jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(rule, ans_avals)
+    jaxpr, _, consts, () = pe.trace_to_jaxpr_dynamic(rule, ans_avals)
     return ans, Residuals(jaxpr, in_tree(), out_tree, consts)
 
   def bwd(res, cts):
@@ -1072,7 +1114,7 @@ def closure_convert(fun: Callable, *example_args) -> tuple[Callable, list[Any]]:
   """
   flat_args, in_tree = tree_flatten(example_args)
   in_avals = tuple(map(abstractify, flat_args))
-  if config.jax_check_tracer_leaks:
+  if config.check_tracer_leaks.value:
     return _closure_convert_for_avals.__wrapped__(fun, in_tree, in_avals)
   else:
     return _closure_convert_for_avals(fun, in_tree, in_avals)
@@ -1100,7 +1142,7 @@ def _maybe_perturbed(x: Any) -> bool:
 @cache()
 def _closure_convert_for_avals(fun, in_tree, in_avals):
   wrapped_fun, out_tree = flatten_fun_nokwargs(lu.wrap_init(fun), in_tree)
-  jaxpr, out_pvals, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals)
+  jaxpr, out_pvals, consts, () = pe.trace_to_jaxpr_dynamic(wrapped_fun, in_avals)
   out_tree = out_tree()
 
   (closure_consts, hoisted_consts), merge = partition_list(_maybe_perturbed, consts)
