@@ -13,23 +13,28 @@
 # limitations under the License.
 """Module for JAX debugging primitives and related functionality."""
 
-from collections.abc import Sequence
+from __future__ import annotations
+
+import importlib.util
+from collections.abc import Callable, Sequence
 import functools
+import logging
 import string
 import sys
-from typing import Any, Callable, Optional, Union
+from typing import Any, Union
 import weakref
 
 import numpy as np
 
+import jax
 import jax.numpy as jnp
 from jax import lax
 
 from jax._src import core
 from jax._src import effects
-from jax._src import linear_util as lu
 from jax._src import mesh as mesh_lib
 from jax._src import sharding_impls
+from jax._src import dispatch
 from jax._src import tree_util
 from jax._src import util
 from jax._src.interpreters import ad
@@ -40,22 +45,9 @@ from jax._src.lib import xla_client as xc
 from jax._src.lib.mlir import ir
 from jax._src.lib.mlir.dialects import hlo
 from jax._src.sharding import Sharding
-from jax._src.sharding_impls import (GSPMDSharding, NamedSharding,
-                                     parse_flatten_op_sharding)
+from jax._src.sharding_impls import NamedSharding, parse_flatten_op_sharding
 
-# pytype: disable=import-error
-try:
-  import rich
-  import rich.align
-  import rich.box
-  import rich.console
-  import rich.padding
-  import rich.style
-  import rich.table
-  RICH_ENABLED = True
-except:
-  RICH_ENABLED = False
-# pytype: enable=import-error
+logger = logging.getLogger(__name__)
 
 class DebugEffect(effects.Effect):
   __str__ = lambda self: "Debug"
@@ -85,7 +77,22 @@ map, unsafe_map = util.safe_map, map
 def debug_callback_impl(*args, callback: Callable[..., Any],
                         effect: DebugEffect):
   del effect
-  return callback(*args)
+  try:
+    cpu_device, *_ = jax.local_devices(backend="cpu")
+  except RuntimeError as e:
+    raise RuntimeError(
+        "jax.debug.callback failed to find a local CPU device to place the"
+        " inputs on. Make sure \"cpu\" is listed in --jax_platforms or the"
+        " JAX_PLATFORMS environment variable."
+    ) from e
+  args = jax.device_put(args, cpu_device)
+  with jax.default_device(cpu_device):
+    try:
+      callback(*args)
+    except BaseException:
+      logger.exception("jax.debug.callback failed")
+      raise
+  return ()
 
 @debug_callback_p.def_effectful_abstract_eval
 def debug_callback_abstract_eval(*flat_avals, callback: Callable[..., Any],
@@ -148,19 +155,19 @@ def debug_callback_lowering(ctx, *args, effect, callback, **params):
     sharding = None
 
   def _callback(*flat_args):
-    return tuple(
-        debug_callback_p.impl(
-            *flat_args, effect=effect, callback=callback, **params))
+    debug_callback_p.impl(
+        *flat_args, effect=effect, callback=callback, **params)
+    return ()
   if effects.ordered_effects.contains(effect):
-    token = ctx.tokens_in.get(effect)[0]
-    result, token, keepalive = mlir.emit_python_callback(
-        ctx, _callback, token, list(args), ctx.avals_in, ctx.avals_out, True)
-    ctx.set_tokens_out(mlir.TokenSet({effect: (token,)}))
+    token = ctx.tokens_in.get(effect)
+    result, token, _ = mlir.emit_python_callback(
+        ctx, _callback, token, list(args), ctx.avals_in, ctx.avals_out,
+        has_side_effect=True)
+    ctx.set_tokens_out(mlir.TokenSet({effect: token}))
   else:
-    result, token, keepalive = mlir.emit_python_callback(
-        ctx, _callback, None, list(args), ctx.avals_in, ctx.avals_out, True,
-        sharding=sharding)
-  ctx.module_context.add_keepalive(keepalive)
+    result, _, _ = mlir.emit_python_callback(
+        ctx, _callback, None, list(args), ctx.avals_in, ctx.avals_out,
+        has_side_effect=True, sharding=sharding)
   return result
 mlir.register_lowering(debug_callback_p, debug_callback_lowering,
                        platform="cpu")
@@ -200,7 +207,7 @@ def _debug_callback_partial_eval_custom(saveable, unks_in, inst_in, eqn):
 pe.partial_eval_jaxpr_custom_rules[debug_callback_p] = (
     _debug_callback_partial_eval_custom)
 
-def debug_callback(callback: Callable[..., Any], *args: Any,
+def debug_callback(callback: Callable[..., None], *args: Any,
                    ordered: bool = False, **kwargs: Any) -> None:
   """Calls a stageable Python callback.
 
@@ -219,7 +226,7 @@ def debug_callback(callback: Callable[..., Any], *args: Any,
   of the computation are duplicated or dropped.
 
   Args:
-    callback: A Python callable. Its return value will be ignored.
+    callback: A Python callable returning None.
     *args: The positional arguments to the callback.
     ordered: A keyword only argument used to indicate whether or not the
       staged out computation will enforce ordering of this callback w.r.t.
@@ -236,12 +243,15 @@ def debug_callback(callback: Callable[..., Any], *args: Any,
 
   .. _External Callbacks: https://jax.readthedocs.io/en/latest/notebooks/external_callbacks.html
   """
+  if not callable(callback):
+    raise TypeError("first argument to jax.debug.callback must be callable, "
+                    f"but got an object of type {type(callback)}")
   flat_args, in_tree = tree_util.tree_flatten((args, kwargs))
   effect = ordered_debug_effect if ordered else debug_effect
   def _flat_callback(*flat_args):
     args, kwargs = tree_util.tree_unflatten(in_tree, flat_args)
     callback(*args, **kwargs)
-    return []
+    return ()
   debug_callback_p.bind(*flat_args, callback=_flat_callback, effect=effect)
 
 class _DebugPrintFormatChecker(string.Formatter):
@@ -266,17 +276,34 @@ def _format_print_callback(fmt: str, *args, **kwargs):
 def debug_print(fmt: str, *args, ordered: bool = False, **kwargs) -> None:
   """Prints values and works in staged out JAX functions.
 
-  Note: This function does *not* work with f-strings because the formatting is
-  done lazily.
+  This function does *not* work with f-strings because formatting is delayed.
+  So instead of ``jax.debug.print(f"hello {bar}")``, write
+  ``jax.debug.print("hello {bar}", bar=bar)``.
+
+  This function is a thin convenience wrapper around :func:`jax.debug.callback`.
+  The implementation is essentially::
+
+    def debug_print(fmt: str, *args, **kwargs):
+      jax.debug.callback(
+          lambda *args, **kwargs: print(fmt.format(*args, **kwargs)),
+          *args, **kwargs)
+
+  It may be useful to call :func:`jax.debug.callback` directly instead of this
+  convenience wrapper. For example, to get debug printing in logs, you might
+  use :func:`jax.debug.callback` together with ``logging.log``.
 
   Args:
     fmt: A format string, e.g. ``"hello {x}"``, that will be used to format
-      input arguments.
-    *args: A list of positional arguments to be formatted.
+      input arguments, like ``str.format``. See the Python docs on
+      `string formatting <https://docs.python.org/3/library/stdtypes.html#str.format>`_
+      and `format string syntax <https://docs.python.org/3/library/string.html#formatstrings>`_.
+    *args: A list of positional arguments to be formatted, as if passed to
+      ``fmt.format``.
     ordered: A keyword only argument used to indicate whether or not the
       staged out computation will enforce ordering of this ``jax.debug.print``
       w.r.t. other ordered ``jax.debug.print`` calls.
-    **kwargs: Additional keyword arguments to be formatted.
+    **kwargs: Additional keyword arguments to be formatted, as if passed to
+      ``fmt.format``.
   """
   # Check that we provide the correct arguments to be formatted
   formatter.format(fmt, *args, **kwargs)
@@ -289,6 +316,7 @@ def debug_print(fmt: str, *args, ordered: bool = False, **kwargs) -> None:
 
 inspect_sharding_p = core.Primitive("inspect_sharding")
 inspect_sharding_p.multiple_results = True
+dispatch.prim_requires_devices_during_lowering.add(inspect_sharding_p)
 
 def _inspect_sharding_impl(value, *, callback):
   callback(value.sharding)
@@ -328,16 +356,21 @@ def _inspect_sharding_lowering_rule(ctx: mlir.LoweringRuleContext, value, *,
 
   if isinstance(axis_context, sharding_impls.ShardingContext):
     devices = axis_context.device_assignment
+    if devices is None:
+      raise AssertionError(
+          'Please file a bug at https://github.com/google/jax/issues')
   elif isinstance(axis_context, sharding_impls.SPMDAxisContext):
-    devices = list(axis_context.mesh.devices.flat)
+    devices = axis_context.mesh._flat_devices_tuple
   else:
     raise NotImplementedError(type(axis_context))
+  assert devices is not None
 
   # If we have a nontrivial parallel computation, we need to wait until the SPMD
   # partitioner calls back with the `HloSharding.
   def _hlo_sharding_callback(hlo_sharding: xc.HloSharding):
     if mesh.empty:
-      return callback(GSPMDSharding(devices, hlo_sharding))
+      return callback(
+          sharding_impls._op_sharding_to_pos_sharding(hlo_sharding, devices))
     pspec = parse_flatten_op_sharding(hlo_sharding, mesh)[0].get_partition_spec()
     return callback(NamedSharding(mesh, pspec))
 
@@ -368,34 +401,6 @@ def inspect_sharding_prop_user_sharding(sharding, backend_string):
   del sharding, backend_string
   return []
 
-def inspect_sharding_partition(shapes, arg_shardings, result_shape,
-                               result_sharding, backend_string):
-  del result_shape, result_sharding
-  sharding_callback_info = sharding_callbacks[backend_string]
-  sharding_callback = sharding_callback_info.callback
-  module_context = sharding_callback_info.module_context
-
-  # Execute callback
-  hlo_sharding, = arg_shardings
-  sharding_callback(hlo_sharding)
-
-  tiled_args = [p.tile(s) for s, p in zip(shapes, arg_shardings)]
-  in_avals = [core.ShapedArray(arg.dimensions(), arg.numpy_dtype())
-              for arg in tiled_args]
-  fun = lu.wrap_init(lambda *args: [])
-  jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(fun, in_avals)
-  closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
-  trivial_comp = mlir.build_xla_computation_helper(closed_jaxpr,
-      name="tmp_xla_computation", platform=module_context.platform,
-      backend_or_name=module_context.backend_or_name,
-      axis_context=module_context.axis_context)
-  # The trivial computation built here has a dummy tuple as the result,
-  # so use sharding compatible with it for the result sharding.
-  empty_tuple_sharding = xc.OpSharding()
-  empty_tuple_sharding.type = xc.OpSharding.Type.TUPLE
-  result_sharding = xc.HloSharding.from_proto(empty_tuple_sharding)
-  return trivial_comp, arg_shardings, result_sharding
-
 def inspect_sharding_infer_sharding_from_operands(arg_shapes, arg_shardings,
                                                   shape, backend_string):
   del arg_shapes, shape, backend_string
@@ -409,7 +414,7 @@ def _slice_to_chunk_idx(size: int, slc: slice) -> int:
   assert size % slice_size == 0
   return slc.start // slice_size
 
-def _raise_to_slice(slc: Union[slice, int]):
+def _raise_to_slice(slc: slice | int):
   if isinstance(slc, int):
     return slice(slc, slc + 1)
   return slc
@@ -440,10 +445,21 @@ def make_color_iter(color_map, num_rows, num_cols):
 def visualize_sharding(shape: Sequence[int], sharding: Sharding, *,
                        use_color: bool = True, scale: float = 1.,
                        min_width: int = 9, max_width: int = 80,
-                       color_map: Optional[ColorMap] = None):
+                       color_map: ColorMap | None = None):
   """Visualizes a ``Sharding`` using ``rich``."""
-  if not RICH_ENABLED:
+  if not importlib.util.find_spec("rich"):
     raise ValueError("`visualize_sharding` requires `rich` to be installed.")
+
+  # These imports are local so that they don't affect JAX import times.
+  # pytype: disable=import-error
+  import rich.align
+  import rich.console
+  import rich.box
+  import rich.padding
+  import rich.style
+  import rich.table
+  # pytype: enable=import-error
+
   if len(shape) > 2 or len(shape) < 1:
     raise ValueError(
         "`visualize_sharding` only works for shapes with 1 and 2 dimensions.")
@@ -466,7 +482,7 @@ def visualize_sharding(shape: Sequence[int], sharding: Sharding, *,
 
   device_indices_map = sharding.devices_indices_map(tuple(shape))
   slices: dict[tuple[int, ...], set[int]] = {}
-  heights: dict[tuple[int, ...], Optional[float]] = {}
+  heights: dict[tuple[int, ...], float | None] = {}
   widths: dict[tuple[int, ...], float] = {}
 
   for i, (dev, slcs) in enumerate(device_indices_map.items()):
@@ -495,11 +511,11 @@ def visualize_sharding(shape: Sequence[int], sharding: Sharding, *,
       heights[chunk_idxs] = None
       widths[chunk_idxs]  = horiz_size / shape[0]
     slices.setdefault(chunk_idxs, set()).add(dev.id)
-  num_rows = max([a[0] for a in slices.keys()]) + 1
+  num_rows = max(a[0] for a in slices.keys()) + 1
   if len(list(slices.keys())[0]) == 1:
     num_cols = 1
   else:
-    num_cols = max([a[1] for a in slices.keys()]) + 1
+    num_cols = max(a[1] for a in slices.keys()) + 1
 
   color_iter = make_color_iter(color_map, num_rows, num_cols)
   table = rich.table.Table(show_header=False, show_lines=not use_color,
