@@ -16,13 +16,13 @@
 from __future__ import annotations
 
 import atexit
-from collections.abc import Iterator, Sequence
+from collections.abc import Callable, Iterator, Sequence
 import contextlib
 import dataclasses
 from functools import partial
 import itertools
 import time
-from typing import Any, Callable, NamedTuple
+from typing import Any, NamedTuple
 import logging
 import threading
 
@@ -30,39 +30,30 @@ import numpy as np
 
 import jax
 from jax._src import basearray
+from jax._src import config
 from jax._src import core
+from jax._src import api
 from jax._src import dtypes
-from jax._src import linear_util as lu
-from jax._src import api_util
-from jax._src import tree_util
 from jax._src import source_info_util
 from jax._src import traceback_util
 from jax._src import util
-from jax._src import xla_bridge as xb
-from jax._src.config import config
 from jax._src.interpreters import ad
 from jax._src.interpreters import batching
+from jax._src.abstract_arrays import array_types
 from jax._src.interpreters import mlir
 from jax._src.interpreters import xla
 from jax._src.interpreters import pxla
+from jax._src import lib
 from jax._src.lib import xla_client as xc
-from jax._src.lib import xla_extension_version
 from jax._src.monitoring import record_event_duration_secs
 from jax._src.partition_spec import PartitionSpec
 from jax._src.sharding import Sharding
 from jax._src.sharding_impls import (
-    PmapSharding, SingleDeviceSharding, NamedSharding, XLACompatibleSharding,
-    UNSPECIFIED, GSPMDSharding, TransferToMemoryKind)
+    SingleDeviceSharding, NamedSharding,
+    GSPMDSharding, TransferToMemoryKind, is_single_device_sharding)
+from jax._src.layout import Layout, DeviceLocalLayout
 
 
-# TODO(b/300274285): Remove when performance of non-tuple and tuple args
-# is matched within each supported backend.
-_JAX_FORCE_TUPLE_ARGS = config.DEFINE_bool(
-    "jax_force_tuple_args",
-    False,
-    help="Force tuplization of arguments to lowered functions.",
-)
-
 JAXPR_TRACE_EVENT = "/jax/core/compile/jaxpr_trace_duration"
 JAXPR_TO_MLIR_MODULE_EVENT = "/jax/core/compile/jaxpr_to_mlir_module_duration"
 BACKEND_COMPILE_EVENT = "/jax/core/compile/backend_compile_duration"
@@ -86,110 +77,25 @@ _on_exit = False
 
 ### op-by-op execution
 
-class _ArgSpec(NamedTuple):
-  aval: core.AbstractValue
-  sharding: XLACompatibleSharding | None
-
-
-def _arg_spec(x: Any) -> _ArgSpec:
-  from jax._src import pjit
-
-  aval = xla.abstractify(x)
-  try:
-    if isinstance(x.sharding, PmapSharding):
-      return _ArgSpec(aval, None)
-    return _ArgSpec(aval, (pjit.to_gspmd_sharding(x.sharding, x.ndim)  # type: ignore
-                          if x._committed else None))
-  except:
-    return _ArgSpec(aval, None)
-
-
-@dataclasses.dataclass(frozen=True)
-class OrigShardings:
-  shardings: Sequence[GSPMDSharding | None]
-
-  def __hash__(self):
-    return hash(tuple(s for s in self.shardings))
-
-  def __eq__(self, other):
-    if not isinstance(other, OrigShardings):
-      return False
-    return all(getattr(s, "_original_sharding", s) == getattr(o, "_original_sharding", o)
-               for s, o in zip(self.shardings, other.shardings))
-
-
 def apply_primitive(prim, *args, **params):
   """Impl rule that compiles and runs a single primitive 'prim' using XLA."""
-  from jax._src import pjit
-
+  fun = xla_primitive_callable(prim, **params)
+  # TODO(yashkatariya): Investigate adding is_primitive to jit and never
+  # triggering the disable jit path instead of messing around with it here.
+  prev = lib.jax_jit.swap_thread_local_state_disable_jit(False)
   try:
-    in_avals, in_shardings = util.unzip2([_arg_spec(a) for a in args])
-    in_tree = tree_util.tree_structure(args)
-    compiled_fun = xla_primitive_callable(
-        prim, in_avals, in_tree, OrigShardings(in_shardings), **params)
-  except pxla.DeviceAssignmentMismatchError as e:
-    fails, = e.args
-    # TODO(yashkatariya): Thread through a signature_fun via every primitive
-    # using apply_primitive so that the error message has the right argument
-    # name instead of `args[0]`, etc.
-    arg_names = api_util._arg_names(prim.impl, args, {}, (), ())
-    msg = pjit._device_assignment_mismatch_error(
-        prim.name, fails, args, 'jit', arg_names)
-    raise ValueError(msg) from None
-
-  return compiled_fun(*args)
-
+    outs = fun(*args)
+  finally:
+    lib.jax_jit.swap_thread_local_state_disable_jit(prev)
+  return outs
 
 @util.cache()
-def xla_primitive_callable(
-    prim: core.Primitive, in_avals: tuple[core.AbstractValue, ...], in_tree,
-    orig_in_shardings: OrigShardings, **params,
-) -> Callable:
+def xla_primitive_callable(prim: core.Primitive, **params):
   def prim_fun(*args):
-    out = prim.bind(*args, **params)
-    if prim.multiple_results:
-      return out
-    else:
-      return out,
-  donated_invars = (False,) * len(in_avals)
-  wrapped_fun = lu.wrap_init(prim_fun)
-  flat_fun, out_tree = api_util.flatten_fun_nokwargs(wrapped_fun, in_tree)
-  computation = sharded_lowering(
-      flat_fun, prim.name, donated_invars, keep_unused=False,
-      inline=True, in_avals=in_avals, in_shardings=orig_in_shardings.shardings,
-      lowering_parameters=mlir.LoweringParameters())
-  compiled = computation.compile()
-  if xla_extension_version >= 192:
-    if config.jax_disable_jit:
-      call = compiled.unsafe_call
-    else:
-      call = compiled.create_cpp_call_for_apply_primitive(out_tree())
-      if call is None:
-        call = compiled.unsafe_call
-  else:
-    call = compiled.unsafe_call
-  if not prim.multiple_results:
-    return lambda *args, **kw: call(*args, **kw)[0]
-  else:
-    return call
-
-
-def sharded_lowering(
-    fun: lu.WrappedFun, name: str, donated_invars: Sequence[bool],
-    keep_unused: bool, inline: bool, in_avals: tuple[core.AbstractValue, ...],
-    in_shardings: Sequence[Sharding | None],
-    lowering_parameters: mlir.LoweringParameters
-) -> pxla.MeshComputation:
-  in_shardings_unspec = [UNSPECIFIED if i is None else i for i in in_shardings]
-
-  # Pass in a singleton `UNSPECIFIED` for out_shardings because we don't know
-  # the number of output avals at this stage. lower_sharding_computation will
-  # apply it to all out_avals.
-  return pxla.lower_sharding_computation(
-      fun, 'jit', name, in_shardings_unspec, UNSPECIFIED, donated_invars,
-      in_avals, keep_unused=keep_unused, inline=inline,
-      devices_from_context=None,
-      lowering_parameters=lowering_parameters)
+    return prim.bind(*args, **params)
+  prim_fun.__name__ = prim.name
+  prim_fun.__qualname__ = prim.name
+  return api.jit(prim_fun)
 
 
 def simple_impl(prim):
@@ -198,11 +104,11 @@ def simple_impl(prim):
 RuntimeToken = Any
 
 class RuntimeTokenSet(threading.local):
-  """See docstring for effect.py module for the calling convention for tokens."""
+  """See docstring for effects.py module for the calling convention for tokens."""
 
   # For each ordered effect, the token returned by the last dispatched
   # computation, sharded over the devices in that computation.
-  current_tokens: dict[core.Effect, jax.Array]
+  current_tokens: dict[core.Effect, core.Token]
 
   # For each device, the runtime token returned by the last dispatched
   # computation on that device.
@@ -212,19 +118,26 @@ class RuntimeTokenSet(threading.local):
     self.current_tokens = {}
     self.output_runtime_tokens = {}
 
-  def get_token_input(self, eff: core.Effect,
-                      devices: list[Device]) -> jax.Array:
+  def get_token_input(
+      self, eff: core.Effect, devices: list[Device]
+  ) -> core.Token:
     tok = self.current_tokens.get(eff, np.zeros(0, np.bool_))
-    s = NamedSharding(pxla.Mesh(devices, axis_names=["dev"]),
-                      PartitionSpec([]))
+
+    if isinstance(tok, core.Token):
+      # The order of devices may change, so we need to reshard if necessary.
+      # TODO(yueshengys): This might still be buggy in a multi-process SPMD
+      # scenario. Revise the logic later. A distributed shutdown barrier inside
+      # the XLA program may be needed.
+      return jax.device_put(tok, jax.sharding.PositionalSharding(devices))
+
+    # We only use replicated sharding for the first time when the token for the
+    # order effect hasn't been created.
     s = jax.sharding.GSPMDSharding.get_replicated(devices)
-    indices = tuple(
-        s.addressable_devices_indices_map(tok.shape).values())
-    sharded_tok = pxla.shard_args(devices, [indices], [s], [tok])[0]
+    sharded_tok = core.Token(pxla.shard_args([s], [tok])[0])
     self.current_tokens[eff] = sharded_tok
     return sharded_tok
 
-  def set_token_result(self, eff: core.Effect, token: jax.Array):
+  def set_token_result(self, eff: core.Effect, token: core.Token):
     self.current_tokens[eff] = token
 
   def set_output_runtime_token(self, device: Device, token: RuntimeToken):
@@ -251,18 +164,12 @@ def wait_for_tokens():
   runtime_tokens.block_until_ready()
 
 
-def is_single_device_sharding(sharding: Sharding) -> bool:
-  # Special case PmapSharding here because PmapSharding maps away an axis
-  # and needs to be handled separately.test_pjit_single_device_sharding_add
-  return len(sharding.device_set) == 1 and not isinstance(sharding, PmapSharding)
-
-
 @contextlib.contextmanager
 def log_elapsed_time(fmt: str, fun_name: str, event: str | None = None):
   if _on_exit:
     yield
   else:
-    log_priority = logging.WARNING if config.jax_log_compiles else logging.DEBUG
+    log_priority = logging.WARNING if config.log_compiles.value else logging.DEBUG
     start_time = time.time()
     yield
     elapsed_time = time.time() - start_time
@@ -278,10 +185,6 @@ def should_tuple_args(num_args: int, platform: str) -> bool:
   # do not have small bounds.
   # TPU only needs a tuple for very long lists
   if platform == "tpu":
-    if _JAX_FORCE_TUPLE_ARGS.value:
-      # TODO(b/300274285): Remove when performance of non-tuple and tuple args
-      # is matched within each supported backend.
-      return True
     return num_args > 2000
   else:
     return False
@@ -297,43 +200,54 @@ def jaxpr_has_primitive(jaxpr: core.Jaxpr, prim_name: str) -> bool:
   return False
 
 
+# Use this registry with caution. It will void the guarantee that lowering to
+# stablehlo is oblivious of physical devices.
+prim_requires_devices_during_lowering: set[core.Primitive] = set()
+
+def jaxpr_has_prim_requiring_devices(jaxpr: core.Jaxpr):
+  for eqn in jaxpr.eqns:
+    if eqn.primitive in prim_requires_devices_during_lowering:
+      return True
+  for subjaxpr in core.subjaxprs(jaxpr):
+    if jaxpr_has_prim_requiring_devices(subjaxpr):
+      return True
+  return False
+
+
 class SourceInfo(NamedTuple):
-  source_info: str
+  source_info: source_info_util.SourceInfo
   eqn_name: str
 
 
-def jaxpr_shardings(
+def get_intermediate_shardings(
     jaxpr: core.Jaxpr,
-) -> Iterator[tuple[XLACompatibleSharding, SourceInfo]]:
+) -> Iterator[tuple[Sharding, SourceInfo]]:
   from jax._src import pjit
   from jax.experimental import shard_map
 
   for eqn in jaxpr.eqns:
     if eqn.primitive is pjit.sharding_constraint_p:
-      source_info = SourceInfo(source_info_util.summarize(eqn.source_info),
-                                eqn.primitive.name)
+      source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       yield (eqn.params['sharding'], source_info)
     elif eqn.primitive is pjit.pjit_p:
-      source_info = SourceInfo(source_info_util.summarize(eqn.source_info),
-                                eqn.primitive.name)
+      source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       yield from ((i, source_info) for i in eqn.params['in_shardings'])
       yield from ((o, source_info) for o in eqn.params['out_shardings'])
     elif eqn.primitive is shard_map.shard_map_p:
-      source_info = SourceInfo(source_info_util.summarize(eqn.source_info),
-                                eqn.primitive.name)
+      if not eqn.params['mesh']._is_jax_device_mesh:
+        continue
+      source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
       def _names_to_pspec(names):
         ndmin = max(names) + 1 if names else 0
         return PartitionSpec(*(names.get(i) for i in range(ndmin)))
       yield from ((NamedSharding(eqn.params['mesh'], _names_to_pspec(names)), source_info)
                   for names in [*eqn.params['in_names'], *eqn.params['out_names']])
     elif eqn.primitive is device_put_p:
-      s = eqn.params['device']
-      if isinstance(s, XLACompatibleSharding) and s.memory_kind is not None:
-        source_info = SourceInfo(source_info_util.summarize(eqn.source_info),
-                                 eqn.primitive.name)
-        yield (s, source_info)
+      source_info = SourceInfo(eqn.source_info, eqn.primitive.name)
+      yield from ((s, source_info) for s in eqn.params['devices']
+                  if isinstance(s, Sharding) and s.memory_kind is not None)
   for subjaxpr in core.subjaxprs(jaxpr):
-    yield from jaxpr_shardings(subjaxpr)
+    yield from get_intermediate_shardings(subjaxpr)
 
 
 def jaxpr_has_bints(jaxpr: core.Jaxpr) -> bool:
@@ -395,7 +309,7 @@ def _initial_style_primitive_replicas(params: dict[str, Any]) -> int:
              default=1)
 
 def needs_check_special() -> bool:
-  return config.jax_debug_infs or config.jax_debug_nans
+  return config.debug_infs.value or config.debug_nans.value
 
 def check_special(name: str, bufs: Sequence[basearray.Array]) -> None:
   if needs_check_special():
@@ -404,27 +318,19 @@ def check_special(name: str, bufs: Sequence[basearray.Array]) -> None:
 
 def _check_special(name: str, dtype: np.dtype, buf: basearray.Array) -> None:
   if dtypes.issubdtype(dtype, np.inexact):
-    if config.jax_debug_nans and np.any(np.isnan(np.asarray(buf))):
+    if config.debug_nans.value and np.any(np.isnan(np.asarray(buf))):
       raise FloatingPointError(f"invalid value (nan) encountered in {name}")
-    if config.jax_debug_infs and np.any(np.isinf(np.asarray(buf))):
+    if config.debug_infs.value and np.any(np.isinf(np.asarray(buf))):
       raise FloatingPointError(f"invalid value (inf) encountered in {name}")
 
 
-def _put_x(x, s: Sharding, aval: core.AbstractValue, committed: bool):
-  result_handler = pxla.global_aval_to_result_handler(aval, s, committed, False)
-  map_ = s.devices_indices_map(aval.shape)  # type: ignore
-  return result_handler(pxla.shard_arg(x, list(map_), list(map_.values()), s))
-
-def _override_get_device_assignment(sharding, *args, **kwargs):
-  da = sharding._device_assignment
-  return xb.get_device_backend(da[0]), da
-
 def _identity_fn(x):
   return x
 
-def _mcjax_reshard(x, target_sharding):
+def _different_device_order_reshard(x, target_sharding):
   from jax._src import api, array
 
+  x._check_if_deleted()
   inp_sharding = x.sharding
 
   if inp_sharding._device_assignment == target_sharding._device_assignment:
@@ -451,63 +357,85 @@ def _mcjax_reshard(x, target_sharding):
     new_op_sharding.iota_reshape_dims = []
     new_op_sharding.iota_transpose_perm = []
     new_op_sharding.tile_assignment_devices = np.take(
-        old_hlo_sharding.tile_assignment_devices(), permute_order
+        permute_order, old_hlo_sharding.tile_assignment_devices()
     )
     new_hlo_sharding = xc.HloSharding.from_proto(new_op_sharding)
+    # TODO(yashkatariya): Enable this when HloSharding conversion is fixed in
+    # XLA.
+    # assert (new_op_sharding.tile_assignment_dimensions
+    #         == new_hlo_sharding.tile_assignment_dimensions())
+    # assert (new_op_sharding.tile_assignment_devices
+    #         == new_hlo_sharding.tile_assignment_devices())
+    assert (list(np.take(inp_sharding._device_assignment,
+                         old_hlo_sharding.tile_assignment_devices()))
+            == list(np.take(target_sharding._device_assignment,
+                            new_op_sharding.tile_assignment_devices)))
 
   new_x = array.make_array_from_single_device_arrays(
       x.shape,
-      GSPMDSharding(target_sharding._device_assignment, new_hlo_sharding),
+      GSPMDSharding(target_sharding._device_assignment, new_hlo_sharding,
+                    memory_kind=target_sharding.memory_kind),
       x._arrays,
   )
-
-  _orig_get_and_check_device_assignment = pxla._get_and_check_device_assignment.fn
-  pxla._get_and_check_device_assignment.fn = partial(
-      _override_get_device_assignment, target_sharding)
-  try:
-    return api.jit(_identity_fn, out_shardings=target_sharding)(new_x)
-  finally:
-    pxla._get_and_check_device_assignment.fn = _orig_get_and_check_device_assignment
+  return api.jit(_identity_fn, out_shardings=target_sharding)(new_x)
 
 
-def _device_put_impl(
-    x,
-    device: Device | Sharding | None = None,
-    src: Device | Sharding | None = None):
+@dataclasses.dataclass(frozen=True)
+class _DeferredShardArg:
+  """Deferred call to `pxla.shard_args`.
+
+  Per-array impls return this object instead of a result array to indicate a
+  deferred `shard_args` call. `_batched_device_put_impl` then batches all
+  `_DeferredShardArg` objects into a single `shard_args` call.
+  """
+
+  x: Any
+  s: Sharding
+  aval: core.AbstractValue
+  committed: bool
+
+  @property
+  def result_handler(self):
+    return pxla.global_aval_to_result_handler(self.aval, self.s, self.committed)
+
+
+def _device_put_sharding_impl(x, aval, device):
   from jax._src import array
-
-  if (isinstance(device, TransferToMemoryKind) or
-      isinstance(src, TransferToMemoryKind)):
-    raise ValueError(
-        "TransferToMemoryKind argument to jax.device_put can only be used"
-        " inside jax.jit. If you are using device_put outside jax.jit, then"
-        " please provide a concrete Sharding with memory_kind.")
-
-  try:
-    aval = xla.abstractify(x)
-  except TypeError as err:
-    raise TypeError(
-        f"Argument '{x}' of type {type(x)} is not a valid JAX type") from err
+  from jax.experimental import multihost_utils
 
   if isinstance(device, Sharding):
     s = device
-    if getattr(x, 'sharding', None) == s:
+    if getattr(x, 'sharding', None) == s and getattr(x, '_committed', False):
       return x
-    if (not s.is_fully_addressable and  # type: ignore
+
+    if (not s.is_fully_addressable and
         isinstance(x, array.ArrayImpl) and not x.is_fully_addressable):
-      # This has to be XLACompatible because _mcjax_reshard will run a
-      # XLA computation.
-      assert isinstance(s, XLACompatibleSharding)
-      return _mcjax_reshard(x, s)
-    if not s.is_fully_addressable:  # type: ignore
+      assert isinstance(s, Sharding)
+      return _different_device_order_reshard(x, s)
+
+    if (s.is_fully_addressable and isinstance(x, array.ArrayImpl) and
+        x.is_fully_addressable and len(s.device_set) > 1 and
+        s._internal_device_list != x.sharding._internal_device_list and  # pytype: disable=attribute-error
+        s.device_set == x.sharding.device_set):
+      assert isinstance(s, Sharding)
+      return _different_device_order_reshard(x, s)
+
+    if not s.is_fully_addressable:
+      if ((isinstance(x, array.ArrayImpl) and not x._committed) or
+          type(x) in array_types):
+        # TODO(yashkatariya): Move this check to `jit`.
+        multihost_utils.assert_equal(
+            x, fail_message=(
+                f"{type(x)} passed to device_put is not the same on each"
+                " process. Make sure you are passing the same value of"
+                f" {type(x)} on each process."))
+        return api.jit(_identity_fn, out_shardings=s)(x)
       # TODO(yashkatariya,mattjj): Link to a doc about McJAX and jax.Array.
       raise ValueError(
           "device_put's second argument must be a Device or a Sharding which"
-          f" represents addressable devices, but got {s}. You are probably"
-          " trying to use device_put in multi-controller JAX which is not"
-          " supported. Please use jax.make_array_from_single_device_arrays API"
-          " or pass device or Sharding which represents addressable devices.")
-    return _put_x(x, s, aval, True)
+          f" represents addressable devices, but got {s}. Please pass device or"
+          " Sharding which represents addressable devices.")
+    return _DeferredShardArg(x, s, aval, True)
 
   # Only `Device` exists below. `Sharding` instance is handled above.
   if isinstance(x, array.ArrayImpl):
@@ -523,37 +451,135 @@ def _device_put_impl(
 
   sh = SingleDeviceSharding(pxla._get_default_device()
                             if device is None else device)
-  return _put_x(x, sh, aval, device is not None)
+  return _DeferredShardArg(x, sh, aval, device is not None)
+
+
+def _device_put_impl(
+    x,
+    *,
+    device: Device | Sharding | Layout | None,
+    src: Device | Sharding | Layout | None,
+):
+  if (isinstance(device, TransferToMemoryKind) or
+      isinstance(src, TransferToMemoryKind)):
+    raise ValueError(
+        "TransferToMemoryKind argument to jax.device_put can only be used"
+        " inside jax.jit. If you are using device_put outside jax.jit, then"
+        " please provide a concrete Sharding with memory_kind.")
+
+  try:
+    aval = xla.abstractify(x)
+  except TypeError as err:
+    raise TypeError(
+        f"Argument '{x}' of type {type(x)} is not a valid JAX type") from err
+
+  if isinstance(device, Layout):
+    l = device
+    dll = l.device_local_layout
+    x_dll = x.layout.device_local_layout if hasattr(x, 'layout') else None
+    if dll is None and l.sharding is None:
+      return _device_put_sharding_impl(x, aval, l.sharding)
+    if (not isinstance(l.sharding, Sharding) or
+        not isinstance(dll, (DeviceLocalLayout, type(None)))):
+      raise ValueError(
+          "sharding and device_local_layout in `Layout` instance should be"
+          f" concrete. Got layout: {l} for input {aval.str_short()}")
+    if getattr(x, 'layout', None) == l and getattr(x, '_committed', False):
+      return x
+    if x_dll is None and dll is None:
+      return _device_put_sharding_impl(x, aval, l.sharding)
+    return api.jit(_identity_fn, out_shardings=l)(x)
+
+  return _device_put_sharding_impl(x, aval, device)
+
+
+def _batched_device_put_impl(
+    *xs,
+    devices: Sequence[Device | Sharding | Layout | None],
+    srcs: Sequence[Device | Sharding | Layout | None],
+):
+  ys = []
+  shard_arg_indices, shard_arg_xs, shard_arg_shardings = [], [], []
+  for i, (x, device, src) in enumerate(zip(xs, devices, srcs)):
+    y = _device_put_impl(x, device=device, src=src)
+    if isinstance(y, _DeferredShardArg):
+      shard_arg_indices.append(i)
+      shard_arg_xs.append(y.x)
+      shard_arg_shardings.append(y.s)
+    ys.append(y)
+
+  if shard_arg_xs:
+    # Batch shard_arg calls. Helps improve efficiency for backends that support
+    # efficient batch transfer.
+    shard_arg_results = pxla.shard_args(shard_arg_shardings, shard_arg_xs)
+    for i, shard_arg_result in zip(shard_arg_indices, shard_arg_results):
+      assert isinstance(ys[i], _DeferredShardArg)
+      ys[i] = ys[i].result_handler(shard_arg_result)
+
+  return ys
 
 
 device_put_p = core.Primitive('device_put')
-device_put_p.def_impl(_device_put_impl)
-device_put_p.def_abstract_eval(lambda x, device=None, src=None: x)
+device_put_p.multiple_results = True
+device_put_p.def_impl(_batched_device_put_impl)
+device_put_p.def_abstract_eval(lambda *xs, devices, srcs: xs)
 
-def device_put_transpose_rule(ct, _, device, src):
-  return [device_put_p.bind(ct, device=src, src=device)]
-ad.deflinear2(device_put_p, device_put_transpose_rule)
-batching.defvectorized(device_put_p)
+def _device_put_transpose(cts, *_, devices, srcs):
+  results = [None] * len(cts)
+  dp_args = []
+  for i, (ct, device, src) in enumerate(zip(cts, devices, srcs)):
+    if type(ct) is not ad.Zero:
+      dp_args.append((i, ct, device, src))
+  if dp_args:
+    indices, args, devices, srcs = list(zip(*dp_args))
+    ys = device_put_p.bind(*args, devices=srcs, srcs=devices)
+    for i, y in zip(indices, ys):
+      results[i] = y
+  return results
+ad.primitive_jvps[device_put_p] = partial(ad.linear_jvp, device_put_p)
+ad.primitive_transposes[device_put_p] = _device_put_transpose
 
-def _tpu_device_put_lowering(ctx, x, *, device, src):
-  if (isinstance(device, (XLACompatibleSharding, TransferToMemoryKind)) and
-      device.memory_kind is not None):
-    aval, = ctx.avals_in
-    out_aval, = ctx.avals_out
-    x = mlir.wrap_with_memory_kind(x, device.memory_kind, out_aval)
-    if isinstance(device, XLACompatibleSharding):
-      x = mlir.wrap_with_sharding_op(
-          ctx, x, out_aval, device._to_xla_hlo_sharding(aval.ndim).to_proto())
-    return [x]
-  return [x]
-mlir.register_lowering(device_put_p, _tpu_device_put_lowering, platform='tpu')
+def _device_put_batcher(batched_args, batch_dims, **params):
+  mapped_batch_dims = [bd for bd in batch_dims if bd is not batching.not_mapped]
+  assert not mapped_batch_dims or all(
+      mapped_batch_dims[0] == bd for bd in mapped_batch_dims[1:]
+  ), batch_dims
+  return device_put_p.bind(*batched_args, **params), batch_dims
+batching.primitive_batchers[device_put_p] = _device_put_batcher
+
+def _tpu_gpu_device_put_lowering(ctx, *xs, devices, srcs):
+  def lower(x, device, src, aval, out_aval):
+    if (isinstance(device, (Sharding, TransferToMemoryKind)) and
+        device.memory_kind is not None):
+      if isinstance(device, Sharding):
+        x = mlir.wrap_with_sharding_op(
+            ctx, x, out_aval, device._to_xla_hlo_sharding(aval.ndim).to_proto())
+      x = mlir.wrap_with_memory_kind(x, device.memory_kind, out_aval)
+      return x
+    return x
+  return list(map(lower, xs, devices, srcs, ctx.avals_in, ctx.avals_out))
+mlir.register_lowering(
+  device_put_p, _tpu_gpu_device_put_lowering, platform='tpu')
+mlir.register_lowering(
+  device_put_p, _tpu_gpu_device_put_lowering, platform='gpu')
 
 
-def _common_device_put_lowering(ctx, x, *, device, src):
-  if (isinstance(device, (XLACompatibleSharding, TransferToMemoryKind)) and
-      device.memory_kind is not None):
-    raise NotImplementedError(
-        "Passing memory_kind to device_put via Shardings is not supported on"
-        f" platform {ctx.module_context.platform}")
-  return [x]
+def _common_device_put_lowering(ctx, *xs, devices, srcs):
+  for device in devices:
+    if (isinstance(device, (Sharding, TransferToMemoryKind)) and
+        device.memory_kind is not None):
+      raise NotImplementedError(
+          "Passing memory_kind to device_put via Shardings is not supported on"
+          f" platforms {ctx.module_context.platforms}")
+  return xs
 mlir.register_lowering(device_put_p, _common_device_put_lowering)
+
+def _propagate_mem_kind_dp(*xm, devices=None, srcs=None):
+  memory_kinds = []
+  for device in devices:
+    if isinstance(device, (Sharding, TransferToMemoryKind)):
+      memory_kinds.append(device.memory_kind)
+    else:
+      memory_kinds.append(None)
+  return memory_kinds
+pxla.memory_kind_propagate_rule[device_put_p] = _propagate_mem_kind_dp
