@@ -12,16 +12,18 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import annotations
+
 import atexit
 from collections.abc import Sequence
 import logging
 import os
-from typing import Any, Optional, Union
+from typing import Any
 
 from jax._src import clusters
-from jax._src.config import config
+from jax._src import config
+from jax._src import xla_bridge
 from jax._src.lib import xla_extension
-from jax._src.lib import xla_extension_version
 
 logger = logging.getLogger(__name__)
 
@@ -29,25 +31,33 @@ logger = logging.getLogger(__name__)
 class State:
   process_id: int = 0
   num_processes: int = 1
-  service: Optional[Any] = None
-  client: Optional[Any] = None
-  preemption_sync_manager: Optional[Any] = None
-  coordinator_address: Optional[str] = None
+  service: Any | None = None
+  client: Any | None = None
+  preemption_sync_manager: Any | None = None
+  coordinator_address: str | None = None
 
   def initialize(self,
-                 coordinator_address: Optional[str] = None,
-                 num_processes: Optional[int] = None,
-                 process_id: Optional[int] = None,
-                 local_device_ids: Optional[Union[int, Sequence[int]]] = None,
-                 initialization_timeout: int = 300):
+                 coordinator_address: str | None = None,
+                 num_processes: int | None = None,
+                 process_id: int | None = None,
+                 local_device_ids: int | Sequence[int] | None = None,
+                 cluster_detection_method: str | None = None,
+                 initialization_timeout: int = 300,
+                 coordinator_bind_address: str | None = None):
     coordinator_address = (coordinator_address or
                            os.environ.get('JAX_COORDINATOR_ADDRESS', None))
     if isinstance(local_device_ids, int):
       local_device_ids = [local_device_ids]
 
+
     (coordinator_address, num_processes, process_id, local_device_ids) = (
         clusters.ClusterEnv.auto_detect_unset_distributed_params(
-            coordinator_address, num_processes, process_id, local_device_ids
+            coordinator_address,
+            num_processes,
+            process_id,
+            local_device_ids,
+            cluster_detection_method,
+            initialization_timeout,
         )
     )
 
@@ -60,37 +70,51 @@ class State:
 
     self.coordinator_address = coordinator_address
 
+    # The default value of [::]:port tells the coordinator to bind to all
+    # available addresses on the same port as coordinator_address.
+    default_coordinator_bind_address = '[::]:' + coordinator_address.rsplit(':', 1)[1]
+    coordinator_bind_address = (coordinator_bind_address or
+                                os.environ.get('JAX_COORDINATOR_BIND_ADDRESS',
+                                               default_coordinator_bind_address))
+    if coordinator_bind_address is None:
+      raise ValueError('coordinator_bind_address should be defined.')
+
     if local_device_ids:
-      visible_devices = ','.join(str(x) for x in local_device_ids) # type: ignore[union-attr]
+      visible_devices = ','.join(str(x) for x in local_device_ids)
       logger.info('JAX distributed initialized with visible devices: %s', visible_devices)
       config.update("jax_cuda_visible_devices", visible_devices)
       config.update("jax_rocm_visible_devices", visible_devices)
 
     self.process_id = process_id
 
+    # Emit a warning about PROXY variables if they are in the user's env:
+    proxy_vars = [ key for key in os.environ.keys() if '_proxy' in key.lower()]
+
+    if len(proxy_vars) > 0:
+      vars = " ".join(proxy_vars) + ". "
+      warning = (
+        f'JAX detected proxy variable(s) in the environment as distributed setup: {vars}'
+        'On some systems, this may cause a hang of distributed.initialize and '
+        'you may need to unset these ENV variable(s)'
+      )
+      logger.warning(warning)
+
     if process_id == 0:
       if self.service is not None:
         raise RuntimeError('distributed.initialize should only be called once.')
-      logger.info('Starting JAX distributed service on %s', coordinator_address)
-      if xla_extension_version >= 179:
-        self.service = xla_extension.get_distributed_runtime_service(
-            coordinator_address, num_processes)
-      else:
-        self.service = xla_extension.get_distributed_runtime_service(
-            coordinator_address, num_processes, True)
+      logger.info(
+          'Starting JAX distributed service on %s', coordinator_bind_address
+      )
+      self.service = xla_extension.get_distributed_runtime_service(
+          coordinator_bind_address, num_processes)
 
     self.num_processes = num_processes
 
     if self.client is not None:
       raise RuntimeError('distributed.initialize should only be called once.')
 
-    if xla_extension_version >= 179:
-      self.client = xla_extension.get_distributed_runtime_client(
-          coordinator_address, process_id, init_timeout=initialization_timeout)
-    else:
-      self.client = xla_extension.get_distributed_runtime_client(
-          coordinator_address, process_id, True,
-          init_timeout=initialization_timeout)
+    self.client = xla_extension.get_distributed_runtime_client(
+        coordinator_address, process_id, init_timeout=initialization_timeout)
     logger.info('Connecting to JAX distributed service on %s', coordinator_address)
     self.client.connect()
 
@@ -117,11 +141,13 @@ class State:
 global_state = State()
 
 
-def initialize(coordinator_address: Optional[str] = None,
-               num_processes: Optional[int] = None,
-               process_id: Optional[int] = None,
-               local_device_ids: Optional[Union[int, Sequence[int]]] = None,
-               initialization_timeout: int = 300):
+def initialize(coordinator_address: str | None = None,
+               num_processes: int | None = None,
+               process_id: int | None = None,
+               local_device_ids: int | Sequence[int] | None = None,
+               cluster_detection_method: str | None = None,
+               initialization_timeout: int = 300,
+               coordinator_bind_address: str | None = None):
   """Initializes the JAX distributed system.
 
   Calling :func:`~jax.distributed.initialize` prepares JAX for execution on
@@ -130,16 +156,27 @@ def initialize(coordinator_address: Optional[str] = None,
 
   The JAX distributed system serves a number of roles:
 
-    * it allows JAX processes to discover each other and share topology information,
-    * it performs health checking, ensuring that all processes shut down if any process dies, and
-    * it is used for distributed checkpointing.
+    * It allows JAX processes to discover each other and share topology information,
+    * It performs health checking, ensuring that all processes shut down if any process dies, and
+    * It is used for distributed checkpointing.
 
   If you are using TPU, Slurm, or Open MPI, all arguments are optional: if omitted, they
   will be chosen automatically.
 
+  The ``cluster_detection_method`` may be used to choose a specific method for detecting those
+  distributed arguments. You may pass any of the automatic ``spec_detect_methods`` to this
+  argument though it is not necessary in the TPU, Slurm, or Open MPI cases.  For other MPI
+  installations, if you have a functional ``mpi4py`` installed, you may pass
+  ``cluster_detection_method="mpi4py"`` to bootstrap the required arguments.
+
   Otherwise, you must provide the ``coordinator_address``,
   ``num_processes``, and ``process_id`` arguments to :func:`~jax.distributed.initialize`.
 
+  Please note: on some systems, particularly HPC clusters that only access external networks
+  through proxy variables such as HTTP_PROXY, HTTPS_PROXY, etc., the call to
+  :func:`~jax.distributed.initialize` may timeout.  You may need to unset these variables
+  prior to application launch.
+
   Args:
     coordinator_address: the IP address of process `0` and a port on which that
       process should launch a coordinator service. The choice of
@@ -156,14 +193,24 @@ def initialize(coordinator_address: Optional[str] = None,
     local_device_ids: Restricts the visible devices of the current process to ``local_device_ids``.
       If ``None``, defaults to all local devices being visible to the process except when processes
       are launched via Slurm and Open MPI on GPUs. In that case, it will default to a single device per process.
+    cluster_detection_method: An optional string to attempt to autodetect the configuration of the distributed
+      run.  Note that "mpi4py" method requires you to have a working ``mpi4py`` install in your environment,
+      and launch the applicatoin with an MPI-compatible job launcher such as ``mpiexec`` or ``mpirun``.
+      Legacy auto-detect options (OMPI, Slurm) remain enabled.
     initialization_timeout: Time period (in seconds) for which connection will
       be retried. If the initialization takes more than the timeout specified,
       the initialization will error. Defaults to 300 secs i.e. 5 mins.
+    coordinator_bind_address: the address and port to which the coordinator service
+      on process `0` should bind. If this is not specified, the default is to bind to
+      all available addresses on the same port as ``coordinator_address``. On systems
+      that have multiple network interfaces per node it may be insufficient to only
+      have the coordinator service listen on one address/interface.
 
   Raises:
-    RuntimeError: If :func:`~jax.distributed.initialize` is called more than once.
+    RuntimeError: If :func:`~jax.distributed.initialize` is called more than once
+      or if called after the backend is already initialized.
 
-  Example:
+  Examples:
 
   Suppose there are two GPU processes, and process 0 is the designated coordinator
   with address ``10.0.0.1:1234``. To initialize the GPU cluster, run the
@@ -177,8 +224,12 @@ def initialize(coordinator_address: Optional[str] = None,
 
   >>> jax.distributed.initialize(coordinator_address='10.0.0.1:1234', num_processes=2, process_id=1)  # doctest: +SKIP
   """
+  if xla_bridge.backends_are_initialized():
+    raise RuntimeError("jax.distributed.initialize() must be called before "
+                        "any JAX computations are executed.")
   global_state.initialize(coordinator_address, num_processes, process_id,
-                          local_device_ids, initialization_timeout)
+                          local_device_ids, cluster_detection_method,
+                          initialization_timeout, coordinator_bind_address)
   atexit.register(shutdown)
 
 
