@@ -24,14 +24,15 @@ from __future__ import annotations
 import abc
 import builtins
 import functools
-from typing import cast, overload, Any, Literal, Optional, Union
+from typing import cast, overload, Any, Literal, Union
 import warnings
 
 import ml_dtypes
 import numpy as np
 
-from jax._src.config import config
-from jax._src.typing import DType, DTypeLike
+from jax._src import config
+from jax._src.typing import Array, DType, DTypeLike
+from jax._src.util import set_module, StrictABC
 
 from jax._src import traceback_util
 traceback_util.register_exclusion(__file__)
@@ -45,7 +46,9 @@ else:
     raise ValueError("JAX requires ml_dtypes version 0.2.0 or newer; "
                      f"installed version is {ml_dtypes.__version__}.")
 
+export = set_module('jax.dtypes')
 
+@export
 class extended(np.generic):
   """Scalar class for extended dtypes.
 
@@ -59,9 +62,9 @@ class extended(np.generic):
     >>> jnp.issubdtype(key.dtype, dtypes.extended)
     True
   """
-  pass
 
 
+@export
 class prng_key(extended):
   """Scalar class for PRNG Key dtypes.
 
@@ -75,10 +78,9 @@ class prng_key(extended):
     >>> jnp.issubdtype(key.dtype, dtypes.prng_key)
     True
   """
-  pass
 
 
-class ExtendedDType(metaclass=abc.ABCMeta):
+class ExtendedDType(StrictABC):
   """Abstract Base Class for extended dtypes"""
   @property
   @abc.abstractmethod
@@ -98,6 +100,13 @@ _float8_e4m3fnuz_dtype: np.dtype = np.dtype(float8_e4m3fnuz)
 _float8_e5m2_dtype: np.dtype = np.dtype(float8_e5m2)
 _float8_e5m2fnuz_dtype: np.dtype = np.dtype(float8_e5m2fnuz)
 
+def supports_inf(dtype: DTypeLike) -> bool:
+  """Return true if the dtype supports infinity, else return False."""
+  typ = np.dtype(dtype).type
+  if typ in {float8_e4m3b11fnuz, float8_e4m3fn, float8_e4m3fnuz, float8_e5m2fnuz}:
+    return False
+  return issubdtype(dtype, np.inexact)
+
 # bfloat16 support
 bfloat16: type[np.generic] = ml_dtypes.bfloat16
 _bfloat16_dtype: np.dtype = np.dtype(bfloat16)
@@ -126,25 +135,54 @@ _float8_dtypes = [
     _float8_e5m2fnuz_dtype,
 ]
 
+# 2-bit integer support
+int2: type[np.generic] | None = None
+uint2: type[np.generic] | None = None
+
+_int2_dtype: np.dtype | None = None
+_uint2_dtype: np.dtype | None = None
+
+_intn_dtypes = []
+
+# Remove the condition once the minimum ml_dtypes version required by JAX
+# contains https://github.com/jax-ml/ml_dtypes/pull/154.
+if hasattr(ml_dtypes, 'int2'):
+  int2 = ml_dtypes.int2
+  uint2 = ml_dtypes.uint2
+  _int2_dtype = np.dtype(int2)
+  _uint2_dtype = np.dtype(uint2)
+  _intn_dtypes.extend([_int2_dtype, _uint2_dtype])
+
 # 4-bit integer support
 int4: type[np.generic] = ml_dtypes.int4
 uint4: type[np.generic] = ml_dtypes.uint4
-
-_int4_dtype: np.dtype = np.dtype(int4)
-_uint4_dtype: np.dtype = np.dtype(uint4)
-
-_int4_dtypes = [
-    _int4_dtype,
-    _uint4_dtype,
-]
+_int4_dtype = np.dtype(int4)
+_uint4_dtype = np.dtype(uint4)
+_intn_dtypes.extend([_int4_dtype, _uint4_dtype])
 
 # Default types.
-bool_: type = np.bool_
-int_: type = np.int32 if config.jax_default_dtype_bits == '32' else np.int64
-uint: type = np.uint32 if config.jax_default_dtype_bits == '32' else np.uint64
-float_: type = np.float32 if config.jax_default_dtype_bits == '32' else np.float64
-complex_: type = np.complex64 if config.jax_default_dtype_bits == '32' else np.complex128
-_default_types: dict[str, type] = {'b': bool_, 'i': int_, 'u': uint, 'f': float_, 'c': complex_}
+bool_ = np.bool_
+int_: type[Any]
+uint: type[Any]
+float_: type[Any]
+complex_: type[Any]
+if config.default_dtype_bits.value == '32':
+  int_ = np.int32
+  uint = np.uint32
+  float_ = np.float32
+  complex_ = np.complex64
+else:
+  int_ = np.int64
+  uint = np.uint64
+  float_ = np.float64
+  complex_ = np.complex128
+_default_types: dict[str, type[Any]] = {
+    'b': bool_,
+    'i': int_,
+    'u': uint,
+    'f': float_,
+    'c': complex_,
+}
 
 # Trivial vectorspace datatype needed for tangent values of int/bool primals
 float0: np.dtype = np.dtype([('float0', np.void, 0)])
@@ -189,7 +227,7 @@ def to_complex_dtype(dtype: DTypeLike) -> DType:
 
 
 @functools.cache
-def _canonicalize_dtype(x64_enabled: bool, allow_extended_dtype: bool, dtype: Any) -> Union[DType, ExtendedDType]:
+def _canonicalize_dtype(x64_enabled: bool, allow_extended_dtype: bool, dtype: Any) -> DType | ExtendedDType:
   if issubdtype(dtype, extended):
     if not allow_extended_dtype:
       raise ValueError(f"Internal: canonicalize_dtype called on extended dtype {dtype} "
@@ -206,20 +244,15 @@ def _canonicalize_dtype(x64_enabled: bool, allow_extended_dtype: bool, dtype: An
     return _dtype_to_32bit_dtype.get(dtype_, dtype_)
 
 @overload
-def canonicalize_dtype(dtype: Any, allow_extended_dtype: Literal[False] = False, allow_opaque_dtype: Any = None) -> DType: ...
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: Literal[False] = False) -> DType: ...
 
 @overload
-def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False, allow_opaque_dtype: Any = None) -> Union[DType, ExtendedDType]: ...
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False) -> DType | ExtendedDType: ...
 
-def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False, allow_opaque_dtype: Any = None) -> Union[DType, ExtendedDType]:
+@export
+def canonicalize_dtype(dtype: Any, allow_extended_dtype: bool = False) -> DType | ExtendedDType:
   """Convert from a dtype to a canonical dtype based on config.x64_enabled."""
-  if allow_opaque_dtype is not None:
-    # TODO(jakevdp): complete the deprecation cycle (Deprecated July 24 2023).
-    warnings.warn(
-      "allow_opaque_dtype argument is deprecated; use allow_extended_dtype.",
-      DeprecationWarning)
-    allow_extended_dtype = allow_opaque_dtype
-  return _canonicalize_dtype(config.x64_enabled, allow_extended_dtype, dtype)  # type: ignore[bad-return-type]
+  return _canonicalize_dtype(config.enable_x64.value, allow_extended_dtype, dtype)  # pytype: disable=bad-return-type
 
 # Default dtypes corresponding to Python scalars.
 python_scalar_dtypes : dict[type, DType] = {
@@ -229,12 +262,13 @@ python_scalar_dtypes : dict[type, DType] = {
   complex: np.dtype('complex128'),
 }
 
+@export
 def scalar_type_of(x: Any) -> type:
   """Return the scalar type associated with a JAX value."""
   typ = dtype(x)
   if typ in _custom_float_dtypes:
     return float
-  elif typ in _int4_dtypes:
+  elif typ in _intn_dtypes:
     return int
   elif np.issubdtype(typ, np.bool_):
     return bool
@@ -278,7 +312,7 @@ def _scalar_type_to_dtype(typ: type, value: Any = None) -> DType:
   return dtype
 
 
-def coerce_to_array(x: Any, dtype: Optional[DTypeLike] = None) -> np.ndarray:
+def coerce_to_array(x: Any, dtype: DTypeLike | None = None) -> np.ndarray:
   """Coerces a scalar or NumPy array to an np.array.
 
   Handles Python scalar type promotion according to JAX's rules, not NumPy's
@@ -303,7 +337,10 @@ def _issubclass(a: Any, b: Any) -> bool:
     return False
 
 
-def issubdtype(a: DTypeLike, b: DTypeLike) -> bool:
+# TODO(jakevdp): consider whether to disallow None here. We allow it
+# because np.issubdtype allows it (and treats it as equivalent to float64).
+def issubdtype(a: DTypeLike | ExtendedDType | None,
+               b: DTypeLike | ExtendedDType | None) -> bool:
   """Returns True if first argument is a typecode lower/equal in type hierarchy.
 
   This is like :func:`numpy.issubdtype`, but can handle dtype extensions such as
@@ -317,30 +354,44 @@ def issubdtype(a: DTypeLike, b: DTypeLike) -> bool:
   #   don't conform to the standard numpy type hierarchy (e.g. the bfloat16 scalar
   #   type is not a subclass of np.floating) so we must also handle these specially.
 
-  # First handle extended dtypes. This is important for performance because
-  # isinstance(x, extended) is called frequently within JAX internals.
-  if _issubclass(b, extended):
+  # We cannot use the cached version directly for all inputs, because some may be
+  # unhashable (e.g. custom objects with a dtype attribute). The following check is
+  # fast and covers the majority of calls to this function within JAX library code.
+  return _issubdtype_cached(
+    a if isinstance(a, (type, np.dtype, ExtendedDType)) else np.dtype(a),  # type: ignore[arg-type]
+    b if isinstance(b, (type, np.dtype, ExtendedDType)) else np.dtype(b),  # type: ignore[arg-type]
+  )
+
+
+@functools.lru_cache(512)  # don't use util.memoize because there is no X64 dependence.
+def _issubdtype_cached(a: type | np.dtype | ExtendedDType,
+                       b: type | np.dtype | ExtendedDType) -> bool:
+  # First handle extended dtypes, which require their own logic.
+  a_is_type = isinstance(a, type)
+  b_is_type = isinstance(b, type)
+  if b_is_type and _issubclass(b, extended):
     if isinstance(a, ExtendedDType):
       return _issubclass(a.type, b)
-    if _issubclass(a, np.generic):
+    if a_is_type and _issubclass(a, np.generic):
       return _issubclass(a, b)
     return _issubclass(np.dtype(a).type, b)
   if isinstance(b, ExtendedDType):
     return isinstance(a, ExtendedDType) and a == b
   if isinstance(a, ExtendedDType):
     a = a.type
+    a_is_type = isinstance(a, type)
 
   # For all others, normalize inputs to scalar types.
-  a_sctype = a if _issubclass(a, np.generic) else np.dtype(a).type
-  b_sctype = b if _issubclass(b, np.generic) else np.dtype(b).type
+  a_sctype = a if a_is_type and _issubclass(a, np.generic) else np.dtype(a).type
+  b_sctype = b if b_is_type and _issubclass(b, np.generic) else np.dtype(b).type
 
   # Now do special handling of custom float and int types, as they don't conform
   # to the normal scalar type hierarchy.
   if a_sctype in _custom_float_scalar_types:
     return b_sctype in {a_sctype, np.floating, np.inexact, np.number, np.generic}
-  if a_sctype == int4:
+  if (int2 is not None and a_sctype == int2) or a_sctype == int4:
     return b_sctype in {a_sctype, np.signedinteger, np.integer, np.number, np.generic}
-  if a_sctype == uint4:
+  if (uint2 is not None and a_sctype == uint2) or a_sctype == uint4:
     return b_sctype in {a_sctype, np.unsignedinteger, np.integer, np.number, np.generic}
 
   # Otherwise, fall back to numpy.issubdtype
@@ -353,31 +404,30 @@ JAXType = Union[type, DType]
 # Enumeration of all valid JAX types in order.
 _weak_types: list[JAXType] = [int, float, complex]
 _bool_types: list[JAXType] = [np.dtype(bool)]
+_signed_types: list[JAXType]
+_unsigned_types: list[JAXType]
 _int_types: list[JAXType]
-if int4 is not None:
-  _int_types = [
-      np.dtype(uint4),
-      np.dtype('uint8'),
-      np.dtype('uint16'),
-      np.dtype('uint32'),
-      np.dtype('uint64'),
-      np.dtype(int4),
-      np.dtype('int8'),
-      np.dtype('int16'),
-      np.dtype('int32'),
-      np.dtype('int64'),
-  ]
-else:
-  _int_types = [
-      np.dtype('uint8'),
-      np.dtype('uint16'),
-      np.dtype('uint32'),
-      np.dtype('uint64'),
-      np.dtype('int8'),
-      np.dtype('int16'),
-      np.dtype('int32'),
-      np.dtype('int64'),
-  ]
+_unsigned_types = [
+    np.dtype(uint4),
+    np.dtype('uint8'),
+    np.dtype('uint16'),
+    np.dtype('uint32'),
+    np.dtype('uint64'),
+]
+_signed_types = [
+    np.dtype(int4),
+    np.dtype('int8'),
+    np.dtype('int16'),
+    np.dtype('int32'),
+    np.dtype('int64'),
+]
+
+if _int2_dtype is not None:
+  _signed_types.insert(0, _int2_dtype)
+if _uint2_dtype is not None:
+  _unsigned_types.insert(0, _uint2_dtype)
+
+_int_types = _unsigned_types + _signed_types
 
 _float_types: list[JAXType] = [
     *_custom_float_dtypes,
@@ -392,6 +442,63 @@ _complex_types: list[JAXType] = [
 _jax_types = _bool_types + _int_types + _float_types + _complex_types
 _jax_dtype_set = {float0, *_bool_types, *_int_types, *_float_types, *_complex_types}
 
+
+_dtype_kinds: dict[str, set] = {
+  'bool': {*_bool_types},
+  'signed integer': {*_signed_types},
+  'unsigned integer': {*_unsigned_types},
+  'integral': {*_signed_types, *_unsigned_types},
+  'real floating': {*_float_types},
+  'complex floating': {*_complex_types},
+  'numeric': {*_signed_types, *_unsigned_types, *_float_types, *_complex_types},
+}
+
+
+def isdtype(dtype: DTypeLike, kind: str | DTypeLike | tuple[str | DTypeLike, ...]) -> bool:
+  """Returns a boolean indicating whether a provided dtype is of a specified kind.
+
+  Args:
+    dtype : the input dtype
+    kind : the data type kind.
+      If ``kind`` is dtype-like, return ``dtype = kind``.
+      If ``kind`` is a string, then return True if the dtype is in the specified category:
+
+      - ``'bool'``: ``{bool}``
+      - ``'signed integer'``: ``{int4, int8, int16, int32, int64}``
+      - ``'unsigned integer'``: ``{uint4, uint8, uint16, uint32, uint64}``
+      - ``'integral'``: shorthand for ``('signed integer', 'unsigned integer')``
+      - ``'real floating'``: ``{float8_*, float16, bfloat16, float32, float64}``
+      - ``'complex floating'``: ``{complex64, complex128}``
+      - ``'numeric'``: shorthand for ``('integral', 'real floating', 'complex floating')``
+
+      If ``kind`` is a tuple, then return True if dtype matches any entry of the tuple.
+
+  Returns:
+    True or False
+  """
+  the_dtype = np.dtype(dtype)
+  kind_tuple: tuple[str | DTypeLike, ...] = (
+    kind if isinstance(kind, tuple) else (kind,)
+  )
+  options: set[DType] = set()
+  for kind in kind_tuple:
+    if isinstance(kind, str) and kind in _dtype_kinds:
+      options.update(_dtype_kinds[kind])
+      continue
+    try:
+      _dtype = np.dtype(kind)
+    except TypeError as e:
+      if isinstance(kind, str):
+        raise ValueError(
+          f"Unrecognized {kind=} expected one of {list(_dtype_kinds.keys())}, "
+          "or a compatible input for jnp.dtype()")
+      raise TypeError(
+        f"Expected kind to be a dtype, string, or tuple; got {kind=}"
+      ) from e
+    options.add(_dtype)
+  return the_dtype in options
+
+
 def _jax_type(dtype: DType, weak_type: bool) -> JAXType:
   """Return the jax type for a dtype and weak type."""
   if weak_type:
@@ -412,10 +519,11 @@ def _type_promotion_lattice(jax_numpy_dtype_promotion: str) -> dict[JAXType, lis
   This DAG maps each type to its immediately higher type on the lattice.
   """
   b1, = _bool_types
-  if int4 is not None:
-    _uint4, u1, u2, u4, u8, _int4, i1, i2, i4, i8 = _int_types  # pytype: disable=bad-unpacking
+  if _int2_dtype is not None:
+    assert _uint2_dtype is not None
+    _uint2, uint4, u1, u2, u4, u8, _int2, int4, i1, i2, i4, i8 = _int_types
   else:
-    u1, u2, u4, u8, i1, i2, i4, i8 = _int_types  # pytype: disable=bad-unpacking
+    uint4, u1, u2, u4, u8, int4, i1, i2, i4, i8 = _int_types
   *f1_types, bf, f2, f4, f8 = _float_types
   c4, c8 = _complex_types
   i_, f_, c_ = _weak_types
@@ -423,18 +531,19 @@ def _type_promotion_lattice(jax_numpy_dtype_promotion: str) -> dict[JAXType, lis
     out: dict[JAXType, list[JAXType]]
     out = {
       b1: [i_],
-      u1: [i2, u2], u2: [i4, u4], u4: [i8, u8], u8: [f_],
-      i_: [u1, i1], i1: [i2], i2: [i4], i4: [i8], i8: [f_],
+      i_: [u1, uint4, i1, int4],
+      uint4: [], u1: [i2, u2], u2: [i4, u4], u4: [i8, u8], u8: [f_],
+      int4: [], i1: [i2], i2: [i4], i4: [i8], i8: [f_],
       f_: [*f1_types, bf, f2, c_],
       **{t: [] for t in f1_types}, bf: [f4], f2: [f4], f4: [f8, c4], f8: [c8],
       c_: [c4], c4: [c8], c8: [],
     }
-    if _int4_dtype is not None:
-      out[i_].append(_int4_dtype)
-      out[_int4_dtype] = []
-    if _uint4_dtype is not None:
-      out[i_].append(_uint4_dtype)
-      out[_uint4_dtype] = []
+    if _int2_dtype is not None:
+      out[i_].append(_int2_dtype)
+      out[_int2_dtype] = []
+    if _uint2_dtype is not None:
+      out[i_].append(_uint2_dtype)
+      out[_uint2_dtype] = []
     return out
   elif jax_numpy_dtype_promotion == 'strict':
     return {
@@ -507,7 +616,7 @@ def _least_upper_bound(jax_numpy_dtype_promotion: str, *nodes: JAXType) -> JAXTy
   if len(LUB) == 1:
     return LUB.pop()
   elif len(LUB) == 0:
-    if config.jax_numpy_dtype_promotion == 'strict':
+    if config.numpy_dtype_promotion.value == 'strict':
       msg = (
         f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
         "promotion path when jax_numpy_dtype_promotion=strict. Try explicitly casting "
@@ -518,12 +627,12 @@ def _least_upper_bound(jax_numpy_dtype_promotion: str, *nodes: JAXType) -> JAXTy
         "promotion path. To avoid unintended promotion, 8-bit floats do not support "
         "implicit promotion. If you'd like your inputs to be promoted to another type, "
         "you can do so explicitly using e.g. x.astype('float32')")
-    elif any(n in _int4_dtypes for n in nodes):
+    elif any(n in _intn_dtypes for n in nodes):
       msg = (
         f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
-        "promotion path. To avoid unintended promotion, 4-bit integers do not support "
-        "implicit promotion. If you'd like your inputs to be promoted to another type, "
-        "you can do so explicitly using e.g. x.astype('int32')")
+        "promotion path. To avoid unintended promotion, 2-bit and 4-bit integers do not "
+        "support implicit promotion. If you'd like your inputs to be promoted to another "
+        "type, you can do so explicitly using e.g. x.astype('int32')")
     else:
       msg = (
         f"Input dtypes {tuple(str(n) for n in nodes)} have no available implicit dtype "
@@ -553,13 +662,15 @@ def promote_types(a: DTypeLike, b: DTypeLike) -> DType:
   # object identity, not object equality, due to the behavior of np.dtype.__eq__
   a_tp = cast(JAXType, a if any(a is t for t in _weak_types) else np.dtype(a))
   b_tp = cast(JAXType, b if any(b is t for t in _weak_types) else np.dtype(b))
-  return np.dtype(_least_upper_bound(config.jax_numpy_dtype_promotion, a_tp, b_tp))
+  return np.dtype(_least_upper_bound(config.numpy_dtype_promotion.value, a_tp, b_tp))
 
 def is_weakly_typed(x: Any) -> bool:
+  if type(x) in _weak_types:
+    return True
   try:
     return x.aval.weak_type
   except AttributeError:
-    return type(x) in _weak_types
+    return False
 
 def is_python_scalar(x: Any) -> bool:
   try:
@@ -576,11 +687,12 @@ def dtype(x: Any, *, canonicalize: bool = False) -> DType:
   """Return the dtype object for a value or type, optionally canonicalized based on X64 mode."""
   if x is None:
     raise ValueError(f"Invalid argument to dtype: {x}.")
-  elif isinstance(x, type) and x in python_scalar_dtypes:
+  is_type = isinstance(x, type)
+  if is_type and x in python_scalar_dtypes:
     dt = python_scalar_dtypes[x]
   elif type(x) in python_scalar_dtypes:
     dt = python_scalar_dtypes[type(x)]
-  elif _issubclass(x, np.generic):
+  elif is_type and _issubclass(x, np.generic):
     return np.dtype(x)
   elif issubdtype(getattr(x, 'dtype', None), extended):
     dt = x.dtype
@@ -604,17 +716,17 @@ def _lattice_result_type(*args: Any) -> tuple[DType, bool]:
     # Trivial promotion case. This allows extended dtypes through.
     out_dtype = dtypes[0]
     out_weak_type = False
-  elif all(weak_types) and config.jax_numpy_dtype_promotion != 'strict':
+  elif all(weak_types) and config.numpy_dtype_promotion.value != 'strict':
     # If all inputs are weakly typed, we compute the bound of the strongly-typed
     # counterparts and apply the weak type at the end. This avoids returning the
     # incorrect result with non-canonical weak types (e.g. weak int16).
     # TODO(jakevdp): explore removing this special case.
-    result_type = _least_upper_bound(config.jax_numpy_dtype_promotion,
+    result_type = _least_upper_bound(config.numpy_dtype_promotion.value,
                                      *{_jax_type(dtype, False) for dtype in dtypes})
     out_dtype = dtype(result_type)
     out_weak_type = True
   else:
-    result_type = _least_upper_bound(config.jax_numpy_dtype_promotion,
+    result_type = _least_upper_bound(config.numpy_dtype_promotion.value,
                                      *{_jax_type(d, w) for d, w in zip(dtypes, weak_types)})
     out_dtype = dtype(result_type)
     out_weak_type = any(result_type is t for t in _weak_types)
@@ -627,9 +739,10 @@ def result_type(*args: Any, return_weak_type_flag: Literal[True]) -> tuple[DType
 def result_type(*args: Any, return_weak_type_flag: Literal[False] = False) -> DType: ...
 
 @overload
-def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, tuple[DType, bool]]: ...
+def result_type(*args: Any, return_weak_type_flag: bool = False) -> DType | tuple[DType, bool]: ...
 
-def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType, tuple[DType, bool]]:
+@export
+def result_type(*args: Any, return_weak_type_flag: bool = False) -> DType | tuple[DType, bool]:
   """Convenience function to apply JAX argument dtype promotion.
 
   Args:
@@ -652,28 +765,37 @@ def result_type(*args: Any, return_weak_type_flag: bool = False) -> Union[DType,
   return (dtype, weak_type) if return_weak_type_flag else dtype  # type: ignore[return-value]
 
 def check_user_dtype_supported(dtype, fun_name=None):
+  if isinstance(dtype, Array):
+    # Deprecation warning added 2024 June 13.
+    warnings.warn("Passing an array as a dtype argument is deprecated; "
+                  "instead of dtype=arr use dtype=arr.dtype.",
+                  category=DeprecationWarning, stacklevel=3)
+    return  # no further check needed, as array dtypes have already been validated.
   if issubdtype(dtype, extended):
     return
   # Avoid using `dtype in [...]` because of numpy dtype equality overloading.
   if isinstance(dtype, type) and dtype in {bool, int, float, builtins.complex}:
     return
   np_dtype = np.dtype(dtype)
-  if int4 is not None:
-    is_custom_dtype = np_dtype.type in [*_custom_float_scalar_types, int4, uint4]
-  else:
-    is_custom_dtype = np_dtype.type in _custom_float_scalar_types
+  is_custom_dtype = np_dtype.type in [
+      *_custom_float_scalar_types,
+      int2,
+      int4,
+      uint2,
+      uint4,
+  ]
   if np_dtype.kind not in "biufc" and not is_custom_dtype:
     msg = f"JAX only supports number and bool dtypes, got dtype {dtype}"
     msg += f" in {fun_name}" if fun_name else ""
     raise TypeError(msg)
-  if dtype is not None and np_dtype != canonicalize_dtype(dtype):
+  if dtype is not None and np_dtype != canonicalize_dtype(np_dtype):
     msg = ("Explicitly requested dtype {} {} is not available, "
            "and will be truncated to dtype {}. To enable more dtypes, set the "
            "jax_enable_x64 configuration option or the JAX_ENABLE_X64 shell "
            "environment variable. "
            "See https://github.com/google/jax#current-gotchas for more.")
     fun_name = f"requested in {fun_name}" if fun_name else ""
-    truncated_dtype = canonicalize_dtype(dtype).name
+    truncated_dtype = canonicalize_dtype(np_dtype).name
     warnings.warn(msg.format(dtype, fun_name, truncated_dtype), stacklevel=3)
 
 def safe_to_cast(input_dtype_or_value: Any,
