@@ -12,25 +12,27 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections.abc import Sequence
+from __future__ import annotations
+
+from collections.abc import Callable, Sequence
 import contextlib
 import functools
 import itertools as it
 from functools import partial
-from typing import Any, Callable, Optional, Union
+from typing import Any
 
 import jax
+from jax._src import config
 from jax._src import linear_util as lu
 from jax._src.interpreters import partial_eval as pe
-from jax import config
 from jax.tree_util import (tree_flatten, tree_unflatten,
                            register_pytree_node, Partial)
 from jax._src import core
 from jax._src import source_info_util
 from jax._src.ad_util import (
-    add_jaxvals, add_jaxvals_p, replace_internal_symbolic_zeros,
-    replace_rule_output_symbolic_zeros, Zero, zeros_like_aval,
-    zeros_like_jaxval, zeros_like_p)
+    add_jaxvals, replace_internal_symbolic_zeros, zeros_like_jaxval,
+    replace_rule_output_symbolic_zeros, Zero, zeros_like_aval)
+from jax._src.ad_util import zeros_like_p, add_jaxvals_p  # noqa: F401
 from jax._src.api_util import flatten_fun, flatten_fun_nokwargs
 from jax._src.core import (Trace, Tracer, get_aval, call_p, Primitive, Literal,
                            raise_to_shaped)
@@ -46,7 +48,7 @@ def identity(x): return x
 
 def _update_annotation(
     f: lu.WrappedFun,
-    orig_type: Optional[tuple[tuple[core.AbstractValue, bool], ...]],
+    orig_type: tuple[tuple[core.AbstractValue, bool], ...] | None,
     explicit_nonzeros: list[bool]
   ) -> lu.WrappedFun:
   if orig_type is None:
@@ -114,6 +116,7 @@ def jvp_subtrace_aux(main, primals, tangents):
                  else x for x in aux]
   yield (out_primals, out_tangents), aux_primals
 
+
 def linearize(traceable, *primals, **kwargs):
   has_aux = kwargs.pop('has_aux', False)
   if not has_aux:
@@ -135,7 +138,7 @@ def linearize(traceable, *primals, **kwargs):
   else:
     return out_primals_consts, out_tangents_pvals, jaxpr, consts, aux()
 
-def vjp(traceable, primals, has_aux=False, reduce_axes=()):
+def vjp(traceable, primals, has_aux=False):
   if not has_aux:
     out_primals, pvals, jaxpr, consts = linearize(traceable, *primals)
   else:
@@ -144,7 +147,7 @@ def vjp(traceable, primals, has_aux=False, reduce_axes=()):
   def unbound_vjp(pvals, jaxpr, consts, *cts):
     cts = tuple(ct for ct, pval in zip(cts, pvals) if not pval.is_known())
     dummy_args = [UndefinedPrimal(v.aval) for v in jaxpr.invars]
-    arg_cts = backward_pass(jaxpr, reduce_axes, True, consts, dummy_args, cts)
+    arg_cts = backward_pass(jaxpr, True, consts, dummy_args, cts)
     return map(instantiate_zeros, arg_cts)
 
   # Ensure that vjp_ is a PyTree so that we can pass it from the forward to the backward
@@ -179,7 +182,7 @@ def recast_to_float0(primal, tangent):
 
 # NOTE: The FIXMEs below are caused by primal/tangent mixups (type
 # errors if you will)
-def backward_pass(jaxpr: core.Jaxpr, reduce_axes, transform_stack,
+def backward_pass(jaxpr: core.Jaxpr, transform_stack,
                   consts, primals_in, cotangents_in):
   if all(type(ct) is Zero for ct in cotangents_in) and not jaxpr.effects:
     return map(lambda v: Zero(v.aval), jaxpr.invars)
@@ -193,20 +196,15 @@ def backward_pass(jaxpr: core.Jaxpr, reduce_axes, transform_stack,
       # FIXME: This triggers a lot of failures!
       # assert v.aval == ct.aval, (prim, v.aval, ct.aval)
       return
-    axes_to_reduce = tuple(axis_name for axis_name in reduce_axes
-                           if axis_name in core.get_aval(ct).named_shape
-                           and axis_name not in v.aval.named_shape)
-    if axes_to_reduce:
-      ct = jax.lax.psum(ct, axis_name=axes_to_reduce)
     ct_env[v] = add_tangents(ct_env[v], ct) if v in ct_env else ct
     # TODO(mattjj): add back these checks for dynamic shapes
-    # if config.jax_enable_checks:
+    # if config.enable_checks.value:
     #   ct_aval = core.get_aval(ct_env[v])
-    #   joined_aval = core.lattice_join(v.aval, ct_aval).strip_weak_type().strip_named_shape()
-    #   assert v.aval.strip_weak_type().strip_named_shape() == joined_aval, (prim, v.aval, ct_aval)
+    #   joined_aval = core.lattice_join(v.aval, ct_aval).strip_weak_type()
+    #   assert v.aval.strip_weak_type() == joined_aval, (prim, v.aval, ct_aval)
 
   def read_cotangent(v):
-    return ct_env.pop(v, Zero(v.aval))
+    return ct_env.pop(v, Zero(v.aval.at_least_vspace()))
 
   def read_primal(v):
     if type(v) is Literal:
@@ -240,16 +238,17 @@ def backward_pass(jaxpr: core.Jaxpr, reduce_axes, transform_stack,
       else:
         cts_in, = map(read_cotangent, eqn.outvars)
       name_stack = source_info_util.current_name_stack() + eqn.source_info.name_stack
-      with source_info_util.user_context(eqn.source_info.traceback, name_stack=name_stack):
+      with source_info_util.user_context(
+          eqn.source_info.traceback, name_stack=name_stack), eqn.ctx.manager:
         if eqn.primitive.call_primitive or eqn.primitive.map_primitive:
           cts_in_avals = [v.aval for v in eqn.outvars]
           params = dict(eqn.params)
           call_jaxpr = params.pop('call_jaxpr')
           cts_out = get_primitive_transpose(eqn.primitive)(
-              params, call_jaxpr, invals, cts_in, cts_in_avals, reduce_axes)
+              params, call_jaxpr, invals, cts_in, cts_in_avals)
         elif eqn.primitive in reducing_transposes:
           cts_out = reducing_transposes[eqn.primitive](
-              reduce_axes, cts_in, *invals, **eqn.params)
+              cts_in, *invals, **eqn.params)
         else:
           cts_out = get_primitive_transpose(eqn.primitive)(
               cts_in, *invals, **eqn.params)
@@ -260,9 +259,9 @@ def backward_pass(jaxpr: core.Jaxpr, reduce_axes, transform_stack,
   cotangents_out = map(read_cotangent, jaxpr.invars)
   return cotangents_out
 
-def closed_backward_pass(jaxpr: core.ClosedJaxpr, reduce_axes, transform_stack,
+def closed_backward_pass(jaxpr: core.ClosedJaxpr, transform_stack,
                          primals_in, cotangents_in):
-  return backward_pass(jaxpr.jaxpr, reduce_axes, transform_stack, jaxpr.consts,
+  return backward_pass(jaxpr.jaxpr, transform_stack, jaxpr.consts,
                        primals_in, cotangents_in)
 
 
@@ -458,7 +457,7 @@ class JVPTracer(Tracer):
   __slots__ = ['primal', 'tangent']
 
   def __init__(self, trace, primal, tangent):
-    if config.jax_enable_checks:
+    if config.enable_checks.value:
       _primal_tangent_shapes_match(primal, tangent)
     self._trace = trace
     self.primal = primal
@@ -581,24 +580,11 @@ def zero_jvp(primitive, primals, tangents, **params):
   r = primitive.bind(*primals, **params)
   return r, Zero.from_value(r)
 
-
-deflinear2(zeros_like_p, lambda t, _: [Zero.from_value(t)])
 deflinear2(add_jaxvals_p, lambda t, *args: (t, t))
 
-def instantiate_zeros(tangent):
-  if type(tangent) is Zero:
-    return zeros_like_aval(tangent.aval)
-  else:
-    return tangent
 
-# This function seems similar to instantiate_zeros, but it is sometimes used
-# to instantiate zero abstract units with a different aval
-def instantiate_zeros_aval(aval, tangent):
-  if type(tangent) is Zero:
-    assert tangent.aval == aval
-    return zeros_like_aval(aval)
-  else:
-    return tangent
+def instantiate_zeros(tangent):
+  return zeros_like_aval(tangent.aval) if type(tangent) is Zero else tangent
 
 @lu.transformation_with_aux
 def traceable(in_tree, *primals_and_tangents):
@@ -611,20 +597,19 @@ def traceable(in_tree, *primals_and_tangents):
   yield out_flat, out_tree
 
 
-def call_transpose(primitive, params, call_jaxpr, args, ct, _, reduce_axes):
+def call_transpose(primitive, params, call_jaxpr, args, ct, _):
   if isinstance(call_jaxpr, core.ClosedJaxpr):
     call_jaxpr, consts = call_jaxpr.jaxpr, call_jaxpr.consts
   else:
     consts = ()
   all_args, in_tree_def = tree_flatten((consts, args, ct))
-  fun = lu.hashable_partial(lu.wrap_init(backward_pass), call_jaxpr,
-                            reduce_axes, False)
+  fun = lu.hashable_partial(lu.wrap_init(backward_pass), call_jaxpr, False)
   fun, out_tree = flatten_fun_nokwargs(fun, in_tree_def)
   update_params = call_transpose_param_updaters.get(primitive)
   if update_params:
     params = update_params(params, map(is_undefined_primal, args),
                            [type(x) is not Zero for x in ct])
-  if config.jax_dynamic_shapes:
+  if config.dynamic_shapes.value:
     # TODO(mattjj,dougalm): handle consts, for now assume just args
     which_lin = [is_undefined_primal(x) for x in args]
     res_invars, _ = partition_list(which_lin, call_jaxpr.invars)
@@ -638,11 +623,11 @@ def call_transpose(primitive, params, call_jaxpr, args, ct, _, reduce_axes):
 primitive_transposes[core.call_p] = partial(call_transpose, call_p)
 
 
-def _closed_call_transpose(params, jaxpr, args, ct, cts_in_avals, reduce_axes):
+def _closed_call_transpose(params, jaxpr, args, ct, cts_in_avals):
   jaxpr_, consts = jaxpr.jaxpr, jaxpr.consts
   jaxpr_ = pe.convert_constvars_jaxpr(jaxpr_)
   return call_transpose(core.closed_call_p, params, jaxpr_, (*consts, *args),
-                        ct, cts_in_avals, reduce_axes)
+                        ct, cts_in_avals)
 primitive_transposes[core.closed_call_p] = _closed_call_transpose
 
 
@@ -651,9 +636,9 @@ def nonzero_outputs(*args, **kwargs):
   results = yield args, kwargs
   yield results, [type(r) is not Zero for r in results]
 
-def map_transpose(primitive, params, call_jaxpr, args, ct, _, reduce_axes):
+def map_transpose(primitive, params, call_jaxpr, args, ct, _):
   all_args, in_tree_def = tree_flatten(((), args, ct))  # empty consts
-  fun = lu.hashable_partial(lu.wrap_init(backward_pass), call_jaxpr, reduce_axes, False)
+  fun = lu.hashable_partial(lu.wrap_init(backward_pass), call_jaxpr, False)
   fun, nz_arg_cts = nonzero_outputs(fun)
   fun, out_tree = flatten_fun_nokwargs(fun, in_tree_def)
   # Preserve axis for primal arguments, skip tangents (represented as undefined primals).
@@ -662,8 +647,10 @@ def map_transpose(primitive, params, call_jaxpr, args, ct, _, reduce_axes):
                    if not is_undefined_primal(x)],
                  *[axis for axis, x in zip(out_axes, ct)
                    if type(x) is not Zero])
-  # The interim strategy we use below (until avals-with-names) only works
-  # when all outputs are mapped.
+  if any(out_axis is None for out_axis in out_axes):
+    raise NotImplementedError(
+        "autodiff of pmap functions with out_axes=None is not supported. "
+        "Consider using shard_map instead.")
   assert all(out_axis is not None for out_axis in out_axes), out_axes
   # NOTE: This assumes that the output cotangents being zero is a deterministic
   #       function of which input cotangents were zero.
@@ -694,7 +681,7 @@ def map_transpose(primitive, params, call_jaxpr, args, ct, _, reduce_axes):
 
 
 def jvp_jaxpr(jaxpr: core.ClosedJaxpr, nonzeros: Sequence[bool],
-              instantiate: Union[bool, Sequence[bool]]
+              instantiate: bool | Sequence[bool]
               ) -> tuple[core.ClosedJaxpr, list[bool]]:
   if type(instantiate) is bool:
     instantiate = (instantiate,) * len(jaxpr.out_avals)
@@ -708,7 +695,7 @@ def _jvp_jaxpr(jaxpr, nonzeros, instantiate):
                                         nonzeros)
   tangent_avals = [aval for aval, nz in zip(jaxpr.in_avals, nonzeros) if nz]
   avals_in = list(it.chain(jaxpr.in_avals, tangent_avals))
-  jaxpr_out, avals_out, literals_out = pe.trace_to_jaxpr_dynamic(f_jvp, avals_in)
+  jaxpr_out, avals_out, literals_out, () = pe.trace_to_jaxpr_dynamic(f_jvp, avals_in)
   return core.ClosedJaxpr(jaxpr_out, literals_out), out_nonzeros()
 
 @lu.transformation_with_aux
@@ -758,7 +745,7 @@ def _custom_lin_transpose(cts_out, *invals, num_res, bwd, out_avals,
   if symbolic_zeros:
     cts_out = map(replace_internal_symbolic_zeros, cts_out)
   else:
-    cts_out = map(instantiate_zeros_aval, out_avals, cts_out)
+    cts_out = map(instantiate_zeros, cts_out)
   cts_in = bwd(*res, *cts_out)
   cts_in = map(replace_rule_output_symbolic_zeros, cts_in)
   return [None] * num_res + list(cts_in)
