@@ -14,21 +14,21 @@
 from __future__ import annotations
 
 import collections
-from collections.abc import Iterable, Sequence
+from collections.abc import Callable, Iterable, Sequence
 import dataclasses
 from functools import partial
-from typing import Any, Callable, Union
+from typing import Any, Union
 
 import numpy as np
 
 import jax
-from jax import config
+from jax._src import config
 from jax._src import core
 from jax._src import source_info_util
 from jax._src import linear_util as lu
-from jax._src.ad_util import (add_jaxvals, add_jaxvals_p, zeros_like_jaxval,
-                              zeros_like_p, Zero, SymbolicZero,
-                              replace_rule_output_symbolic_zeros, instantiate)
+from jax._src.ad_util import (Zero, instantiate, SymbolicZero,
+                              replace_rule_output_symbolic_zeros,
+                              add_jaxvals, add_jaxvals_p)
 from jax._src.core import raise_to_shaped, Trace, Tracer, AxisName
 from jax._src.interpreters import partial_eval as pe
 from jax._src.tree_util import (tree_unflatten, tree_flatten,
@@ -61,7 +61,7 @@ class IndexedAxisSize:
   idx: core.Var
   lengths: Array | core.Var | Tracer
   def __repr__(self) -> str:
-    return f'{str(self.lengths)}.Var{id(self.idx)}'
+    return f'{self.lengths}.Var{id(self.idx)}'
   replace = dataclasses.replace
 
 # Jumble(aval=a:3 => f32[[3 1 4].a],
@@ -99,7 +99,7 @@ def _jumble_unflatten(aval, x):
 register_pytree_node(Jumble, _jumble_flatten, _jumble_unflatten)
 
 def _jumble_result(axis_size, stacked_axis, ragged_axes, x):
-  binder = core.Var(0, '', core.ShapedArray((), np.dtype('int32')))
+  binder = core.Var('', core.ShapedArray((), np.dtype('int32')))
   if stacked_axis != 0:
     raise NotImplementedError  # TODO Transpose x so the stacked axis is axis 0
   shape = list(x.shape)
@@ -162,7 +162,7 @@ def bdim_as_shape(
     bdim: int | RaggedAxis, data_shape: core.Shape) -> core.Shape:
   if isinstance(bdim, RaggedAxis):
     result = list(data_shape)
-    binder = core.Var(0, '', core.ShapedArray((), np.dtype('int32')))
+    binder = core.Var('', core.ShapedArray((), np.dtype('int32')))
     for ragged_axis, segment_lens in bdim.ragged_axes:
       result[ragged_axis] = IndexedAxisSize(binder, segment_lens)
     return tuple(result)
@@ -200,7 +200,7 @@ def _update_annotation(
   class Name:
     def __init__(self, a): self.a = a
   names = [Name(a) for a, _  in orig_type]
-  avals = [a.update(shape=tuple(names[d.val] if type(d) is pe.DBIdx else d  # type: ignore
+  avals = [a.update(shape=tuple(names[d.val] if type(d) is pe.DBIdx else d
                                 for d in a.shape))
            if type(a) is core.DShapedArray else a for a, e in orig_type if e]
 
@@ -245,7 +245,7 @@ def to_elt(trace: Trace, get_idx: GetIdx, x: Vmappable, spec: MapSpec) -> Elt:
                  for i, sz in enumerate(x.aval.elt_ty.shape)
                  if type(sz) is IndexedAxisSize)
     batch_axis = make_batch_axis(x.data.ndim, 0, [(d+1, ias.lengths)])
-    return BatchTracer(trace, x.data, batch_axis)  # type: ignore
+    return BatchTracer(trace, x.data, batch_axis)
   elif isinstance(spec, int) or spec is None:
     spec = spec and canonicalize_axis(spec, len(np.shape(x)))
     return (BatchTracer(trace, x, spec, source_info_util.current())
@@ -254,11 +254,13 @@ def to_elt(trace: Trace, get_idx: GetIdx, x: Vmappable, spec: MapSpec) -> Elt:
     assert False
 to_elt_handlers: dict[type, ToEltHandler] = {}
 
-def from_elt(trace: BatchTrace, axis_size: AxisSize, x: Elt, spec: MapSpec
-             ) -> Vmappable:
+def from_elt(trace: BatchTrace, axis_size: AxisSize, i: int,
+             x: Elt, spec: MapSpec) -> Vmappable:
   handler = from_elt_handlers.get(type(x))
   if handler:
-    return handler(partial(from_elt, trace), axis_size, x, spec)
+    def _cont(axis_size, elt, axis):
+      return from_elt(trace, axis_size, i, elt, axis)
+    return handler(_cont, axis_size, x, spec)
   x_ = trace.full_raise(x)
   val, bdim = x_.val, x_.batch_dim
   if type(bdim) is RaggedAxis:
@@ -267,7 +269,10 @@ def from_elt(trace: BatchTrace, axis_size: AxisSize, x: Elt, spec: MapSpec
       raise TypeError("ragged output without using jumble_axis out_axes spec")
     return _jumble_result(axis_size, bdim.stacked_axis, bdim.ragged_axes, val)
   else:
-    return matchaxis(trace.axis_name, axis_size, x_.batch_dim, spec, x_.val)
+    try:
+      return matchaxis(trace.axis_name, axis_size, x_.batch_dim, spec, x_.val)
+    except SpecMatchError:
+      raise SpecMatchError(i, x_.batch_dim, spec) from None
 from_elt_handlers: dict[type, FromEltHandler] = {}
 
 def make_iota(axis_size: AxisSize) -> Array:
@@ -318,11 +323,11 @@ class BatchTracer(Tracer):
 
   def __init__(self, trace, val, batch_dim: NotMapped | int | RaggedAxis,
                source_info: source_info_util.SourceInfo | None = None):
-    if config.jax_enable_checks:
+    if config.enable_checks.value:
       assert type(batch_dim) in (NotMapped, int, RaggedAxis)
       if type(batch_dim) is int:
         aval = raise_to_shaped(core.get_aval(val))
-        assert 0 <= batch_dim < len(aval.shape)  # type: ignore
+        assert 0 <= batch_dim < len(aval.shape)
     self._trace = trace
     self.val = val
     self.batch_dim = batch_dim
@@ -338,7 +343,7 @@ class BatchTracer(Tracer):
     elif type(self.batch_dim) is RaggedAxis:
       new_aval = core.mapped_aval(
         aval.shape[self.batch_dim.stacked_axis], self.batch_dim.stacked_axis, aval)
-      shape = list(new_aval.shape)  # type: ignore
+      shape = list(new_aval.shape)  # pytype: disable=attribute-error
       for ragged_axis, segment_lengths in self.batch_dim.ragged_axes:
         size_tracer = BatchTracer(self._trace, segment_lengths, 0)
         if self.batch_dim.stacked_axis < ragged_axis:
@@ -416,7 +421,7 @@ class BatchTrace(Trace):
     return frame
 
   def process_primitive(self, primitive, tracers, params):
-    if config.jax_dynamic_shapes:
+    if config.dynamic_shapes.value:
       primitive.abstract_eval(*(t.aval for t in tracers), **params)
     vals_in, dims_in = unzip2((t.val, t.batch_dim) for t in tracers)
     is_axis_primitive = primitive in axis_primitive_batchers
@@ -635,7 +640,8 @@ def _batch_inner(axis_size, out_dim_dests, main, in_dims, *in_vals):
   in_tracers = map(partial(to_elt, trace, idx), in_vals, in_dims)
   outs = yield in_tracers, {}
   out_dim_dests = out_dim_dests() if callable(out_dim_dests) else out_dim_dests
-  out_vals = map(partial(from_elt, trace, axis_size), outs, out_dim_dests)
+  out_vals = map(partial(from_elt, trace, axis_size), range(len(outs)),
+                 outs, out_dim_dests)
   yield out_vals
 
 # NOTE: This divides the in_axes by the tile_size and multiplies the out_axes by it.
@@ -789,7 +795,7 @@ def _batch_jaxpr2(
   avals_in2 = [core.unmapped_aval(axis_size, axis_name, b, aval)
                if b is not not_mapped else aval
                for aval, b in unsafe_zip(avals_in, in_axes2)]
-  jaxpr_out, _, consts = pe.trace_to_jaxpr_dynamic(f, avals_in2)
+  jaxpr_out, _, consts, () = pe.trace_to_jaxpr_dynamic(f, avals_in2)
   return core.ClosedJaxpr(jaxpr_out, consts), out_axes()
 
 def handle_ragged(in_avals: list[core.AbstractValue], dim: RaggedAxis,
@@ -834,7 +840,7 @@ def _batch_jaxpr_axes(closed_jaxpr, axis_size, in_axes, out_axes_dest,
                          main_type)
   avals_in = [core.unmapped_aval(axis_size, axis_name, b, aval) if b is not not_mapped
               else aval for aval, b in unsafe_zip(closed_jaxpr.in_avals, in_axes)]
-  jaxpr_out, _, consts = pe.trace_to_jaxpr_dynamic(f, avals_in)
+  jaxpr_out, _, consts, () = pe.trace_to_jaxpr_dynamic(f, avals_in)
   return core.ClosedJaxpr(jaxpr_out, consts), out_batched()
 
 @lu.transformation_with_aux
@@ -966,7 +972,10 @@ def _matchaxis_symbolic_zeros(axis_name, sz, name, src, dst, x, sum_match=False)
 
 ### utilities for defining primitives' batching rules
 
-BatchingRule = Callable[..., tuple[Any, Union[None, int, tuple[Union[None, int], ...]]]]
+BatchingRule = Callable[
+    ...,
+    tuple[Any, Union[int, None, tuple[Union[int, None], ...]]]
+]
 primitive_batchers : dict[core.Primitive, BatchingRule] = {}
 axis_primitive_batchers: dict[core.Primitive, Callable] = {}
 spmd_axis_primitive_batchers: dict[core.Primitive, Callable] = {}
@@ -1095,13 +1104,13 @@ def matchaxis(axis_name, sz, src, dst, x, sum_match=False):
   if dst == jumble_axis:
     x = bdim_at_front(x, src, sz)
     elt_ty = x.aval.update(shape=x.shape[1:])
-    aval = JumbleTy(core.Var(0, '', core.ShapedArray((), np.dtype('int32'))),
+    aval = JumbleTy(core.Var('', core.ShapedArray((), np.dtype('int32'))),
                     x.shape[0], elt_ty)
     return Jumble(aval, x)
   try:
     _ = core.get_aval(x)
   except TypeError as e:
-    raise TypeError(f"Output from batched function {repr(x)} with type "
+    raise TypeError(f"Output from batched function {x!r} with type "
                     f"{type(x)} is not a valid JAX type") from e
   if src == dst:
     return x
@@ -1116,7 +1125,13 @@ def matchaxis(axis_name, sz, src, dst, x, sum_match=False):
         axis_name is not core.no_axis_name):
       raise ValueError(f'vmap has mapped output ({axis_name=}) but out_axes is {dst}')
     else:
-      raise ValueError(f'vmap has mapped output but out_axes is {dst}')
+      raise SpecMatchError(None, None, None)
+
+class SpecMatchError(Exception):
+  def __init__(self, leaf_idx, src, dst):
+    self.leaf_idx = leaf_idx
+    self.src = src
+    self.dst = dst
 
 def bdim_at_front(x, bdim, size):
   if bdim is not_mapped:
@@ -1124,7 +1139,6 @@ def bdim_at_front(x, bdim, size):
   else:
     return moveaxis(x, bdim, 0)
 
-# sets up primitive batchers for ad_util and xla primitives
 
 def add_batched(batched_args, batch_dims):
   bdx, bdy = batch_dims
@@ -1141,9 +1155,3 @@ def add_batched(batched_args, batch_dims):
     x = moveaxis(x, bdx, bdy)
     return add_jaxvals(x, y), bdy
 primitive_batchers[add_jaxvals_p] = add_batched
-
-def zeros_like_batched(batched_args, batch_dims):
-  val, = batched_args
-  bdim, = batch_dims
-  return zeros_like_jaxval(val), bdim
-primitive_batchers[zeros_like_p] = zeros_like_batched
