@@ -14,14 +14,15 @@
 
 # Lowering of jaxprs into XLA (HLO) computations.
 
+from __future__ import annotations
+
 from collections import defaultdict
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import dataclasses
 import functools
 from functools import partial
 import itertools as it
-import operator
-from typing import Any, Callable, Optional, Protocol, Union
+from typing import Any, Protocol, Union
 
 import numpy as np
 
@@ -78,9 +79,8 @@ def parameter(builder, num, shape, name=None, replicated=None):
 # arbitrary tuple nesting, but JAX only uses one level of tupling (and our type
 # checkers don't support recursive types), so we only represent one level of
 # nesting in this type definition.
-SpatialSharding = Union[Shape,
-                        None,
-                        tuple[Optional[Shape], ...]]
+SpatialSharding = Union[Shape, None, tuple[Union[Shape, None], ...]]
+
 
 def sharding_to_proto(sharding: SpatialSharding):
   """Converts a SpatialSharding to an OpSharding.
@@ -92,7 +92,7 @@ def sharding_to_proto(sharding: SpatialSharding):
   proto = xc.OpSharding()
   if isinstance(sharding, tuple) and not isinstance(sharding[0], int):
     assert all(s is None or isinstance(s, tuple) for s in sharding)
-    return tuple_sharding_proto(list(map(sharding_to_proto, sharding)))  # type: ignore
+    return tuple_sharding_proto(list(map(sharding_to_proto, sharding)))
 
   if sharding is None:
     proto.type = xc.OpSharding.Type.REPLICATED
@@ -110,8 +110,6 @@ def tuple_sharding_proto(elems):
   return proto
 
 
-
-
 ### handlers
 
 # JAX abstract values -> XLA shapes
@@ -132,6 +130,10 @@ _xla_shape_handlers[core.AbstractToken] = lambda _: (xc.Shape.token_shape(),)
 
 # IR constants
 
+class InvalidInputException(Exception):
+  pass
+
+
 # TODO(mattjj): try to remove this canonicalize_dtype stuff
 def canonicalize_dtype(x):
   typ = type(x)
@@ -142,8 +144,8 @@ def canonicalize_dtype(x):
     if handler: return handler(x)
   if hasattr(x, '__jax_array__'):
     return canonicalize_dtype(x.__jax_array__())
-  raise TypeError(f"Argument '{x}' of type {type(x)} is not a valid "
-                  "JAX type.")
+  raise InvalidInputException(
+      f"Argument '{x}' of type {type(x)} is not a valid JAX type.")
 
 def _canonicalize_masked_array_dtype(x):
   raise ValueError("numpy masked arrays are not supported as direct inputs to JAX functions. "
@@ -165,6 +167,7 @@ canonicalize_dtype_handlers.update(
     (t, partial(_canonicalize_python_scalar_dtype, t)) for t in _scalar_types)
 canonicalize_dtype_handlers[core.Token] = identity
 canonicalize_dtype_handlers[core.DArray] = identity
+canonicalize_dtype_handlers[core.MutableArray] = identity
 
 def abstractify(x) -> Any:
   typ = type(x)
@@ -195,7 +198,8 @@ def _make_shaped_array_for_numpy_array(x: np.ndarray) -> ShapedArray:
 
 
 pytype_aval_mappings: dict[Any, Callable[[Any], core.AbstractValue]] = {}
-pytype_aval_mappings[core.DArray] = operator.attrgetter('_aval')
+pytype_aval_mappings[core.DArray] = lambda x: x._aval
+pytype_aval_mappings[core.MutableArray] = lambda x: x._aval
 pytype_aval_mappings[core.Token] = lambda _: core.abstract_token
 pytype_aval_mappings.update((t, _make_shaped_array_for_numpy_scalar)
                             for t in numpy_scalar_types)
@@ -204,7 +208,7 @@ pytype_aval_mappings.update(
     (t, partial(_make_abstract_python_scalar, t)) for t in _scalar_types)
 
 
-def primitive_subcomputation(platform: str, axis_env: 'AxisEnv',
+def primitive_subcomputation(platform: str, axis_env: AxisEnv,
                              prim: core.Primitive,
                              avals_in: Sequence[core.AbstractValue],
                              avals_out: Sequence[core.AbstractValue],
@@ -237,9 +241,9 @@ class TranslationContext:
   builder: xc.XlaBuilder
   # TODO(phawkins): make platform non-optional. We should always be translating
   # with a specific platform in mind.
-  platform: Optional[str]
+  platform: str | None
   axis_env: AxisEnv
-  name_stack: Union[str, source_info_util.NameStack]
+  name_stack: str | source_info_util.NameStack
 
   def replace(self, **kw): return dataclasses.replace(self, **kw)
 
@@ -272,7 +276,7 @@ def register_initial_style_primitive(prim: core.Primitive):
   initial_style_primitives.add(prim)
 
 def register_translation(prim: core.Primitive, rule: TranslationRule, *,
-                         platform: Optional[str] = None) -> None:
+                         platform: str | None = None) -> None:
   if platform is None:
     _translations[prim] = rule
   else:
