@@ -229,11 +229,19 @@ def _approx_top_k_abstract_eval(operand, *, k, reduction_dimension,
   if not dtypes.issubdtype(operand.dtype, np.floating):
     raise ValueError('operand must be a floating type')
   reduction_input_size = dims[reduction_dimension]
-  dims[reduction_dimension] = xc.ops.ApproxTopKReductionOutputSize(
-      reduction_input_size, len(dims), k, recall_target, aggregate_to_topk,
-      reduction_input_size_override)[0]
-  return (operand.update(
-      shape=dims, dtype=operand.dtype, weak_type=operand.weak_type),
+  if aggregate_to_topk:
+    dims[reduction_dimension] = k
+  elif core.is_constant_shape((reduction_input_size, k)):
+    dims[reduction_dimension] = xc.ops.ApproxTopKReductionOutputSize(
+        reduction_input_size, len(dims), k, recall_target, aggregate_to_topk,
+        reduction_input_size_override)[0]
+  else:
+    raise NotImplementedError(
+         "approx_top_k with aggregate_to_topk=False not yet implemented when "
+         f"either the `k` ({k}) or the "
+         f" reduction dimension size ({reduction_input_size}) are symbolic")
+  return (operand.update(shape=dims, dtype=operand.dtype,
+                         weak_type=operand.weak_type),
           operand.update(shape=dims, dtype=np.dtype(np.int32)))
 
 
@@ -254,30 +262,6 @@ def _comparator_builder(op_type, is_max_k):
 def _get_init_val_literal(op_type, is_max_k):
   return np.array(-np.inf if is_max_k else np.inf, dtype=op_type)
 
-def _approx_top_k_tpu_translation(ctx, avals_in, avals_out, operand, *, k,
-                                  reduction_dimension, recall_target, is_max_k,
-                                  reduction_input_size_override,
-                                  aggregate_to_topk):
-  c = ctx.builder
-  op_shape = c.get_shape(operand)
-  if not op_shape.is_array():
-    raise ValueError(f'operand must be an array, but was {op_shape}')
-  op_dims = op_shape.dimensions()
-  op_type = op_shape.element_type()
-  if reduction_dimension < 0:
-    reduction_dimension = len(op_dims) + reduction_dimension
-  comparator = _comparator_builder(op_type, is_max_k)
-  init_val_literal = _get_init_val_literal(op_type, is_max_k)
-  iota = xc.ops.Iota(c, xc.Shape.array_shape(np.dtype(np.int32), op_dims),
-                     reduction_dimension)
-  init_val = xc.ops.Constant(c, init_val_literal)
-  init_arg = xc.ops.Constant(c, np.int32(-1))
-  out = xc.ops.ApproxTopK(c, [operand, iota], [init_val, init_arg], k,
-                          reduction_dimension, comparator, recall_target,
-                          aggregate_to_topk, reduction_input_size_override)
-  return xla.xla_destructure(c, out)
-
-
 def _comparator_builder_mlir(ctx, op_type, is_max_k):
   scalar = ir.RankedTensorType.get([], op_type)
   index = ir.RankedTensorType.get([], ir.IntegerType.get_signless(32))
@@ -295,8 +279,8 @@ def _comparator_builder_mlir(ctx, op_type, is_max_k):
   with ir.InsertionPoint(entry_block):
     p0, p1, _, _ = entry_block.arguments
     direction = hlo.ComparisonDirectionAttr.get('GT' if is_max_k else 'LT')
-    cmp_result = hlo.CompareOp(p0, p1, comparison_direction=direction)
-    hlo.ReturnOp(cmp_result)
+    cmp_result = hlo.compare(p0, p1, comparison_direction=direction)
+    hlo.return_([cmp_result])
 
   return comparator
 
@@ -321,12 +305,11 @@ def _approx_top_k_lowering(ctx, operand, *, k,
   iota = mlir.iota(ctx, core.ShapedArray(ctx.avals_in[0].shape, np.int32),
                    dimension=reduction_dimension)
 
-  init_arg = hlo.ConstantOp(ir.DenseElementsAttr.get(np.int32(-1))).result
+  init_arg = hlo.constant(ir.DenseElementsAttr.get(np.int32(-1)))
   init_val_array = _get_init_val_literal(ctx.avals_in[0].dtype, is_max_k)
   init_val = mlir.ir_constant(init_val_array.reshape(()))
 
   backend_config = {
-    "top_k" : mlir.i64_attr(k),
     "reduction_dim" : mlir.i64_attr(reduction_dimension),
     "recall_target" : mlir.ir.FloatAttr.get(recall_type, recall_target),
     "aggregate_to_topk" : mlir.ir.BoolAttr.get(aggregate_to_topk),
@@ -342,13 +325,24 @@ def _approx_top_k_lowering(ctx, operand, *, k,
         mlir.shape_tensor(mlir.eval_dynamic_shape(ctx, aval_out.shape))
         for aval_out in ctx.avals_out]
 
-  out = mlir.custom_call(
-      "ApproxTopK",
-      result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
-      operands=[operand, iota, init_val, init_arg],
-      called_computations=[comparator.name.value],
-      backend_config=backend_config,
-      result_shapes=result_shapes)
+  if core.is_constant_dim(k):
+    backend_config["top_k"] = mlir.i64_attr(k)
+    out = mlir.custom_call(
+        "ApproxTopK",
+        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
+        operands=[operand, iota, init_val, init_arg],
+        called_computations=[comparator.name.value],
+        backend_config=backend_config,
+        result_shapes=result_shapes)
+  else:
+    k_value, = mlir.eval_dynamic_shape_as_vals(ctx, (k,))
+    out = mlir.custom_call(
+        "stablehlo.dynamic_approx_top_k",
+        result_types=[mlir.aval_to_ir_type(aval) for aval in ctx.avals_out],
+        operands=[operand, iota, init_val, init_arg, k_value],
+        called_computations=[comparator.name.value],
+        backend_config=backend_config,
+        result_shapes=result_shapes)
 
   return out.results
 
