@@ -27,7 +27,6 @@ from jax._src.interpreters import ad
 from jax._src.interpreters import batching
 from jax._src.interpreters import mlir
 from jax._src.interpreters import xla
-from jax._src.lax import lax
 from jax._src.traceback_util import api_boundary
 from jax._src.util import split_list, safe_map
 import numpy as np
@@ -50,7 +49,7 @@ def _split_root_args(args, const_lengths):
 
 @api_boundary
 def custom_root(f, initial_guess, solve, tangent_solve, has_aux=False):
-  """Differentiably solve for a roots of a function.
+  """Differentiably solve for the roots of a function.
 
   This is a low-level routine, mostly intended for internal use in JAX.
   Gradients of custom_root() are defined with respect to closed-over variables
@@ -152,7 +151,7 @@ def _root_jvp(const_lengths, jaxprs, primals, tangents):
       operator.neg, linearize_and_solve(*solution, *rhs))
   # append aux, create symbolic zero tangents for the aux values
   solution += aux
-  solution_dot += _map(lax.zeros_like_array, aux)
+  solution_dot += _map(ad_util.zeros_like_jaxval, aux)
 
   return solution, solution_dot
 
