@@ -15,11 +15,12 @@
 # This module is largely a wrapper around `jaxlib` that performs version
 # checking on import.
 
-import datetime
+from __future__ import annotations
+
 import gc
 import pathlib
 import re
-from typing import Optional
+from typing import Any
 
 try:
   import jaxlib as jaxlib
@@ -82,16 +83,10 @@ version = check_jaxlib_version(
 import jaxlib.cpu_feature_guard as cpu_feature_guard
 cpu_feature_guard.check_cpu_features()
 
-try:
-  import jaxlib.cuda_plugin_extension as cuda_plugin_extension  # pytype: disable=import-error
-except ModuleNotFoundError:
-  cuda_plugin_extension = None
 import jaxlib.utils as utils
 import jaxlib.xla_client as xla_client
 import jaxlib.lapack as lapack
 
-import jaxlib.ducc_fft as ducc_fft
-
 xla_extension = xla_client._xla
 pytree = xla_client._xla.pytree
 jax_jit = xla_client._xla.jax_jit
@@ -105,7 +100,10 @@ gc.callbacks.append(_xla_gc_callback)
 try:
   import jaxlib.cuda._versions as cuda_versions  # pytype: disable=import-error
 except ImportError:
-  cuda_versions = None
+  try:
+    import jax_cuda12_plugin._versions as cuda_versions  # pytype: disable=import-error
+  except ImportError:
+    cuda_versions = None
 
 import jaxlib.gpu_solver as gpu_solver  # pytype: disable=import-error
 import jaxlib.gpu_sparse as gpu_sparse  # pytype: disable=import-error
@@ -122,28 +120,19 @@ xla_extension_version: int = getattr(xla_client, '_version', 0)
 import jaxlib.gpu_rnn as gpu_rnn  # pytype: disable=import-error
 import jaxlib.gpu_triton as gpu_triton # pytype: disable=import-error
 
-if version >= (0, 4, 14):
-  import jaxlib.tpu_mosaic as tpu_mosaic # pytype: disable=import-error
-else:
-  # Jaxlib doesn't contain Mosaic bindings
-  tpu_mosaic = None  # type: ignore
+import jaxlib.mosaic.python.tpu as tpu # pytype: disable=import-error
 
 # Version number for MLIR:Python APIs, provided by jaxlib.
 mlir_api_version = xla_client.mlir_api_version
 
 # TODO(rocm): check if we need the same for rocm.
 
-def _cuda_path() -> Optional[str]:
+def _cuda_path() -> str | None:
   _jaxlib_path = pathlib.Path(jaxlib.__file__).parent
   # If the pip package nvidia-cuda-nvcc-cu11 is installed, it should have
   # both of the things XLA looks for in the cuda path, namely bin/ptxas and
   # nvvm/libdevice/libdevice.10.bc
   path = _jaxlib_path.parent / "nvidia" / "cuda_nvcc"
-  if path.is_dir():
-    return str(path)
-  # Failing that, we use the copy of libdevice.10.bc we include with jaxlib and
-  # hope that the user has ptxas in their PATH.
-  path = _jaxlib_path / "cuda"
   if path.is_dir():
     return str(path)
   return None
@@ -151,3 +140,5 @@ def _cuda_path() -> Optional[str]:
 cuda_path = _cuda_path()
 
 transfer_guard_lib = xla_client._xla.transfer_guard_lib
+
+Device = xla_client._xla.Device
