@@ -63,16 +63,17 @@ data must be immutable, because it will be stored in function memoization tables
 """
 from __future__ import annotations
 
+from collections.abc import Callable
 from functools import partial
-import operator
-from typing import Any, Callable, NamedTuple
+from typing import Any, NamedTuple
 import weakref
 
-from jax._src.tree_util import tree_map
-from jax._src.config import config
+from jax._src import config
 from jax._src import core
 from jax._src import traceback_util
-from jax._src.util import curry
+from jax._src.tree_util import tree_map
+from jax._src.util import curry, cache_clearing_funs
+
 
 traceback_util.register_exclusion(__file__)
 
@@ -116,7 +117,9 @@ class EqualStore:
   def __init__(self):
     self._store = Store()
 
-  val = property(operator.attrgetter('_store.val'))
+  @property
+  def val(self):
+    return self._store.val
 
   def store(self, val):
     try:
@@ -244,8 +247,9 @@ def transformation(gen, fun: WrappedFun, *gen_static_args) -> WrappedFun:
   return fun.wrap(gen, gen_static_args, None)
 
 @curry
-def transformation_with_aux(gen, fun: WrappedFun, *gen_static_args,
-                            use_eq_store=False) -> tuple[WrappedFun, Any]:
+def transformation_with_aux(
+    gen, fun: WrappedFun, *gen_static_args, use_eq_store: bool = False
+) -> tuple[WrappedFun, Callable[[], Any]]:
   """Adds one more transformation with auxiliary output to a WrappedFun."""
   out_store = Store() if not use_eq_store else EqualStore()
   out_thunk = lambda: out_store.val
@@ -305,7 +309,7 @@ class TracingDebugInfo(NamedTuple):
   # formed just before staging to a jaxpr and read in trace-time error messages.
   # TODO(mattjj): delete partial_eval.DebugInfo, replace all uses with this cls
   traced_for: str             # e.g. 'jit', 'scan', etc
-  func_src_info: str          # e.g. f'{fun.__name__} at {filename}:{lineno}'
+  func_src_info: str | None   # e.g. f'{fun.__name__} at {filename}:{lineno}'
   arg_names: tuple[str, ...]  # e.g. ('args[0]', ... )
   result_paths: Callable[[], tuple[str, ...]] | None
 
@@ -318,7 +322,7 @@ def add_debug_info(f: WrappedFun, debug_info: TracingDebugInfo | None
   return WrappedFun(f.f, f.transforms, f.stores, f.params, f.in_type, debug_info)
 
 
-def cache(call: Callable):
+def cache(call: Callable, *, explain: Callable | None = None):
   """Memoization decorator for functions taking a WrappedFun as first argument.
 
   Args:
@@ -332,20 +336,22 @@ def cache(call: Callable):
   fun_caches: weakref.WeakKeyDictionary = weakref.WeakKeyDictionary()
 
   def memoized_fun(fun: WrappedFun, *args):
-    cache = fun_caches.setdefault(fun.f, {})
-    if config.jax_check_tracer_leaks:
+    cache = fun_caches.setdefault(fun.f, new_cache := {})  # type: ignore
+    if config.check_tracer_leaks.value:
       key = (_copy_main_traces(fun.transforms), fun.params, fun.in_type, args,
-             config.x64_enabled, config.jax_default_device,
-             config._trace_context())
+             config.enable_x64.value, config.default_device.value,
+             config.trace_context())
     else:
-      key = (fun.transforms, fun.params, fun.in_type, args, config.x64_enabled,
-             config.jax_default_device, config._trace_context())
+      key = (fun.transforms, fun.params, fun.in_type, args, config.enable_x64.value,
+             config.default_device.value, config.trace_context())
     result = cache.get(key, None)
     if result is not None:
       ans, stores = result
       fun.populate_stores(stores)
     else:
       ans = call(fun, *args)
+      if explain and config.explain_cache_misses.value:
+        explain(fun.f, cache is new_cache, cache, key)
       cache[key] = (ans, fun.stores)
 
     return ans
@@ -355,17 +361,9 @@ def cache(call: Callable):
 
   memoized_fun.cache_clear = fun_caches.clear  # type: ignore
   memoized_fun.evict_function = _evict_function  # type: ignore
-
   cache_clearing_funs.add(memoized_fun.cache_clear)
-
   return memoized_fun
 
-cache_clearing_funs = weakref.WeakSet()  # type: ignore
-
-def clear_all_caches():
-  global cache_clearing_funs
-  for clear in cache_clearing_funs:
-    clear()
 
 @partial(partial, tree_map)
 def _copy_main_traces(x):
