@@ -14,25 +14,37 @@
 
 """Shared neural network activations and other functions."""
 
+from __future__ import annotations
+
+from collections.abc import Sequence
 from functools import partial
 import operator
-import warnings
 import numpy as np
-from typing import Any, Optional, Union
+from typing import Any, Literal
+import warnings
 
 import jax
 import jax.numpy as jnp
 from jax import custom_jvp
 from jax import lax
+from jax._src import config
 from jax._src import core
 from jax._src import dtypes
 from jax._src import util
 from jax._src.core import AxisName
+from jax._src.cudnn.fused_attention_stablehlo import (
+    dot_product_attention as cudnn_dot_product_attention, MaskType)
 from jax._src.numpy import util as numpy_util
 from jax._src.typing import Array, ArrayLike
 from jax._src.ops.special import logsumexp as _logsumexp
 
 
+class Unspecified:
+  def __repr__(self):
+    return "_UNSPECIFIED"
+_UNSPECIFIED = Unspecified()
+
+
 # activations
 
 @custom_jvp
@@ -60,7 +72,7 @@ def relu(x: ArrayLike) -> Array:
   Returns:
     An array.
 
-  Example:
+  Examples:
     >>> jax.nn.relu(jax.numpy.array([-2., -1., -0.5, 0, 0.5, 1., 2.]))
     Array([0. , 0. , 0. , 0. , 0.5, 1. , 2. ], dtype=float32)
 
@@ -72,6 +84,28 @@ def relu(x: ArrayLike) -> Array:
 # For behavior at 0, see https://openreview.net/forum?id=urrcVI-_jRm
 relu.defjvps(lambda g, ans, x: lax.select(x > 0, g, lax.full_like(g, 0)))
 
+@jax.jit
+def squareplus(x: ArrayLike, b: ArrayLike = 4) -> Array:
+  r"""Squareplus activation function.
+
+  Computes the element-wise function
+
+  .. math::
+    \mathrm{squareplus}(x) = \frac{x + \sqrt{x^2 + b}}{2}
+
+  as described in https://arxiv.org/abs/2112.11687.
+
+  Args:
+    x : input array
+    b : smoothness parameter
+  """
+  numpy_util.check_arraylike("squareplus", x)
+  numpy_util.check_arraylike("squareplus", b)
+  x = jnp.asarray(x)
+  b = jnp.asarray(b)
+  y = x + jnp.sqrt(jnp.square(x) + b)
+  return y / 2
+
 @jax.jit
 def softplus(x: ArrayLike) -> Array:
   r"""Softplus activation function.
@@ -86,6 +120,32 @@ def softplus(x: ArrayLike) -> Array:
   """
   return jnp.logaddexp(x, 0)
 
+@jax.jit
+def sparse_plus(x: ArrayLike) -> Array:
+  r"""Sparse plus function.
+
+  Computes the function:
+
+  .. math::
+
+    \mathrm{sparse\_plus}(x) = \begin{cases}
+      0, & x \leq -1\\
+      \frac{1}{4}(x+1)^2, & -1 < x < 1 \\
+      x, & 1 \leq x
+    \end{cases}
+
+  This is the twin function of the softplus activation ensuring a zero output
+  for inputs less than -1 and a linear output for inputs greater than 1,
+  while remaining smooth, convex, monotonic by an adequate definition between
+  -1 and 1.
+
+  Args:
+    x: input (float)
+  """
+  numpy_util.check_arraylike("sparse_plus", x)
+  x = jnp.asarray(x)
+  return jnp.where(x <= -1.0, 0.0, jnp.where(x >= 1.0, x, (x + 1.0)**2/4))
+
 @jax.jit
 def soft_sign(x: ArrayLike) -> Array:
   r"""Soft-sign activation function.
@@ -102,7 +162,7 @@ def soft_sign(x: ArrayLike) -> Array:
   x_arr = jnp.asarray(x)
   return x_arr / (jnp.abs(x_arr) + 1)
 
-@jax.jit
+@partial(jax.jit, inline=True)
 def sigmoid(x: ArrayLike) -> Array:
   r"""Sigmoid activation function.
 
@@ -123,9 +183,41 @@ def sigmoid(x: ArrayLike) -> Array:
   """
   return lax.logistic(x)
 
+@jax.jit
+def sparse_sigmoid(x: ArrayLike) -> Array:
+  r"""Sparse sigmoid activation function.
+
+  Computes the function:
+
+  .. math::
+
+    \mathrm{sparse\_sigmoid}(x) = \begin{cases}
+      0, & x \leq -1\\
+      \frac{1}{2}(x+1), & -1 < x < 1 \\
+      1, & 1 \leq x
+    \end{cases}
+
+  This is the twin function of the ``sigmoid`` activation ensuring a zero output
+  for inputs less than -1, a 1 output for inputs greater than 1, and a linear
+  output for inputs between -1 and 1. It is the derivative of ``sparse_plus``.
+
+  For more information, see `Learning with Fenchel-Young Losses (section 6.2)
+  <https://arxiv.org/abs/1901.02324>`_.
+
+  Args:
+    x : input array
+
+  Returns:
+    An array.
+
+  See also:
+    :func:`sigmoid`
+  """
+  return 0.5 * jnp.clip(x + 1.0, 0.0, 2.0)
+
 @jax.jit
 def silu(x: ArrayLike) -> Array:
-  r"""SiLU (a.k.a. swish) activation function.
+  r"""SiLU (aka swish) activation function.
 
   Computes the element-wise function:
 
@@ -149,6 +241,29 @@ def silu(x: ArrayLike) -> Array:
 
 swish = silu
 
+@jax.jit
+def mish(x: ArrayLike) -> Array:
+  r"""Mish activation function.
+
+  Computes the element-wise function:
+
+  .. math::
+    \mathrm{mish}(x) = x \cdot \mathrm{tanh}(\mathrm{softplus}(x))
+
+  For more information, see
+  `Mish: A Self Regularized Non-Monotonic Activation Function
+  <https://arxiv.org/abs/1908.08681>`_.
+
+  Args:
+    x : input array
+
+  Returns:
+    An array.
+  """
+  numpy_util.check_arraylike("mish", x)
+  x_arr = jnp.asarray(x)
+  return x_arr * jnp.tanh(softplus(x_arr))
+
 @jax.jit
 def log_sigmoid(x: ArrayLike) -> Array:
   r"""Log-sigmoid activation function.
@@ -264,7 +379,7 @@ def celu(x: ArrayLike, alpha: ArrayLike = 1.0) -> Array:
 
   For more information, see
   `Continuously Differentiable Exponential Linear Units
-  <https://arxiv.org/pdf/1704.07483.pdf>`_.
+  <https://arxiv.org/abs/1704.07483>`_.
 
   Args:
     x : input array
@@ -292,7 +407,7 @@ def selu(x: ArrayLike) -> Array:
 
   For more information, see
   `Self-Normalizing Neural Networks
-  <https://papers.nips.cc/paper/6698-self-normalizing-neural-networks.pdf>`_.
+  <https://arxiv.org/abs/1706.02515>`_.
 
   Args:
     x : input array
@@ -379,9 +494,9 @@ logsumexp = _logsumexp
 
 @partial(jax.jit, static_argnames=("axis",))
 def log_softmax(x: ArrayLike,
-                axis: Optional[Union[int, tuple[int, ...]]] = -1,
-                where: Optional[ArrayLike] = None,
-                initial: Optional[ArrayLike] = None) -> Array:
+                axis: int | tuple[int, ...] | None = -1,
+                where: ArrayLike | None = None,
+                initial: ArrayLike | None | Unspecified = _UNSPECIFIED) -> Array:
   r"""Log-Softmax function.
 
   Computes the logarithm of the :code:`softmax` function, which rescales
@@ -396,19 +511,27 @@ def log_softmax(x: ArrayLike,
     axis: the axis or axes along which the :code:`log_softmax` should be
       computed. Either an integer or a tuple of integers.
     where: Elements to include in the :code:`log_softmax`.
-    initial: The minimum value used to shift the input array. Must be present
-      when :code:`where` is not None.
 
   Returns:
     An array.
 
+  Note:
+    If any input values are ``+inf``, the result will be all ``NaN``: this reflects the
+    fact that ``inf / inf`` is not well-defined in the context of floating-point math.
+
   See also:
     :func:`softmax`
   """
+  if initial is not _UNSPECIFIED:
+    # Added 2024-4-10
+    warnings.warn("The initial argument to log_softmax is deprecated, and no longer has any effect.",
+                  DeprecationWarning, stacklevel=2)
+  del initial
   numpy_util.check_arraylike("log_softmax", x)
   x_arr = jnp.asarray(x)
-  x_max = jnp.max(x_arr, axis, where=where, initial=initial, keepdims=True)
-  shifted = x_arr - lax.stop_gradient(x_max)
+  x_max = jnp.max(x_arr, axis, where=where, initial=-jnp.inf, keepdims=True)
+  x_safe = x_arr if where is None else jnp.where(where, x_arr, -jnp.inf)
+  shifted = x_safe - lax.stop_gradient(x_max)
   shifted_logsumexp = jnp.log(
       jnp.sum(jnp.exp(shifted), axis, where=where, keepdims=True))
   result = shifted - shifted_logsumexp
@@ -420,9 +543,9 @@ def log_softmax(x: ArrayLike,
 # TODO(phawkins): this jit was found to change numerics in a test. Debug this.
 #@partial(jax.jit, static_argnames=("axis",))
 def softmax(x: ArrayLike,
-            axis: Optional[Union[int, tuple[int, ...]]] = -1,
-            where: Optional[ArrayLike] = None,
-            initial: Optional[ArrayLike] = None) -> Array:
+            axis: int | tuple[int, ...] | None = -1,
+            where: ArrayLike | None = None,
+            initial: ArrayLike | None | Unspecified = _UNSPECIFIED) -> Array:
   r"""Softmax function.
 
   Computes the function which rescales elements to the range :math:`[0, 1]`
@@ -437,32 +560,40 @@ def softmax(x: ArrayLike,
       softmax output summed across these dimensions should sum to :math:`1`.
       Either an integer or a tuple of integers.
     where: Elements to include in the :code:`softmax`.
-    initial: The minimum value used to shift the input array. Must be present
-      when :code:`where` is not None.
 
   Returns:
     An array.
 
+  Note:
+    If any input values are ``+inf``, the result will be all ``NaN``: this reflects the
+    fact that ``inf / inf`` is not well-defined in the context of floating-point math.
+
   See also:
     :func:`log_softmax`
   """
-  if jax.config.jax_softmax_custom_jvp:
+  if initial is not _UNSPECIFIED:
+    # Added 2024-4-10
+    warnings.warn("The initial argument to softmax is deprecated, and no longer has any effect.",
+                  DeprecationWarning, stacklevel=2)
+  del initial
+  if config.softmax_custom_jvp.value:
     # mypy is confused by the `functools.partial` application in the definition
     # of `_softmax` and incorrectly concludes that `_softmax` returns
     # `ReturnValue` -- the unsubstituted type parameter of `custom_jvp`.
-    return _softmax(x, axis, where, initial)  # type: ignore[return-value]
+    return _softmax(x, axis, where)
   else:
-    return _softmax_deprecated(x, axis, where, initial)
+    return _softmax_deprecated(x, axis, where)
 
 # TODO(mattjj): replace softmax with _softmax when deprecation flag is removed
 @partial(jax.custom_jvp, nondiff_argnums=(1,))
 def _softmax(
     x: ArrayLike,
-    axis: Optional[Union[int, tuple[int, ...]]] = -1,
-    where: Optional[ArrayLike] = None,
-    initial: Optional[ArrayLike] = None) -> Array:
+    axis: int | tuple[int, ...] | None = -1,
+    where: ArrayLike | None = None,
+    initial: ArrayLike | None = -jnp.inf) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
-  unnormalized = jnp.exp(x - x_max)
+  x_safe = x if where is None else jnp.where(where, x, initial)
+  unnormalized = jnp.exp(x_safe - x_max)
   result = unnormalized / jnp.sum(unnormalized, axis, where=where, keepdims=True)
   if where is not None:
     result = jnp.where(where, result, 0)
@@ -476,11 +607,12 @@ def _softmax_jvp(axis, primals, tangents):
 
 def _softmax_deprecated(
     x: ArrayLike,
-    axis: Optional[Union[int, tuple[int, ...]]] = -1,
-    where: Optional[ArrayLike] = None,
-    initial: Optional[ArrayLike] = None) -> Array:
+    axis: int | tuple[int, ...] | None = -1,
+    where: ArrayLike | None = None,
+    initial: ArrayLike | None = -jnp.inf) -> Array:
   x_max = jnp.max(x, axis, where=where, initial=initial, keepdims=True)
-  unnormalized = jnp.exp(x - lax.stop_gradient(x_max))
+  x_safe = x if where is None else jnp.where(where, x, initial)
+  unnormalized = jnp.exp(x_safe - lax.stop_gradient(x_max))
   result = unnormalized / jnp.sum(unnormalized, axis, where=where, keepdims=True)
   if where is not None:
     result = jnp.where(where, result, 0)
@@ -489,11 +621,11 @@ def _softmax_deprecated(
 
 @partial(jax.jit, static_argnames=("axis",))
 def standardize(x: ArrayLike,
-              axis: Optional[Union[int, tuple[int, ...]]] = -1,
-              mean: Optional[ArrayLike] = None,
-              variance: Optional[ArrayLike] = None,
+              axis: int | tuple[int, ...] | None = -1,
+              mean: ArrayLike | None = None,
+              variance: ArrayLike | None = None,
               epsilon: ArrayLike = 1e-5,
-              where: Optional[ArrayLike] = None) -> Array:
+              where: ArrayLike | None = None) -> Array:
   r"""Normalizes an array by subtracting ``mean`` and dividing by :math:`\sqrt{\mathrm{variance}}`."""
   numpy_util.check_arraylike("standardize", x)
   numpy_util.check_arraylike_or_none("standardize", mean, variance, where)
@@ -508,20 +640,10 @@ def standardize(x: ArrayLike,
         jnp.square(x), axis, keepdims=True, where=where) - jnp.square(mean)
   return jnp.subtract(x, jnp.asarray(mean)) * lax.rsqrt(jnp.asarray(variance) + epsilon)
 
-def normalize(x: ArrayLike,
-            axis: Optional[Union[int, tuple[int, ...]]] = -1,
-            mean: Optional[ArrayLike] = None,
-            variance: Optional[ArrayLike] = None,
-            epsilon: ArrayLike = 1e-5,
-            where: Optional[ArrayLike] = None) -> Array:
-  r"""Normalizes an array by subtracting ``mean`` and dividing by :math:`\sqrt{\mathrm{variance}}`."""
-  warnings.warn("jax.nn.normalize will be deprecated. Use jax.nn.standardize instead.", DeprecationWarning)
-  return standardize(x, axis, mean, variance, epsilon, where)
-
 # TODO(slebedev): Change the type of `x` to `ArrayLike`.
 @partial(jax.jit, static_argnames=("num_classes", "dtype", "axis"))
 def _one_hot(x: Any, num_classes: int, *,
-             dtype: Any, axis: Union[int, AxisName]) -> Array:
+             dtype: Any, axis: int | AxisName) -> Array:
   num_classes = core.concrete_dim_or_error(
       num_classes,
       "The error arose in jax.nn.one_hot argument `num_classes`.")
@@ -545,7 +667,7 @@ def _one_hot(x: Any, num_classes: int, *,
 
 # TODO(slebedev): Change the type of `x` to `ArrayLike`.
 def one_hot(x: Any, num_classes: int, *,
-            dtype: Any = jnp.float_, axis: Union[int, AxisName] = -1) -> Array:
+            dtype: Any = jnp.float_, axis: int | AxisName = -1) -> Array:
   """One-hot encodes the given indices.
 
   Each index in the input ``x`` is encoded as a vector of zeros of length
@@ -654,3 +776,146 @@ def hard_silu(x: ArrayLike) -> Array:
   return x_arr * hard_sigmoid(x_arr)
 
 hard_swish = hard_silu
+
+def _get_large_negative(dtype):
+  dtype_max = jnp.finfo(dtype).max
+  return jnp.asarray(-0.7 * dtype_max, dtype=dtype)
+
+def _get_causal_mask(T, S, dtype):
+  pred = jnp.tril(jnp.ones((T, S), dtype=jnp.bool_))
+  mask = jnp.where(pred, jnp.asarray(0.0, dtype), _get_large_negative(dtype))
+  return mask[jnp.newaxis, jnp.newaxis, :, :]
+
+def _dot_product_attention_xla(
+    query: Array,
+    key: Array,
+    value: Array,
+    bias: Array | None,
+    mask: Array | None,
+    is_causal: bool,
+    scale: float):
+  logits_dtype = jnp.promote_types(query.dtype, jnp.float32)
+  logits = jnp.einsum('BTNH,BSNH->BNTS', query, key,
+                      preferred_element_type=logits_dtype)
+
+  logits *= jnp.array(scale, dtype=logits.dtype)
+
+  if bias is not None:
+    logits = (logits + bias).astype(logits.dtype)
+
+  if mask is not None:
+    assert mask.dtype == jnp.bool_
+    large_negative_number = _get_large_negative(logits.dtype)
+    padded_logits = jnp.where(mask, logits, large_negative_number)
+  else:
+    padded_logits = logits
+
+  if is_causal:
+    T, S = query.shape[-3], key.shape[-3]
+    mask = _get_causal_mask(T, S, logits.dtype)
+    padded_logits = padded_logits + mask
+
+  # Softmax and it is always carried out in fp32.
+  padded_logits = padded_logits.astype(jnp.float32)
+  probs = jax.nn.softmax(padded_logits, axis=-1).astype(key.dtype)
+
+  encoded = jnp.einsum('BNTS,BSNH->BTNH', probs, value)
+  return encoded
+
+def dot_product_attention(
+    query: ArrayLike,
+    key: ArrayLike,
+    value: ArrayLike,
+    *,
+    bias: ArrayLike | None = None,
+    mask: ArrayLike | None = None,
+    scale: float | None = None,
+    is_causal: bool = False,
+    implementation: Literal['xla', 'cudnn'] | None = None) -> Array:
+  r"""Scaled dot product attention function.
+
+  Computes the attention function on Query, Key, and Value tensors:
+
+  .. math::
+
+    \mathrm{Attention}(Q, K, V)=\mathrm{softmax}(\frac{QK^T}{\sqrt{d_k}})V
+
+  If we define :code:`logits` as the output of :math:`QK^T` and the
+  :code:`probs` as the output of :math:`softmax`.
+
+  Throughout this function, we utilize the following uppercase letters to
+  represent the shape of array:
+
+    B = batch size
+    S = length of the key/value (source)
+    T = length of the query (target)
+    N = number of attention heads
+    H = dimensions of each attention head
+
+  Args:
+    query: query array; shape :code:`(BTNH)`
+    key: key array; shape :code:`(BSNH)`
+    value: value array; shape :code:`(BSNH)`
+    bias: optional, bias array to be added to logits; shape broadcastable to
+      :code:`(BNTS)`.
+    mask: optional, mask array used to filter out logits. It is a boolean mask
+      where `True` indicates the element should take part in attention. For an
+      additive mask, users should pass it to `bias`. The shape is broadcastable
+      to :code:`(BNTS)`.
+    scale: scale for the logits. If None, the scale will be set to 1 divided by
+      the square root of query's head dimension (i.e. H).
+    is_causal: If true, causal attention will be applied. Note, some
+      implementations like `xla` will generate a mask tensor and apply it to the
+      logits to mask out the non-causal parts of the attention matrix, but other
+      implementations like `cudnn` will avoid computing the non-causal regions,
+      providing speedups.
+    implementation: A string to control which implementation backend to use.
+      Supported strings are `xla`, `cudnn` (cuDNN flash attention). It defaults
+      to `None`, which will automatically select the best available backend.
+      Note, `cudnn` supports only a subset of shapes/dtypes, and an exception
+      will be thrown if its not supported.
+
+  Returns:
+    An array of the attention output with the same shape as :code:`query`.
+  """
+  def _check_has_shape(t: Array, shape: Sequence[int], name: str) -> None:
+    if t.ndim != len(shape):
+      raise ValueError(f"{name} ndim should be {len(shape)}, but got {t.ndim}")
+    for i in range(t.ndim):
+      if shape[i] != -1 and t.shape[i] != shape[i]:
+        raise ValueError(f"{name} shape should be {shape}: but got {t.shape}")
+
+  query = jnp.asarray(query)
+  key = jnp.asarray(key)
+  value = jnp.asarray(value)
+  bias = bias if bias is None else jnp.asarray(bias)
+  mask = mask if mask is None else jnp.asarray(mask)
+
+  B, S, N, H = key.shape
+  _check_has_shape(value, [B, S, N, H], 'value')
+  _check_has_shape(query, [B, -1, N, H], 'query')
+  scale_val = (1.0 / np.sqrt(H)) if scale is None else scale
+  if not (query.dtype == key.dtype == value.dtype):
+    raise ValueError(f"query/key/value should have the same dtype, but got "
+                     f"{query.dtype} vs {key.dtype} vs {value.dtype}.")
+  if mask is not None and mask.dtype != jnp.bool_:
+    raise ValueError(f"Mask must be boolean dtype, but got {mask.dtype}.")
+
+  match implementation:
+    case 'xla':
+      return _dot_product_attention_xla(
+          query, key, value, bias, mask, is_causal=is_causal, scale=scale_val,
+      )
+    case 'cudnn':
+      mask_type = MaskType.CAUSAL if is_causal else MaskType.NO_MASK
+      return cudnn_dot_product_attention(
+          query, key, value, bias, mask, scale=scale_val, mask_type=mask_type
+      )
+    case None:
+      # TODO(kaixih@nvidia) Defaults to XLA for now. Will automatically select
+      # best backend.
+      return _dot_product_attention_xla(
+          query, key, value, bias, mask, is_causal=is_causal, scale=scale_val,
+      )
+    case _:
+      raise ValueError(f"Unsupported implementation option: {implementation}")
