@@ -19,20 +19,26 @@
 This is done dynamically in order to avoid circular imports.
 """
 
+from __future__ import annotations
+
 __all__ = ['register_jax_array_methods']
 
 import abc
 from functools import partial, wraps
-from typing import Any, Optional, Union
+import math
+from typing import Any
 
 import numpy as np
 import jax
 from jax import lax
+from jax.sharding import Sharding
+from jax._src import api
 from jax._src import core
 from jax._src import dtypes
 from jax._src.api_util import _ensure_index_tuple
 from jax._src.array import ArrayImpl
 from jax._src.lax import lax as lax_internal
+from jax._src.lib import xla_client as xc
 from jax._src.numpy import lax_numpy
 from jax._src.numpy import reductions
 from jax._src.numpy import ufuncs
@@ -52,7 +58,7 @@ zip, unsafe_zip = safe_zip, zip
 # functions, which can themselves handle instances from any of these classes.
 
 
-def _astype(arr: ArrayLike, dtype: DTypeLike) -> Array:
+def _astype(arr: ArrayLike, dtype: DTypeLike, copy: bool = False, device: xc.Device | Sharding | None = None) -> Array:
   """Copy the array and cast to a specified dtype.
 
   This is implemented via :func:`jax.lax.convert_element_type`, which may
@@ -60,10 +66,13 @@ def _astype(arr: ArrayLike, dtype: DTypeLike) -> Array:
   some cases. In particular, the details of float-to-int and int-to-float
   casts are implementation dependent.
   """
-  if dtype is None:
-    dtype = dtypes.canonicalize_dtype(lax_numpy.float_)
-  dtypes.check_user_dtype_supported(dtype, "astype")
-  return lax.convert_element_type(arr, dtype)
+  return lax_numpy.astype(arr, dtype, copy=copy, device=device)
+
+def _to_device(arr: ArrayLike, device: xc.Device | Sharding, *,
+               stream: int | Any | None = None):
+  if stream is not None:
+    raise NotImplementedError("stream argument of array.to_device()")
+  return api.device_put(arr, device)
 
 
 def _nbytes(arr: ArrayLike) -> int:
@@ -71,19 +80,12 @@ def _nbytes(arr: ArrayLike) -> int:
   return np.size(arr) * dtypes.dtype(arr, canonicalize=True).itemsize
 
 
-def _item(a: Array) -> Any:
+def _item(a: Array, *args) -> bool | int | float | complex:
   """Copy an element of an array to a standard Python scalar and return it."""
-  if dtypes.issubdtype(a.dtype, np.complexfloating):
-    return complex(a)
-  elif dtypes.issubdtype(a.dtype, np.floating):
-    return float(a)
-  elif dtypes.issubdtype(a.dtype, np.integer):
-    return int(a)
-  elif dtypes.issubdtype(a.dtype, np.bool_):
-    return bool(a)
-  else:
-    raise TypeError(a.dtype)
-
+  arr = core.concrete_or_error(np.asarray, a, context="This occurred in the item() method of jax.Array")
+  if dtypes.issubdtype(a.dtype, dtypes.extended):
+    raise TypeError(f"No Python scalar type for {a.dtype=}")
+  return arr.item(*args)
 
 def _itemsize(arr: ArrayLike) -> int:
   """Length of one array element in bytes."""
@@ -91,12 +93,11 @@ def _itemsize(arr: ArrayLike) -> int:
 
 
 def _clip(number: ArrayLike,
-          min: Optional[ArrayLike] = None, max: Optional[ArrayLike] = None,
-          out: None = None) -> Array:
+          min: ArrayLike | None = None, max: ArrayLike | None = None) -> Array:
   """Return an array whose values are limited to a specified range.
 
   Refer to :func:`jax.numpy.clip` for full documentation."""
-  return lax_numpy.clip(number, a_min=min, a_max=max, out=out)
+  return lax_numpy.clip(number, min=min, max=max)
 
 
 def _transpose(a: Array, *args: Any) -> Array:
@@ -113,24 +114,37 @@ def _transpose(a: Array, *args: Any) -> Array:
   return lax_numpy.transpose(a, axis)
 
 
-def _compute_newshape(a: ArrayLike, newshape: Union[DimSize, Shape]) -> Shape:
+def _compute_newshape(a: ArrayLike, newshape: DimSize | Shape) -> Shape:
   """Fixes a -1 value in newshape, if present."""
-  # other errors, like having more than one -1, are caught downstream, in
-  # reshape_shape_rule.
+  orig_newshape = newshape  # for error messages
   try:
     iter(newshape)  # type: ignore[arg-type]
   except:
     newshape = [newshape]
   newshape = core.canonicalize_shape(newshape)  # type: ignore[arg-type]
   neg1s = [i for i, d in enumerate(newshape) if type(d) is int and d == -1]
-  if len(neg1s) == 1:
+  if len(neg1s) > 1:
+    raise TypeError("can only specify one unknown axis size with a `-1` value, "
+                    f"got {orig_newshape}")
+  if neg1s:
     i, = neg1s
-    sz = core.cancel_divide_tracers(np.shape(a), (*newshape[:i], *newshape[i+1:]))
+    other_sizes = (*newshape[:i], *newshape[i+1:])
+    if (all(isinstance(d, int) for d in (*np.shape(a), *other_sizes)) and
+        np.size(a) % math.prod(other_sizes) != 0):
+      raise TypeError(f"cannot reshape array of shape {np.shape(a)} (size {np.size(a)}) "
+                      f"into shape {orig_newshape} because the product of "
+                      f"specified axis sizes ({math.prod(other_sizes)}) does "
+                      f"not evenly divide {np.size(a)}")
+    sz = core.cancel_divide_tracers(np.shape(a), other_sizes)
     if sz is not None:
       return (*newshape[:i], sz, *newshape[i+1:])
+  else:
+    if (all(isinstance(d, int) for d in (*np.shape(a), *newshape)) and
+        np.size(a) != math.prod(newshape)):
+      raise TypeError(f"cannot reshape array of shape {np.shape(a)} (size {np.size(a)}) "
+                      f"into shape {orig_newshape} (size {math.prod(newshape)})")
   return tuple(-core.divide_shape_sizes(np.shape(a), newshape)
-               if core.definitely_equal(d, -1) else d
-               for d in newshape)
+               if core.definitely_equal(d, -1) else d for d in newshape)
 
 
 def _reshape(a: Array, *args: Any, order: str = "C") -> Array:
@@ -138,6 +152,7 @@ def _reshape(a: Array, *args: Any, order: str = "C") -> Array:
 
   Refer to :func:`jax.numpy.reshape` for full documentation.
   """
+  __tracebackhide__ = True
   newshape = _compute_newshape(a, args[0] if len(args) == 1 else args)
   if order == "C":
     return lax.reshape(a, newshape, None)
@@ -150,7 +165,7 @@ def _reshape(a: Array, *args: Any, order: str = "C") -> Array:
     raise ValueError(f"Unexpected value for 'order' argument: {order}.")
 
 
-def _view(arr: Array, dtype: DTypeLike = None, type: None = None) -> Array:
+def _view(arr: Array, dtype: DTypeLike | None = None, type: None = None) -> Array:
   """Return a bitwise copy of the array, viewed as a new dtype.
 
   This is fuller-featured wrapper around :func:`jax.lax.bitcast_convert_type`.
@@ -240,6 +255,7 @@ def _view(arr: Array, dtype: DTypeLike = None, type: None = None) -> Array:
 
 
 def _notimplemented_flat(self):
+  """Not implemented: Use :meth:`~jax.Array.flatten` instead."""
   raise NotImplementedError("JAX Arrays do not implement the arr.flat property: "
                             "consider arr.flatten() instead.")
 
@@ -269,7 +285,7 @@ def _unimplemented_setitem(self, i, x):
          "https://jax.readthedocs.io/en/latest/_autosummary/jax.numpy.ndarray.at.html")
   raise TypeError(msg.format(type(self)))
 
-def _operator_round(number: ArrayLike, ndigits: Optional[int] = None) -> Array:
+def _operator_round(number: ArrayLike, ndigits: int | None = None) -> Array:
   out = lax_numpy.round(number, decimals=ndigits or 0)
   # If `ndigits` is None, for a builtin float round(7.5) returns an integer.
   return out.astype(int) if ndigits is None else out
@@ -295,11 +311,13 @@ def __array_module__(self, types):
 
 
 def _compress_method(a: ArrayLike, condition: ArrayLike,
-                     axis: Optional[int] = None, out: None = None) -> Array:
+                     axis: int | None = None, *, out: None = None,
+                     size: int | None = None, fill_value: ArrayLike = 0) -> Array:
   """Return selected slices of this array along given axis.
 
   Refer to :func:`jax.numpy.compress` for full documentation."""
-  return lax_numpy.jaxcompress(condition, a, axis, out)
+  return lax_numpy.compress(condition, a, axis=axis, out=out,
+                            size=size, fill_value=fill_value)
 
 
 @core.stash_axis_env()
@@ -435,7 +453,7 @@ class _IndexUpdateHelper:
     return _IndexUpdateRef(self.array, index)
 
   def __repr__(self):
-    return f"_IndexUpdateHelper({repr(self.array)})"
+    return f"_IndexUpdateHelper({self.array!r})"
 
 
 class _IndexUpdateRef:
@@ -452,7 +470,7 @@ class _IndexUpdateRef:
     self.index = index
 
   def __repr__(self):
-    return f"_IndexUpdateRef({repr(self.array)}, {repr(self.index)})"
+    return f"_IndexUpdateRef({self.array!r}, {self.index!r})"
 
   def get(self, *, indices_are_sorted=False, unique_indices=False,
           mode=None, fill_value=None):
@@ -683,6 +701,7 @@ _array_methods = {
   "sum": reductions.sum,
   "swapaxes": lax_numpy.swapaxes,
   "take": lax_numpy.take,
+  "to_device": _to_device,
   "trace": lax_numpy.trace,
   "transpose": _transpose,
   "var": reductions.var,
@@ -730,6 +749,7 @@ def _forward_operator_to_aval(name):
 
 def _forward_method_to_aval(name):
   def meth(self, *args, **kwargs):
+    __tracebackhide__ = True
     return getattr(self.aval, name).fun(self, *args, **kwargs)
   return meth
 
@@ -800,5 +820,3 @@ def register_jax_array_methods():
   _set_array_attributes(ArrayImpl)
 
   _set_array_abstract_methods(Array)
-
-  Array.at.__doc__ = _IndexUpdateHelper.__doc__
