@@ -11,22 +11,23 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 from functools import partial
 import re
 import textwrap
-from typing import Any, Callable, NamedTuple, Optional, TypeVar
+from typing import Any, NamedTuple, TypeVar
 
 import warnings
 
-from jax._src import dtypes
 from jax._src import api
+from jax._src import config
 from jax._src import core
-from jax._src.config import config
+from jax._src import dtypes
 from jax._src.lax import lax
 from jax._src.util import safe_zip, safe_map
-from jax._src.typing import Array, ArrayLike, DType, DTypeLike, Shape
+from jax._src.typing import Array, ArrayLike, DimSize, DType, DTypeLike, Shape
 
 import numpy as np
 
@@ -49,14 +50,14 @@ class ParsedDoc(NamedTuple):
   front_matter: front matter before sections.
   sections: dictionary of section titles to section content.
   """
-  docstr: Optional[str]
+  docstr: str | None
   signature: str = ""
   summary: str = ""
   front_matter: str = ""
   sections: dict[str, str] = {}
 
 
-def _parse_numpydoc(docstr: Optional[str]) -> ParsedDoc:
+def _parse_numpydoc(docstr: str | None) -> ParsedDoc:
   """Parse a standard numpy-style docstring.
 
   Args:
@@ -111,58 +112,60 @@ def _parse_parameters(body: str) -> dict[str, str]:
 
 
 def _parse_extra_params(extra_params: str) -> dict[str, str]:
-  """Parse the extra parameters passed to _wraps()"""
+  """Parse the extra parameters passed to implements()"""
   parameters = _parameter_break.split(extra_params.strip('\n'))
   return {p.partition(' : ')[0].partition(', ')[0]: p for p in parameters}
 
 
-def _wraps(
-    fun: Optional[Callable[..., Any]],
+def implements(
+    original_fun: Callable[..., Any] | None,
     update_doc: bool = True,
     lax_description: str = "",
     sections: Sequence[str] = ('Parameters', 'Returns', 'References'),
     skip_params: Sequence[str] = (),
-    extra_params: Optional[str] = None,
-    module: Optional[str] = None,
+    extra_params: str | None = None,
+    module: str | None = None,
 ) -> Callable[[_T], _T]:
-  """Specialized version of functools.wraps for wrapping numpy functions.
+  """Decorator for JAX functions which implement a specified NumPy function.
 
-  This produces a wrapped function with a modified docstring. In particular, if
-  `update_doc` is True, parameters listed in the wrapped function that are not
-  supported by the decorated function will be removed from the docstring. For
-  this reason, it is important that parameter names match those in the original
-  numpy function.
+  This mainly contains logic to copy and modify the docstring of the original
+  function. In particular, if `update_doc` is True, parameters listed in the
+  original function that are not supported by the decorated function will
+  be removed from the docstring. For this reason, it is important that parameter
+  names match those in the original numpy function.
 
   Args:
-    fun: The function being wrapped
+    original_fun: The original function being implemented
     update_doc: whether to transform the numpy docstring to remove references of
       parameters that are supported by the numpy version but not the JAX version.
       If False, include the numpy docstring verbatim.
     lax_description: a string description that will be added to the beginning of
       the docstring.
     sections: a list of sections to include in the docstring. The default is
-      ["Parameters", "returns", "References"]
+      ["Parameters", "Returns", "References"]
     skip_params: a list of strings containing names of parameters accepted by the
       function that should be skipped in the parameter list.
     extra_params: an optional string containing additional parameter descriptions.
       When ``update_doc=True``, these will be added to the list of parameter
       descriptions in the updated doc.
-    module: an optional string specifying the module from which the wrapped function
+    module: an optional string specifying the module from which the original function
       is imported. This is useful for objects such as ufuncs, where the module cannot
-      be determined from the wrapped function itself.
+      be determined from the original function itself.
   """
-  def wrap(op):
-    op.__np_wrapped__ = fun
-    # Allows this pattern: @wraps(getattr(np, 'new_function', None))
-    if fun is None:
-      return op
-    docstr = getattr(fun, "__doc__", None)
-    name = getattr(fun, "__name__", getattr(op, "__name__", str(op)))
+  def decorator(wrapped_fun):
+    wrapped_fun.__np_wrapped__ = original_fun
+    # Allows this pattern: @implements(getattr(np, 'new_function', None))
+    if original_fun is None:
+      if lax_description:
+        wrapped_fun.__doc__ = lax_description
+      return wrapped_fun
+    docstr = getattr(original_fun, "__doc__", None)
+    name = getattr(original_fun, "__name__", getattr(wrapped_fun, "__name__", str(wrapped_fun)))
     try:
-      mod = module or fun.__module__
+      mod = module or original_fun.__module__
     except AttributeError:
-      if config.jax_enable_checks:
-        raise ValueError(f"function {fun} defines no __module__; pass module keyword to _wraps.")
+      if config.enable_checks.value:
+        raise ValueError(f"function {original_fun} defines no __module__; pass module keyword to implements().")
     else:
       name = f"{mod}.{name}"
     if docstr:
@@ -170,7 +173,7 @@ def _wraps(
         parsed = _parse_numpydoc(docstr)
 
         if update_doc and 'Parameters' in parsed.sections:
-          code = getattr(getattr(op, "__wrapped__", op), "__code__", None)
+          code = getattr(getattr(wrapped_fun, "__wrapped__", wrapped_fun), "__code__", None)
           # Remove unrecognized parameter descriptions.
           parameters = _parse_parameters(parsed.sections['Parameters'])
           if extra_params:
@@ -206,20 +209,20 @@ def _wraps(
         if kept_sections:
           docstr += "\n" + "\n\n".join(kept_sections) + "\n"
       except:
-        if config.jax_enable_checks:
+        if config.enable_checks.value:
           raise
-        docstr = fun.__doc__
+        docstr = original_fun.__doc__
 
-    op.__doc__ = docstr
+    wrapped_fun.__doc__ = docstr
     for attr in ['__name__', '__qualname__']:
       try:
-        value = getattr(fun, attr)
+        value = getattr(original_fun, attr)
       except AttributeError:
         pass
       else:
-        setattr(op, attr, value)
-    return op
-  return wrap
+        setattr(wrapped_fun, attr, value)
+    return wrapped_fun
+  return decorator
 
 _dtype = partial(dtypes.dtype, canonicalize=True)
 
@@ -229,7 +232,7 @@ def promote_shapes(fun_name: str, *args: ArrayLike) -> list[Array]:
     return [lax.asarray(arg) for arg in args]
   else:
     shapes = [np.shape(arg) for arg in args]
-    if config.jax_dynamic_shapes:
+    if config.dynamic_shapes.value:
       # With dynamic shapes we don't support singleton-dimension broadcasting;
       # we instead broadcast out to the full shape as a temporary workaround.
       # TODO(mattjj): revise this workaround
@@ -242,21 +245,20 @@ def promote_shapes(fun_name: str, *args: ArrayLike) -> list[Array]:
       if len(nonscalar_ranks) < 2:
         return [lax.asarray(arg) for arg in args]  # rely on lax scalar promotion
       else:
-        if config.jax_numpy_rank_promotion != "allow":
+        if config.numpy_rank_promotion.value != "allow":
           _rank_promotion_warning_or_error(fun_name, shapes)
         result_rank = len(lax.broadcast_shapes(*shapes))
-        return [_broadcast_to(arg, (1,) * (result_rank - len(shp)) + shp)
-                for arg, shp in zip(args, shapes)]
+        return [lax.broadcast_to_rank(arg, result_rank) for arg in args]
 
 
 def _rank_promotion_warning_or_error(fun_name: str, shapes: Sequence[Shape]):
-  if config.jax_numpy_rank_promotion == "warn":
+  if config.numpy_rank_promotion.value == "warn":
     msg = ("Following NumPy automatic rank promotion for {} on shapes {}. "
            "Set the jax_numpy_rank_promotion config option to 'allow' to "
            "disable this warning; for more information, see "
            "https://jax.readthedocs.io/en/latest/rank_promotion_warning.html.")
     warnings.warn(msg.format(fun_name, ' '.join(map(str, shapes))))
-  elif config.jax_numpy_rank_promotion == "raise":
+  elif config.numpy_rank_promotion.value == "raise":
     msg = ("Operands could not be broadcast together for {} on shapes {} "
            "and with the config option jax_numpy_rank_promotion='raise'. "
            "For more information, see "
@@ -326,7 +328,7 @@ def check_arraylike(fun_name: str, *args: Any, emit_warning=False, stacklevel=3)
                     if not _arraylike(arg))
     msg = f"{fun_name} requires ndarray or scalar arguments, got {type(arg)} at position {pos}."
     if emit_warning:
-      warnings.warn(msg + "In a future JAX release this will be an error.",
+      warnings.warn(msg + " In a future JAX release this will be an error.",
                     category=DeprecationWarning, stacklevel=stacklevel)
     else:
       raise TypeError(msg.format(fun_name, type(arg), pos))
@@ -355,16 +357,33 @@ def check_no_float0s(fun_name: str, *args: Any):
 _check_no_float0s = check_no_float0s
 
 
+def check_for_prngkeys(fun_name: str, *args: Any):
+  """Check if args don't match and none of the args have typed prng dtype"""
+  arg_dtypes = [dtypes.dtype(arg) for arg in args]
+  if len(set(arg_dtypes)) < 2:
+    return  # Will be caught by extended dtype impl rules.
+  if any(dtypes.issubdtype(dt, dtypes.prng_key) for dt in arg_dtypes):
+    if len(arg_dtypes) == 1:
+      raise TypeError(
+        f"{fun_name} does not accept dtype {str(arg_dtypes[0])}.")
+    else:
+      raise TypeError(
+        f"{fun_name} does not accept dtypes {', '.join(map(str, arg_dtypes))}."
+      )
+
+
 def promote_args(fun_name: str, *args: ArrayLike) -> list[Array]:
   """Convenience function to apply Numpy argument shape and dtype promotion."""
   check_arraylike(fun_name, *args)
   _check_no_float0s(fun_name, *args)
+  check_for_prngkeys(fun_name, *args)
   return promote_shapes(fun_name, *promote_dtypes(*args))
 
 
 def promote_args_numeric(fun_name: str, *args: ArrayLike) -> list[Array]:
   check_arraylike(fun_name, *args)
   _check_no_float0s(fun_name, *args)
+  check_for_prngkeys(fun_name, *args)
   return promote_shapes(fun_name, *promote_dtypes_numeric(*args))
 
 
@@ -374,6 +393,7 @@ def promote_args_inexact(fun_name: str, *args: ArrayLike) -> list[Array]:
   Promotes non-inexact types to an inexact type."""
   check_arraylike(fun_name, *args)
   _check_no_float0s(fun_name, *args)
+  check_for_prngkeys(fun_name, *args)
   return promote_shapes(fun_name, *promote_dtypes_inexact(*args))
 
 
@@ -387,15 +407,18 @@ def _broadcast_arrays(*args: ArrayLike) -> list[Array]:
   return [_broadcast_to(arg, result_shape) for arg in args]
 
 
-def _broadcast_to(arr: ArrayLike, shape: Shape) -> Array:
+def _broadcast_to(arr: ArrayLike, shape: DimSize | Shape) -> Array:
   check_arraylike("broadcast_to", arr)
   arr = arr if isinstance(arr, Array) else lax.asarray(arr)
   if not isinstance(shape, tuple) and np.ndim(shape) == 0:
     shape = (shape,)
-  shape = core.canonicalize_shape(shape)  # check that shape is concrete
+  # check that shape is concrete
+  shape = core.canonicalize_shape(shape)  # type: ignore[arg-type]
   arr_shape = np.shape(arr)
   if core.definitely_equal_shape(arr_shape, shape):
     return arr
+  elif len(shape) < len(arr_shape):
+    raise ValueError(f"Cannot broadcast to shape with fewer dimensions: {arr_shape=} {shape=}")
   else:
     nlead = len(shape) - len(arr_shape)
     shape_tail = shape[nlead:]
@@ -404,11 +427,7 @@ def _broadcast_to(arr: ArrayLike, shape: Shape) -> Array:
     if nlead < 0 or not compatible:
       msg = "Incompatible shapes for broadcasting: {} and requested shape {}"
       raise ValueError(msg.format(arr_shape, shape))
-    diff, = np.where(tuple(not core.definitely_equal(arr_d, shape_d)
-                           for arr_d, shape_d in safe_zip(arr_shape, shape_tail)))
-    new_dims = tuple(range(nlead)) + tuple(nlead + diff)
-    kept_dims = tuple(np.delete(np.arange(len(shape)), new_dims))
-    return lax.broadcast_in_dim(lax.squeeze(arr, tuple(diff)), shape, kept_dims)
+    return lax.broadcast_in_dim(arr, shape, tuple(range(nlead, len(shape))))
 
 
 # The `jit` on `where` exists to avoid materializing constants in cases like
@@ -423,9 +442,13 @@ def _where(condition: ArrayLike, x: ArrayLike, y: ArrayLike) -> Array:
   if not np.issubdtype(_dtype(condition), np.bool_):
     condition = lax.ne(condition, lax._zero(condition))
   x, y = promote_dtypes(x, y)
-  condition_arr, x_arr, y_arr = _broadcast_arrays(condition, x, y)
+  if np.ndim(condition) == 0:
+    # lax.select() handles scalar conditions without broadcasting.
+    x_arr, y_arr = _broadcast_arrays(x, y)
+  else:
+    condition, x_arr, y_arr = _broadcast_arrays(condition, x, y)
   try:
     is_always_empty = core.is_empty_shape(x_arr.shape)
   except:
     is_always_empty = False  # can fail with dynamic shapes
-  return lax.select(condition_arr, x_arr, y_arr) if not is_always_empty else x_arr
+  return lax.select(condition, x_arr, y_arr) if not is_always_empty else x_arr
