@@ -30,17 +30,22 @@ executable protocols described above.
 """
 from __future__ import annotations
 
+import functools
 from collections.abc import Sequence
 from dataclasses import dataclass
-from typing import Any, NamedTuple, Protocol, Union
+from typing import Any, NamedTuple, Protocol, Union, runtime_checkable
+import warnings
 
 import jax
 
 from jax._src import core
+from jax._src import config
 from jax._src import source_info_util
 from jax._src import traceback_util
 from jax._src import tree_util
 from jax._src import util
+from jax._src.sharding_impls import is_unspecified_or_auto
+from jax._src.layout import Layout
 from jax._src.interpreters import mlir
 from jax._src.lib.mlir import ir
 from jax._src.lib import xla_client as xc
@@ -67,7 +72,7 @@ class Executable(Protocol):
     # TODO(frostig): improve annotation (sequences of arrays/buffers)
     raise NotImplementedError
 
-  def input_shardings(self) -> Sequence[jax.sharding.XLACompatibleSharding]:
+  def input_shardings(self) -> Sequence[jax.sharding.Sharding]:
     """Flat sequence of input shardings.
 
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
@@ -75,7 +80,7 @@ class Executable(Protocol):
     """
     raise NotImplementedError
 
-  def output_shardings(self) -> Sequence[jax.sharding.XLACompatibleSharding]:
+  def output_shardings(self) -> Sequence[jax.sharding.Sharding]:
     """Flat sequence of output shardings.
 
     May raise ``NotImplementedError`` if unavailable, e.g. based on backend,
@@ -83,6 +88,12 @@ class Executable(Protocol):
     """
     raise NotImplementedError
 
+  def input_layouts(self):
+    raise NotImplementedError
+
+  def output_layouts(self):
+    raise NotImplementedError
+
   def as_text(self) -> str:
     """A human-readable text representation of this executable.
 
@@ -207,14 +218,22 @@ class XlaExecutable(Executable):
   def call(self, *args_flat) -> Sequence[Any]:
     raise NotImplementedError("must override")
 
-  def input_shardings(self) -> Sequence[jax.sharding.XLACompatibleSharding]:
+  def input_shardings(self) -> Sequence[jax.sharding.Sharding]:
     raise NotImplementedError(
         "compiled executable carries no input sharding information")
 
-  def output_shardings(self) -> Sequence[jax.sharding.XLACompatibleSharding]:
+  def output_shardings(self) -> Sequence[jax.sharding.Sharding]:
     raise NotImplementedError(
         "compiled executable carries no output sharding information")
 
+  def input_layouts(self):
+    raise NotImplementedError(
+        "compiled executable carries no input layout information")
+
+  def output_layouts(self):
+    raise NotImplementedError(
+        "compiled executable carries no input layout information")
+
   def as_text(self) -> str:
     xla_ext_exe = self.xla_extension_executable()
     err_msg = ("text view unsupported on current XLA backend: "
@@ -253,7 +272,10 @@ class XlaExecutable(Executable):
         ]
       except xla_extension.XlaRuntimeError as e:
         msg, *_ = e.args
-        if not (type(msg) is str and msg.startswith("UNIMPLEMENTED")):
+        supported = not (type(msg) is str and
+                         (msg.startswith("UNIMPLEMENTED") or
+                          "PjRt-compatible backend only" in msg))
+        if supported:
           raise
 
     if (
@@ -294,12 +316,19 @@ class XlaLowering(Lowering):
 
   def hlo(self) -> xc.XlaComputation:
     """Return an HLO representation of this computation."""
+    hlo = self.stablehlo()
+    m: str | bytes
+    m = mlir.module_to_bytecode(hlo)
     return xla_extension.mlir.mlir_module_to_xla_computation(
-        mlir.module_to_string(self.stablehlo()),
-        use_tuple_args=self.compile_args["tuple_args"])
+        m, use_tuple_args=self.compile_args["tuple_args"])
 
   def mhlo(self) -> ir.Module:
     """Return an MHLO representation of this computation."""
+    warnings.warn(
+        "mhlo support is deprecated and will be removed "
+        "from a future release of JAX. Use stablehlo instead.",
+        DeprecationWarning,
+    )
     module_str = xla_extension.mlir.stablehlo_to_mhlo(
         mlir.module_to_bytecode(self.stablehlo()))
     with self.stablehlo().context:
@@ -343,11 +372,26 @@ class XlaLowering(Lowering):
 
 # -- Public-facing API, plus helpers
 
-@dataclass
+@dataclass(frozen=True)
 class ArgInfo:
-  aval: core.AbstractValue
+  _aval: core.AbstractValue
   donated: bool
 
+  @property
+  def shape(self):
+    return self._aval.shape  # pytype: disable=attribute-error
+
+  @property
+  def dtype(self):
+    return self._aval.dtype  # pytype: disable=attribute-error
+
+
+@dataclass(frozen=True)
+class OutInfo:
+  shape: tuple[int, ...]
+  dtype: jax.typing.DTypeLike
+  sharding: jax.sharding.Sharding | None = None
+
 
 class Stage:
   args_info: Any  # PyTree of ArgInfo
@@ -360,7 +404,7 @@ class Stage:
   @property
   def in_avals(self):
     """Tree of input avals."""
-    return tree_util.tree_map(lambda x: x.aval, self.args_info)
+    return tree_util.tree_map(lambda x: x._aval, self.args_info)
 
   @property
   def donate_argnums(self):
@@ -377,6 +421,7 @@ def make_args_info(in_tree, in_avals, donate_argnums):
       ArgInfo(aval, i in donate_argnums)
       for i, aval in enumerate(flat_avals)])
 
+
 class CompiledCallParams(NamedTuple):
   executable: Executable
   no_kwargs: bool
@@ -471,15 +516,35 @@ class Compiled(Stage):
     return self._executable.runtime_executable()
 
   @property
-  def input_shardings(self):  # PyTree[sharding.XLACompatibleSharding]
+  def input_shardings(self):  # PyTree[sharding.Sharding]
     shardings_flat = self._executable.input_shardings()
+    # Some input shardings got DCE'd
+    if self.in_tree.num_leaves > len(shardings_flat):
+      iter_shardings_flat = iter(shardings_flat)
+      shardings_flat = [next(iter_shardings_flat) if i in self._executable._kept_var_idx
+                        else None for i in range(self.in_tree.num_leaves)]
     return tree_util.tree_unflatten(self.in_tree, shardings_flat)  # pytype: disable=attribute-error
 
   @property
-  def output_shardings(self):  # PyTree[sharding.XLACompatibleSharding]
+  def output_shardings(self):  # PyTree[sharding.Sharding]
     shardings_flat = self._executable.output_shardings()
     return tree_util.tree_unflatten(self.out_tree, shardings_flat)  # pytype: disable=attribute-error
 
+  def input_layouts(self):
+    layouts_flat = self._executable.input_layouts()
+    assert all(isinstance(l, Layout) for l in layouts_flat)
+    # Some input layouts got DCE'd
+    if self.in_tree.num_leaves > len(layouts_flat):
+      iter_layouts_flat = iter(layouts_flat)
+      layouts_flat = [next(iter_layouts_flat) if i in self._executable._kept_var_idx
+                      else Layout() for i in range(self.in_tree.num_leaves)]
+    return tree_util.tree_unflatten(self.in_tree, layouts_flat)  # pytype: disable=attribute-error
+
+  def output_layouts(self):
+    layouts_flat = self._executable.output_layouts()
+    assert all(isinstance(l, Layout) for l in layouts_flat)
+    return tree_util.tree_unflatten(self.out_tree, layouts_flat)  # pytype: disable=attribute-error
+
   @staticmethod
   def call(*args, **kwargs):
     # This is because `__call__` passes in `self._params` as the first argument.
@@ -488,7 +553,7 @@ class Compiled(Stage):
     # which might conflict here.
     params = args[0]
     args = args[1:]
-    if jax.config.jax_dynamic_shapes:
+    if config.dynamic_shapes.value:
       raise NotImplementedError
     if params.no_kwargs and kwargs:
       kws = ', '.join(kwargs.keys())
@@ -497,10 +562,18 @@ class Compiled(Stage):
           f"keyword arguments, but called with keyword arguments: {kws}")
     args_flat, in_tree = tree_util.tree_flatten((args, kwargs))
     if in_tree != params.in_tree:
-      # TODO(frostig): provide more info about the source function
-      # and transformation
-      raise TypeError(
-          f"function compiled for {params.in_tree}, called with {in_tree}")
+      errs = list(tree_util.equality_errors_pytreedef(in_tree, params.in_tree))
+      msg = []
+      msg.append(
+          "Function compiled with input pytree does not match the input pytree"
+          f" it was called with. There are {len(errs)} mismatches, including:")
+      for path, thing1, thing2, explanation in errs:
+        fst, *rest = path
+        base = ['args', 'kwargs'][fst.idx]
+        msg.append(
+            f"    * at {base}{tree_util.keystr(tuple(rest))}, seen {thing2} but now"
+            f" given {thing1}, so {explanation}")
+      raise TypeError('\n'.join(msg))
     try:
       out_flat = params.executable.call(*args_flat)
     except TypeError as e:
@@ -545,11 +618,10 @@ class Lowered(Stage):
   querying properties of lowered computations across JAX's various
   lowering paths (:func:`~jax.jit`, :func:`~jax.pmap`, etc.).
   """
-  __slots__ = ["args_info", "out_tree", "_lowering", "_no_kwargs"]
-
+  __slots__ = ["_lowering", "args_info", "out_tree", "_no_kwargs"]
+  _lowering: XlaLowering
   args_info: Any                # PyTree of ArgInfo
   out_tree: tree_util.PyTreeDef
-  _lowering: XlaLowering
   _no_kwargs: bool
 
   def __init__(
@@ -558,10 +630,11 @@ class Lowered(Stage):
       args_info,  # PyTree of ArgInfo
       out_tree: tree_util.PyTreeDef,
       no_kwargs: bool = False):
+
     self._lowering = lowering
-    self._no_kwargs = no_kwargs
     self.args_info = args_info
     self.out_tree = out_tree
+    self._no_kwargs = no_kwargs
 
   @classmethod
   def from_flat_info(cls,
@@ -586,6 +659,14 @@ class Lowered(Stage):
         out_tree,
         no_kwargs=no_kwargs)
 
+  @property
+  def out_info(self):  # PyTree of OutInfo
+    out_avals = self._lowering.compile_args["global_out_avals"]
+    out_shardings = self._lowering.compile_args["out_shardings"]
+    return self.out_tree.unflatten(
+        [OutInfo(o.shape, o.dtype, None if is_unspecified_or_auto(s) else s)
+         for o, s in zip(out_avals, out_shardings)])
+
   def compile(
       self, compiler_options: CompilerOptions | None = None) -> Compiled:
     """Compile, returning a corresponding ``Compiled`` instance."""
@@ -645,8 +726,40 @@ class Lowered(Stage):
       return None
 
 
+class Traced(Stage):
+  __slots__ = ["jaxpr", "args_info", "fun_name", "_out_tree", "_lower_callable",
+               "_args_flat", "_arg_names", "_num_consts"]
+
+  def __init__(self, jaxpr: core.ClosedJaxpr, args_info, fun_name, out_tree,
+               lower_callable, args_flat=None, arg_names=None,
+               num_consts: int = 0):
+    self.jaxpr = jaxpr
+    self.args_info = args_info
+    self.fun_name = fun_name
+    self._out_tree = out_tree
+    self._lower_callable = lower_callable
+    self._args_flat = args_flat
+    self._arg_names = arg_names
+    self._num_consts = num_consts
+
+  @property
+  def out_info(self):
+    return self._out_tree.unflatten(
+        [OutInfo(o.shape, o.dtype) for o in self.jaxpr.out_avals])
+
+  def lower(self, *, lowering_platforms: tuple[str, ...] | None = None,
+            _private_parameters: mlir.LoweringParameters | None = None):
+    if _private_parameters is None:
+      _private_parameters = mlir.LoweringParameters()
+    new_callable = functools.partial(
+        self._lower_callable, lowering_platforms=lowering_platforms,
+        lowering_parameters=_private_parameters)
+    return Lowered(new_callable(), self.args_info, self._out_tree)
+
+
+@runtime_checkable
 class Wrapped(Protocol):
-  """A function ready to be specialized, lowered, and compiled.
+  """A function ready to be traced, lowered, and compiled.
 
   This protocol reflects the output of functions such as
   ``jax.jit``. Calling it results in JIT (just-in-time) lowering,
@@ -658,6 +771,17 @@ class Wrapped(Protocol):
     """Executes the wrapped function, lowering and compiling as needed."""
     raise NotImplementedError
 
+  def trace(self, *args, **kwargs) -> Traced:
+    """Trace this function explicitly for the given arguments.
+
+    A traced function is staged out of Python and translated to a jaxpr. It is
+    ready for lowering but not yet lowered.
+
+    Returns:
+      A ``Traced`` instance representing the tracing.
+    """
+    raise NotImplementedError
+
   def lower(self, *args, **kwargs) -> Lowered:
     """Lower this function explicitly for the given arguments.
 
