@@ -12,16 +12,18 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 """Module for state types."""
+
 from __future__ import annotations
 
 from collections.abc import Sequence
 import dataclasses
 import math
-from typing import Any, Generic, TypeVar, Union
+from typing import Any, Union
 
 from jax._src import core
 from jax._src import effects
 from jax._src import pretty_printer as pp
+from jax._src.state import indexing
 from jax._src.util import safe_map, safe_zip
 from jax._src.typing import Array
 
@@ -72,34 +74,75 @@ StateEffect = Union[ReadEffect, WriteEffect, AccumEffect]
 
 # ## `Ref`s
 
-Aval = TypeVar("Aval", bound=core.AbstractValue)
-
 @dataclasses.dataclass
 class RefIndexer:
-  ref: Any
+  ref_or_view: Any
 
   def __getitem__(self, slc):
     if not isinstance(slc, tuple):
       slc = (slc,)
-    return RefView(self.ref, slc)
+    indexer = indexing.NDIndexer.from_indices_shape(slc, self.ref_or_view.shape)
+    if isinstance(self.ref_or_view, RefView):
+      view = self.ref_or_view
+      return RefView(view.ref, (*view.indexers, indexer))
+    return RefView(self.ref_or_view, (indexer,))
+
+Indexer = Any
 
 @dataclasses.dataclass
 class RefView:
   ref: Any
-  indexer: Any
+  indexers: tuple[indexing.NDIndexer, ...]
 
   @property
-  def at(self):
-    raise NotImplementedError("Can't call `.at` multiple times.")
+  def is_dynamic_size(self):
+    return self.indexers[-1].is_dynamic_size
+
+  @property
+  def shape(self) -> tuple[int | Array, ...]:
+    assert (
+        len(self.indexers) > 0
+    ), "Should not be able to create a trivial RefView"
+    return self.indexers[-1].get_indexer_shape()
+
+  @property
+  def dtype(self):
+    return self.ref.dtype
+
+  @property
+  def at(self) -> RefIndexer:
+    return RefIndexer(self)
+
+  def __getattr__(self, name):
+    return getattr(self.ref, name)
+
+  def __getitem__(self, slc):
+    from jax._src.state.primitives import ref_get  # pytype: disable=import-error
+    return ref_get(self, slc)
+
+  def __setitem__(self, slc, value):
+    from jax._src.state.primitives import ref_set # pytype: disable=import-error
+    return ref_set(self, slc, value)
 
 
 # We need an aval for `Ref`s so we can represent `get` and `swap` in Jaxprs.
-class AbstractRef(core.AbstractValue, Generic[Aval]):
+class AbstractRef(core.AbstractValue):
   __slots__ = ["inner_aval"]
 
   def __init__(self, inner_aval: core.AbstractValue):
     self.inner_aval = inner_aval
 
+  @property
+  def weak_type(self) -> bool:
+    if not hasattr(self.inner_aval, "weak_type"):
+      raise AttributeError
+    return self.inner_aval.weak_type
+
+  def update(self, inner_aval=None):
+    if inner_aval is None:
+      return AbstractRef(self.inner_aval)
+    return AbstractRef(inner_aval)
+
   def join(self, other):
     assert isinstance(other, AbstractRef)
     return AbstractRef(self.inner_aval.join(other.inner_aval))
@@ -110,13 +153,13 @@ class AbstractRef(core.AbstractValue, Generic[Aval]):
   @property
   def shape(self):
     if not isinstance(self.inner_aval, core.ShapedArray):
-      raise ValueError(f"`Ref{{{self.inner_aval.str_short()}}} has no `shape`.")
+      raise AttributeError(f"`Ref{{{self.inner_aval.str_short()}}} has no `shape`.")
     return self.inner_aval.shape
 
   @property
   def dtype(self):
     if not isinstance(self.inner_aval, core.UnshapedArray):
-      raise ValueError(f"`Ref{{{self.inner_aval.str_short()}}} has no `dtype`.")
+      raise AttributeError(f"`Ref{{{self.inner_aval.str_short()}}} has no `dtype`.")
     return self.inner_aval.dtype
 
   @core.aval_property
@@ -136,14 +179,10 @@ class AbstractRef(core.AbstractValue, Generic[Aval]):
     return ref_set(tracer, idx, value)
 
   def _getitem(self, tracer, idx) -> Array:
-    if not isinstance(idx, tuple):
-      idx = idx,
     from jax._src.state.primitives import ref_get  # pytype: disable=import-error
     return ref_get(tracer, idx)
 
   def _setitem(self, tracer, idx, value) -> None:
-    if not isinstance(idx, tuple):
-      idx = idx,
     from jax._src.state.primitives import ref_set  # pytype: disable=import-error
     return ref_set(tracer, idx, value)
 
@@ -179,8 +218,22 @@ def get_ref_state_effects(
            if isinstance(eff, (ReadEffect, WriteEffect, AccumEffect))
            and eff.input_index == i} for i, _ in enumerate(avals)]
 
-def shaped_array_ref(shape: tuple[int, ...], dtype,
-                     weak_type: bool = False,
-                     named_shape = None) -> AbstractRef[core.AbstractValue]:
-  return AbstractRef(core.ShapedArray(shape, dtype, weak_type=weak_type,
-                                      named_shape=named_shape))
+def shaped_array_ref(
+    shape: tuple[int, ...], dtype, weak_type: bool = False) -> AbstractRef:
+  return AbstractRef(core.ShapedArray(shape, dtype, weak_type=weak_type))
+
+def _shard_ref(mesh, names, ref_aval: AbstractRef):
+  del mesh
+  if names:
+    # Can't actually shard a ref, can only close over it.
+    raise NotImplementedError("Can't shard a Ref.")
+  return ref_aval
+core.shard_aval_handlers[AbstractRef] = _shard_ref
+
+def _unshard_ref(mesh, names, ref_aval: AbstractRef):
+  del mesh
+  if names:
+    # Can't actually shard a ref, can only close over it.
+    raise NotImplementedError("Can't unshard a Ref")
+  return ref_aval
+core.unshard_aval_handlers[AbstractRef] = _unshard_ref
