@@ -27,19 +27,24 @@ https://github.com/google/jax/pull/11859/.
 from __future__ import annotations
 
 from collections.abc import Sequence
+import enum
+import typing
 from typing import Any, Protocol, Union
-import numpy as np
 
 from jax._src.basearray import (
-    Array as Array,
     ArrayLike as ArrayLike,
+    Array as Array,
+    StaticScalar as StaticScalar,
 )
+import numpy as np
 
 DType = np.dtype
 
 # TODO(jakevdp, froystig): make ExtendedDType a protocol
 ExtendedDType = Any
 
+
+@typing.runtime_checkable
 class SupportsDType(Protocol):
   @property
   def dtype(self) -> DType: ...
@@ -47,14 +52,19 @@ class SupportsDType(Protocol):
 # DTypeLike is meant to annotate inputs to np.dtype that return
 # a valid JAX dtype. It's different than numpy.typing.DTypeLike
 # because JAX doesn't support objects or structured dtypes.
-# It does not include JAX dtype extensions such as KeyType and others.
-# For now, we use Any to allow scalar types like np.int32 & jnp.int32.
-# TODO(jakevdp) specify these more strictly.
-DTypeLike = Union[Any, str, np.dtype, SupportsDType]
+# Unlike np.typing.DTypeLike, we exclude None, and instead require
+# explicit annotations when None is acceptable.
+# TODO(jakevdp): consider whether to add ExtendedDtype to the union.
+DTypeLike = Union[
+  str,            # like 'float32', 'int32'
+  type[Any],      # like np.float32, np.int32, float, int
+  np.dtype,       # like np.dtype('float32'), np.dtype('int32')
+  SupportsDType,  # like jnp.float32, jnp.int32
+]
 
 # Shapes are tuples of dimension sizes, which are normally integers. We allow
 # modules to extend the set of dimension sizes to contain other types, e.g.,
-# symbolic dimensions in jax2tf.shape_poly.DimVar and masking.Poly.
+# symbolic dimensions in export.DimExpr.
 DimSize = Union[int, Any]  # extensible
 Shape = Sequence[DimSize]
 
@@ -72,3 +82,15 @@ class DuckTypedArray(Protocol):
 # JAX array (i.e. not including future non-standard array types like KeyArray and BInt).
 # It's different than np.typing.ArrayLike in that it doesn't accept arbitrary sequences,
 # nor does it accept string data.
+
+# We use a class for deprecated args to avoid using Any/object types which can
+# introduce complications and mistakes in static analysis
+class DeprecatedArg:
+  def __repr__(self):
+    return "Deprecated"
+
+# Mirror of dlpack.h enum
+class DLDeviceType(enum.IntEnum):
+  kDLCPU = 1
+  kDLCUDA = 2
+  kDLROCM = 10
