@@ -12,20 +12,23 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections.abc import Iterable, Iterator, Sequence
+from __future__ import annotations
+
+import abc
+from collections.abc import Callable, Iterable, Iterator, Sequence
 import functools
 from functools import partial
 import itertools as it
 import logging
 import operator
-from typing import (Any, Callable, Generic, Optional, TypeVar, overload, TYPE_CHECKING, cast)
+from typing import (Any, Generic, TypeVar, overload, TYPE_CHECKING, cast)
 import weakref
 
 import numpy as np
 
+from jax._src import config
 from jax._src.lib import xla_client as xc
 from jax._src.lib import utils as jaxlib_utils
-from jax._src.config import config
 
 logger = logging.getLogger(__name__)
 
@@ -128,6 +131,15 @@ def split_list(args: Sequence[T], ns: Sequence[int]) -> list[list[T]]:
   lists.append(args)
   return lists
 
+def split_list_checked(args: Sequence[T], ns: Sequence[int]) -> list[list[T]]:
+  args = list(args)
+  assert sum(ns) == len(args)
+  lists = []
+  for n in ns:
+    lists.append(args[:n])
+    args = args[n:]
+  return lists
+
 def partition_list(bs: Sequence[bool], l: Sequence[T]) -> tuple[list[T], list[T]]:
   assert len(bs) == len(l)
   lists = [], []  # type: ignore
@@ -135,14 +147,38 @@ def partition_list(bs: Sequence[bool], l: Sequence[T]) -> tuple[list[T], list[T]
     lists[b].append(x)
   return lists
 
-def merge_lists(bs: Sequence[bool], l0: Sequence[T], l1: Sequence[T]) -> list[T]:
+def merge_lists(bs: Sequence[bool],
+                l0: Sequence[T1],
+                l1: Sequence[T2]
+                ) -> list[T1 | T2]:
   assert sum(bs) == len(l1) and len(bs) - sum(bs) == len(l0)
   i0, i1 = iter(l0), iter(l1)
-  out = [next(i1) if b else next(i0) for b in bs]
+  out: list[T1 | T2] = [next(i1) if b else next(i0) for b in bs]
   sentinel = object()
   assert next(i0, sentinel) is next(i1, sentinel) is sentinel
   return out
 
+def subs_list(
+    subs: Sequence[int | None], src: Sequence[T], base: Sequence[T],
+) -> list[T]:
+  base_ = iter(base)
+  out = [src[i] if i is not None else next(base_) for i in subs]
+  sentinel = object()
+  assert next(base_, sentinel) is sentinel
+  return out
+
+def subs_list2(
+    subs1: Sequence[int | None], subs2: Sequence[int | None],
+    src1: Sequence[T], src2: Sequence[T], base: Sequence[T],
+) -> list[T]:
+  assert len(subs1) == len(subs2)
+  base_ = iter(base)
+  out = [src1[f1] if f1 is not None else src2[f2] if f2 is not None else
+         next(base_) for f1, f2, in zip(subs1, subs2)]
+  sentinel = object()
+  assert next(base_, sentinel) is sentinel
+  return out
+
 def split_dict(dct, names):
   dct = dict(dct)
   lst = [dct.pop(name) for name in names]
@@ -249,7 +285,11 @@ def split_merge(predicate, xs):
 
   return lhs, rhs, merge
 
-def cache(max_size=4096):
+
+def _ignore(): return None
+
+
+def cache(max_size=4096, trace_context_in_key=True):
   def wrap(f):
     @functools.lru_cache(max_size)
     def cached(_, *args, **kwargs):
@@ -257,19 +297,28 @@ def cache(max_size=4096):
 
     @functools.wraps(f)
     def wrapper(*args, **kwargs):
-      if config.jax_check_tracer_leaks:
+      if config.check_tracer_leaks.value:
         return f(*args, **kwargs)
-      else:
-        return cached(config._trace_context(), *args, **kwargs)
+      return cached(config.trace_context() if trace_context_in_key else _ignore(),
+                    *args, **kwargs)
 
     wrapper.cache_clear = cached.cache_clear
     wrapper.cache_info = cached.cache_info
+    cache_clearing_funs.add(wrapper.cache_clear)
     return wrapper
   return wrap
 
+cache_clearing_funs = weakref.WeakSet()  # type: ignore
+
+def clear_all_caches():
+  global cache_clearing_funs
+  for clear in cache_clearing_funs:
+    clear()
+
 memoize = cache(max_size=None)
 
-def weakref_lru_cache(call: Callable, maxsize=2048):
+def weakref_lru_cache(call: Callable, maxsize=2048,
+                      trace_context_in_key: bool = True):
   """
   Least recently used cache decorator with weakref support.
 
@@ -278,7 +327,9 @@ def weakref_lru_cache(call: Callable, maxsize=2048):
   behave similar to `functools.lru_cache`.
   """
   global _weakref_lru_caches
-  cached_call = xc.weakref_lru_cache(config._trace_context, call, maxsize)
+  cached_call = xc.weakref_lru_cache(
+      config.trace_context if trace_context_in_key else _ignore,
+      call, maxsize)
   _weakref_lru_caches.add(cached_call)
   return cached_call
 
@@ -324,6 +375,13 @@ class WrapKwArgs:
 def wrap_name(name, transform_name):
   return transform_name + '(' + name + ')'
 
+def fun_name(fun: Callable):
+  return getattr(fun, "__name__", "<unnamed function>")
+
+def fun_qual_name(fun: Callable):
+  return getattr(fun, "__qualname__",
+                 getattr(fun, "__name__", "<unnamed function>"))
+
 def canonicalize_axis(axis, num_dims) -> int:
   """Canonicalize an axis in [-num_dims, num_dims) to [0, num_dims)."""
   axis = operator.index(axis)
@@ -353,8 +411,8 @@ def ceil_of_ratio(x, y):
 
 def wraps(
     wrapped: Callable,
-    namestr: Optional[str] = None,
-    docstr: Optional[str] = None,
+    namestr: str | None = None,
+    docstr: str | None = None,
     **kwargs,
 ) -> Callable[[T], T]:
   """
@@ -363,7 +421,7 @@ def wraps(
   """
   def wrapper(fun: T) -> T:
     try:
-      name = getattr(wrapped, "__name__", "<unnamed function>")
+      name = fun_name(wrapped)
       doc = getattr(wrapped, "__doc__", "") or ""
       fun.__dict__.update(getattr(wrapped, "__dict__", {}))
       fun.__annotations__ = getattr(wrapped, "__annotations__", {})
@@ -391,6 +449,10 @@ def tuple_delete(t, idx):
   assert 0 <= idx < len(t), (idx, len(t))
   return t[:idx] + t[idx + 1:]
 
+def tuple_update(t, idx, val):
+  assert 0 <= idx < len(t), (idx, len(t))
+  return t[:idx] + (val,) + t[idx+1:]
+
 class HashableFunction:
   """Decouples function equality and hash from its identity.
 
@@ -444,7 +506,13 @@ class HashablePartial:
             self.args == other.args and self.kwargs == other.kwargs)
 
   def __hash__(self):
-    return hash((self.f.__code__, self.args, tuple(self.kwargs.items())))
+    return hash(
+      (
+        self.f.__code__,
+        self.args,
+        tuple(sorted(self.kwargs.items(), key=lambda kv: kv[0])),
+      ),
+    )
 
   def __call__(self, *args, **kwargs):
     return self.f(*self.args, *args, **self.kwargs, **kwargs)
@@ -464,7 +532,7 @@ def distributed_debug_log(*pairs):
     pairs: A sequence of label/value pairs to log. The first pair is treated as
     a heading for subsequent pairs.
   """
-  if config.jax_distributed_debug:
+  if config.distributed_debug.value:
     lines = ["\nDISTRIBUTED_DEBUG_BEGIN"]
     try:
       lines.append(f"{pairs[0][0]}: {pairs[0][1]}")
@@ -477,6 +545,14 @@ def distributed_debug_log(*pairs):
     logger.warning("\n".join(lines))
 
 
+def stable_unique(it: Iterable[T]) -> Iterable[T]:
+  """Returns unique elements from `it` in the order of occurrence.
+
+  The elements must be hashable.
+  """
+  return dict.fromkeys(it).keys()
+
+
 class OrderedSet(Generic[T]):
   elts_set: set[T]
   elts_list: list[T]
@@ -506,7 +582,7 @@ class OrderedSet(Generic[T]):
 
 class HashableWrapper:
   x: Any
-  hash: Optional[int]
+  hash: int | None
   def __init__(self, x):
     self.x = x
     try: self.hash = hash(x)
@@ -592,3 +668,22 @@ try:
 except AttributeError:
   # legacy numpy
   NumpyComplexWarning = np.ComplexWarning
+
+
+class StrictABCMeta(abc.ABCMeta):
+  """A variant of `abc.ABCMeta` which does not allow virtual subclasses.
+
+  Virtual subclasses support require `abc.ABCMeta` to roundtrip through
+  pure Python when doing instance/subclass checking. This if fine for ABCs
+  which need virtual subclasses, but is wasteful for the ones which don't.
+  """
+  def register(cls, subclass):
+    del subclass  # Unused.
+    raise NotImplementedError(f"{cls} does not support virtual subclasses")
+
+  __instancecheck__ = type.__instancecheck__  # type: ignore[assignment]
+  __subclasscheck__ = type.__subclasscheck__  # type: ignore[assignment]
+
+
+class StrictABC(metaclass=StrictABCMeta):
+  __slots__ = ()
