@@ -15,8 +15,6 @@
 # Note: import <name> as <name> is required for names to be exported.
 # See PEP 484 & https://github.com/google/jax/issues/7570
 
-from __future__ import annotations
-
 from jax._src.core import (
   AbstractToken as AbstractToken,
   AbstractValue as AbstractValue,
@@ -27,7 +25,6 @@ from jax._src.core import (
   ConcreteArray as ConcreteArray,
   ConcretizationTypeError as ConcretizationTypeError,
   DShapedArray as DShapedArray,
-  DimSize as DimSize,
   DropVar as DropVar,
   Effect as Effect,
   Effects as Effects,
@@ -45,12 +42,10 @@ from jax._src.core import (
   MainTrace as MainTrace,
   MapPrimitive as MapPrimitive,
   NameGatheringSubst as NameGatheringSubst,
-  NamedShape as NamedShape,
   OutDBIdx as OutDBIdx,
   OutputType as OutputType,
   ParamDict as ParamDict,
   Primitive as Primitive,
-  Shape as Shape,
   ShapedArray as ShapedArray,
   Sublevel as Sublevel,
   TRACER_LEAK_DEBUGGER_WARNING as TRACER_LEAK_DEBUGGER_WARNING,
@@ -60,39 +55,29 @@ from jax._src.core import (
   TraceStack as TraceStack,
   TraceState as TraceState,
   Tracer as Tracer,
-  TracerArrayConversionError as TracerArrayConversionError,
-  TracerIntegerConversionError as TracerIntegerConversionError,
-  UnexpectedTracerError as UnexpectedTracerError,
   UnshapedArray as UnshapedArray,
   Value as Value,
   Var as Var,
   abstract_token as abstract_token,
   apply_todos as apply_todos,
-  as_hashable_function as as_hashable_function,
-  as_named_shape as as_named_shape,
   aval_mapping_handlers as aval_mapping_handlers,
   axis_frame as axis_frame,
   call as call,
   call_bind_with_continuation as call_bind_with_continuation,
   call_impl as call_impl,
   call_p as call_p,
-  canonicalize_shape as canonicalize_shape,  # TODO(necula): remove this API
   check_eqn as check_eqn,
   check_jaxpr as check_jaxpr,
   check_type as check_type,
   check_valid_jaxtype as check_valid_jaxtype,
   closed_call_p as closed_call_p,
-  collections as collections,
   concrete_aval as concrete_aval,
   concrete_or_error as concrete_or_error,
   concretization_function_error as concretization_function_error,
   cur_sublevel as cur_sublevel,
   custom_typechecks as custom_typechecks,
   dedup_referents as dedup_referents,
-  definitely_equal as definitely_equal,  # TODO(necula): remove this API
-  dimension_as_value as dimension_as_value,  # TODO(necula): remove this API
   do_subst_axis_names_jaxpr as do_subst_axis_names_jaxpr,
-  dtypes as dtypes,
   ensure_compile_time_eval as ensure_compile_time_eval,
   escaped_tracer_error as escaped_tracer_error,
   eval_context as eval_context,
@@ -110,38 +95,23 @@ from jax._src.core import (
   jaxpr_uses_outfeed as jaxpr_uses_outfeed,
   jaxprs_in_params as jaxprs_in_params,
   join_effects as join_effects,
-  join_named_shapes as join_named_shapes,
   lattice_join as lattice_join,
   leaked_tracer_error as leaked_tracer_error,
   literalable_types as literalable_types,
-  lu as lu,
-  map as map,
   map_bind as map_bind,
   map_bind_with_continuation as map_bind_with_continuation,
   mapped_aval as mapped_aval,
   maybe_find_leaked_tracers as maybe_find_leaked_tracers,
-  namedtuple as namedtuple,
+  max_dim as max_dim,
+  min_dim as min_dim,
   new_base_main as new_base_main,
   new_jaxpr_eqn as new_jaxpr_eqn,
   new_main as new_main,
   new_sublevel as new_sublevel,
   no_axis_name as no_axis_name,
   no_effects as no_effects,
+  non_negative_dim as _deprecated_non_negative_dim,
   outfeed_primitives as outfeed_primitives,
-  partial as partial,
-  pp as pp,
-  pp_aval as pp_aval,
-  pp_eqn as pp_eqn,
-  pp_eqn_rules as pp_eqn_rules,
-  pp_eqns as pp_eqns,
-  pp_jaxpr as pp_jaxpr,
-  pp_jaxpr_eqn_range as pp_jaxpr_eqn_range,
-  pp_jaxpr_skeleton as pp_jaxpr_skeleton,
-  pp_jaxprs as pp_jaxprs,
-  pp_kv_pair as pp_kv_pair,
-  pp_kv_pairs as pp_kv_pairs,
-  pp_var as pp_var,
-  pp_vars as pp_vars,
   primal_dtype_to_tangent_dtype as primal_dtype_to_tangent_dtype,
   primitive_uses_outfeed as primitive_uses_outfeed,
   process_env_traces_call as process_env_traces_call,
@@ -150,11 +120,7 @@ from jax._src.core import (
   raise_as_much_as_possible as raise_as_much_as_possible,
   raise_to_shaped as raise_to_shaped,
   raise_to_shaped_mappings as raise_to_shaped_mappings,
-  ref as ref,
   reset_trace_state as reset_trace_state,
-  safe_map as safe_map,
-  safe_zip as safe_zip,
-  source_info_util as source_info_util,
   stash_axis_env as stash_axis_env,
   str_eqn_compact as str_eqn_compact,
   subjaxprs as subjaxprs,
@@ -164,13 +130,8 @@ from jax._src.core import (
   subst_axis_names_var as subst_axis_names_var,
   substitute_vars_in_output_ty as substitute_vars_in_output_ty,
   thread_local_state as thread_local_state,
-  token as token,
-  total_ordering as total_ordering,
   trace_state_clean as trace_state_clean,
-  traceback_util as traceback_util,
   traverse_jaxpr_params as traverse_jaxpr_params,
-  tuple_delete as tuple_delete,
-  tuple_insert as tuple_insert,
   typecheck as typecheck,
   typecompat as typecompat,
   typematch as typematch,
@@ -178,7 +139,70 @@ from jax._src.core import (
   used_axis_names as used_axis_names,
   used_axis_names_jaxpr as used_axis_names_jaxpr,
   valid_jaxtype as valid_jaxtype,
-  zip as zip,
 )
 
-symbolic_equal_dim = definitely_equal  # TODO(necula): remove this API
+
+from jax._src import core as _src_core
+_deprecations = {
+    # Added 2024-06-12
+    "pp_aval": ("jax.core.pp_aval is deprecated.", _src_core.pp_aval),
+    "pp_eqn": ("jax.core.pp_eqn is deprecated.", _src_core.pp_eqn),
+    "pp_eqn_rules": ("jax.core.pp_eqn_rules is deprecated.", _src_core.pp_eqn_rules),
+    "pp_eqns": ("jax.core.pp_eqns is deprecated.", _src_core.pp_eqns),
+    "pp_jaxpr": ("jax.core.pp_jaxpr is deprecated.", _src_core.pp_jaxpr),
+    "pp_jaxpr_eqn_range": ("jax.core.pp_jaxpr_eqn_range is deprecated.", _src_core.pp_jaxpr_eqn_range),
+    "pp_jaxpr_skeleton": ("jax.core.pp_jaxpr_skeleton is deprecated.", _src_core.pp_jaxpr_skeleton),
+    "pp_jaxprs": ("jax.core.pp_jaxprs is deprecated.", _src_core.pp_jaxprs),
+    "pp_kv_pair": ("jax.core.pp_kv_pair is deprecated.", _src_core.pp_kv_pair),
+    "pp_kv_pairs": ("jax.core.pp_kv_pairs is deprecated.", _src_core.pp_kv_pairs),
+    "pp_var": ("jax.core.pp_var is deprecated.", _src_core.pp_var),
+    "pp_vars": ("jax.core.pp_vars is deprecated.", _src_core.pp_vars),
+    # Finalized 2024-05-13; remove after 2024-08-13
+    "DimSize": (
+        "jax.core.DimSize is deprecated. Use DimSize = int | Any.",
+        None,
+    ),
+    "Shape": (
+        "jax.core.Shape is deprecated. Use Shape = Sequence[int | Any].",
+        None,
+    ),
+    # Finalized 2024-06-24; remove after 2024-09-24
+    "canonicalize_shape": (
+      "jax.core.canonicalize_shape is deprecated.", None,
+    ),
+    "dimension_as_value": (
+      "jax.core.dimension_as_value is deprecated. Use jnp.array.", None,
+    ),
+    "definitely_equal": (
+      "jax.core.definitely_equal is deprecated. Use ==.", None,
+    ),
+    "symbolic_equal_dim": (
+      "jax.core.symbolic_equal_dim is deprecated. Use ==.", None,
+    ),
+    # Added Jan 8, 2024
+    "non_negative_dim": (
+      "jax.core.non_negative_dim is deprecated. Use max_dim(..., 0).", _deprecated_non_negative_dim,
+    ),
+}
+
+import typing
+if typing.TYPE_CHECKING:
+  non_negative_dim = _deprecated_non_negative_dim
+  pp_aval = _src_core.pp_aval
+  pp_eqn = _src_core.pp_eqn
+  pp_eqn_rules = _src_core.pp_eqn_rules
+  pp_eqns = _src_core.pp_eqns
+  pp_jaxpr = _src_core.pp_jaxpr
+  pp_jaxpr_eqn_range = _src_core.pp_jaxpr_eqn_range
+  pp_jaxpr_skeleton = _src_core.pp_jaxpr_skeleton
+  pp_jaxprs = _src_core.pp_jaxprs
+  pp_kv_pair = _src_core.pp_kv_pair
+  pp_kv_pairs = _src_core.pp_kv_pairs
+  pp_var = _src_core.pp_var
+  pp_vars = _src_core.pp_vars
+else:
+  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+  __getattr__ = _deprecation_getattr(__name__, _deprecations)
+  del _deprecation_getattr
+del typing
+del _src_core
