@@ -89,7 +89,10 @@ Example Usage:
 .. _Optax: https://github.com/deepmind/optax
 """
 
-from typing import Any, Callable, NamedTuple, Union
+from __future__ import annotations
+
+from collections.abc import Callable
+from typing import Any, NamedTuple
 
 from collections import namedtuple
 import functools
@@ -117,7 +120,7 @@ OptimizerState = namedtuple("OptimizerState",
 register_pytree_node(
     OptimizerState,
     lambda xs: ((xs.packed_state,), (xs.tree_def, xs.subtree_defs)),
-    lambda data, xs: OptimizerState(xs[0], data[0], data[1]))  # type: ignore[index]
+    lambda data, xs: OptimizerState(xs[0], data[0], data[1]))
 
 
 Array = Any
@@ -550,7 +553,7 @@ def piecewise_constant(boundaries: Any, values: Any):
     return values[jnp.sum(i > boundaries)]
   return schedule
 
-def make_schedule(scalar_or_schedule: Union[float, Schedule]) -> Schedule:
+def make_schedule(scalar_or_schedule: float | Schedule) -> Schedule:
   if callable(scalar_or_schedule):
     return scalar_or_schedule
   elif jnp.ndim(scalar_or_schedule) == 0:
