@@ -13,13 +13,15 @@
 # limitations under the License.
 """Workarounds for jax2tf transforms when XLA is not linked in."""
 
+from __future__ import annotations
+
 import builtins
-from collections.abc import Sequence
+from collections.abc import Callable, Sequence
 import dataclasses
 from functools import partial, wraps
 import math
 import string
-from typing import Any, Callable, Optional
+from typing import Any
 
 from jax._src import core
 from jax import lax
@@ -30,7 +32,7 @@ from jax._src import util
 from jax.experimental.jax2tf import jax2tf
 
 import numpy as np
-import tensorflow as tf  # type: ignore[import]
+import tensorflow as tf
 
 
 # Implementation rules for primitives when XLA is not linked in. These
@@ -250,8 +252,8 @@ def _conv_general_dilated(
     lhs, rhs, *, window_strides, padding, lhs_dilation, rhs_dilation,
     dimension_numbers: lax.ConvDimensionNumbers, feature_group_count: int,
     batch_group_count: int,
-    precision: Optional[tuple[PrecisionType, PrecisionType]],
-    preferred_element_type: Optional[DType],
+    precision: tuple[PrecisionType, PrecisionType] | None,
+    preferred_element_type: DType | None,
     _in_avals: Sequence[core.ShapedArray], _out_aval: core.ShapedArray):
   """Implementation of lax.conv_general_dilated_p using XlaConv."""
   # In presence of shape polymorphism, lhs.shape and rhs.shape may contain
@@ -272,8 +274,8 @@ def _conv_general_dilated(
   in_channels = lhs_shape[-1]
   *rhs_spatial_shapes, _, rhs_out_channel = rhs_shape
 
-  is_transpose = any([d != 1 for d in lhs_dilation])
-  is_atrous = any([d != 1 for d in rhs_dilation])
+  is_transpose = any(d != 1 for d in lhs_dilation)
+  is_atrous = any(d != 1 for d in rhs_dilation)
   is_depthwise = in_channels == feature_group_count and feature_group_count > 1
   _validate_conv_features(is_transpose, is_atrous, is_depthwise,
                           feature_group_count, batch_group_count,
@@ -360,8 +362,8 @@ tf_impl_no_xla[lax.conv_general_dilated_p] = _conv_general_dilated
 
 
 def _dot_general(lhs, rhs, *, dimension_numbers,
-                 precision: Optional[tuple[PrecisionType, PrecisionType]],
-                 preferred_element_type: Optional[DType],
+                 precision: tuple[PrecisionType, PrecisionType] | None,
+                 preferred_element_type: DType | None,
                  _in_avals: Sequence[core.ShapedArray],
                  _out_aval: core.ShapedArray):
   """Implementation of lax.dot_general_p in terms of tf.linalg.einsum."""
@@ -416,7 +418,7 @@ def _dot_general(lhs, rhs, *, dimension_numbers,
       squeeze_idxs.append(len(rhs.shape) - 1)
     result = tf.linalg.matmul(lhs, rhs)
     if len(squeeze_idxs) != 0:
-      assert all([result.shape[i] == 1 for i in squeeze_idxs])
+      assert all(result.shape[i] == 1 for i in squeeze_idxs)
       result = tf.squeeze(result, squeeze_idxs)
     return convert_result(result)
 
@@ -457,7 +459,7 @@ tf_impl_no_xla[lax.dot_general_p] = _dot_general
 def _interior_padding(operand, padding_value, padding_config, operand_shape):
   # Used only when enable_xla=False
   # Applies only the interior padding from the padding_config.
-  # We do this somewhat inefficiently, as as a scatter.
+  # We do this somewhat inefficiently, as a scatter.
   # For each dimension we compute the indices_by_dim as [0, f, 2f, 3f, ...] where
   # f is the dilation factor for the dimension, i.e., 1 + interior_padding.
   # Then we compute the cartesian production of the indices (using broadcast
@@ -545,7 +547,12 @@ def _validate_reduce_window_inputs(operand_shape, computation_name, dtype,
     # tf.math.reduce_min.
     raise _reduce_error(f"Min pool does not support operands of type {dtype}")
   if computation_name == "add" and dtype not in [
-      tf.float16, tf.float32, tf.float64
+      tf.bfloat16,
+      tf.float16,
+      tf.float32,
+      tf.float64,
+      tf.int16,
+      tf.int32,
   ]:
     raise _reduce_error("Add pooling does not support operands of type "
                         f"{dtype}")
@@ -651,6 +658,13 @@ def _reduce_monoid(operand, window_dimensions, window_strides, padding,
       raise NotImplementedError(
           f"TODO: use tf.nn.pool with dynamic shapes¨{window_dimensions=} "
           f" {window_strides=} {dilations=}")
+    # tf.nn.pool() currently does not suport tf.int32 and so we cast back and
+    # forth in order to be able to convert.
+    if (inputs.dtype in [tf.int16, tf.int32]) and computation_name == "add":
+      original_dtype = inputs.dtype
+      inputs = tf.cast(inputs, dtype=tf.float32)
+    else:
+      original_dtype = None
     result = tf.nn.pool(
         inputs,
         window_shape=window_dimensions,
@@ -658,6 +672,8 @@ def _reduce_monoid(operand, window_dimensions, window_strides, padding,
         padding=padding_type,
         strides=window_strides,
         dilations=dilations)
+    if original_dtype:
+      result = tf.cast(result, dtype=original_dtype)
 
     if has_only_spatial_dims:
       # If the input only had spatial dimensions we need to contract the batch
@@ -1090,7 +1106,7 @@ def _gather(operand, start_indices, *, dimension_numbers,
     try:
       return gather_fn(gather_args)
     except ValueError as e:
-      errors.append(f"{gather_fn}: {repr(e)}")
+      errors.append(f"{gather_fn}: {e!r}")
 
   error_msg = (f"Unsupported arguments for gather: {gather_args}, errors:\n" +
                "\n".join(errors))
