@@ -13,23 +13,25 @@
 # limitations under the License.
 """See primitives_test docstring for how the Jax2TfLimitations are used."""
 
-from collections.abc import Sequence
+from __future__ import annotations
+
+from collections.abc import Callable, Sequence
 import itertools
-from typing import Any, Callable, Optional, Union
+from typing import Any
 
 import jax
 from jax import lax
 from jax import numpy as jnp
-from jax import config
-from jax._src import test_util as jtu
+from jax._src import config
 from jax._src import dtypes
-from jax.experimental.jax2tf.tests import primitive_harness
+from jax._src import test_util as jtu
+from jax._src.internal_test_util import test_harnesses
 import numpy as np
 
 DType = Any
 
 
-class Jax2TfLimitation(primitive_harness.Limitation):
+class Jax2TfLimitation(test_harnesses.Limitation):
   """Specific primitive limitations for jax2tf.
 
   See the primitive_test module docstring for details.
@@ -43,7 +45,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       self,
       description: str,
       *,
-      devices: Union[str, Sequence[str]] = ("cpu", "gpu", "tpu"),
+      devices: str | Sequence[str] = ("cpu", "gpu", "tpu"),
       dtypes: Sequence[DType] = (),
       enabled: bool = True,
       # jax2tf specific
@@ -52,9 +54,9 @@ class Jax2TfLimitation(primitive_harness.Limitation):
       skip_tf_run=False,
       expect_tf_error: bool = True,
       skip_comparison=False,
-      custom_assert: Optional[Callable] = None,
+      custom_assert: Callable | None = None,
       tol=None):
-    """See the primitive_harness.Limitation common arguments.
+    """See the test_harnesses.Limitation common arguments.
 
     Args :
       modes: one of "eager", "graph", "compiled"
@@ -92,8 +94,8 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     self.skip_comparison = skip_comparison
 
   def get_max_tolerance_limitation(
-      self, limitations: Sequence["Jax2TfLimitation"]
-  ) -> Optional["Jax2TfLimitation"]:
+      self, limitations: Sequence[Jax2TfLimitation]
+  ) -> Jax2TfLimitation | None:
     """Pick the tolerance limitation that establishes the maximum tolerance."""
     # TODO: it would be best if the limitations with tolerance are mutually exclusive
     # and we don't have to compute the maximum
@@ -108,13 +110,13 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
   def filter(  # type: ignore[override]
       self,
-      dtype: Optional[DType] = None,
-      device: Optional[str] = None,
-      mode: Optional[str] = None) -> bool:
+      dtype: DType | None = None,
+      device: str | None = None,
+      mode: str | None = None) -> bool:
     """Checks if this limitation is enabled for dtype and device and mode."""
     native_serialization_mask = (
         Jax2TfLimitation.FOR_NATIVE
-        if config.jax2tf_default_native_serialization
+        if config.jax2tf_default_native_serialization.value
         else Jax2TfLimitation.FOR_NON_NATIVE)
     return ((mode is None or mode in self.modes) and
             (self.native_serialization & native_serialization_mask) and
@@ -122,7 +124,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
 
   @classmethod
   def limitations_for_harness(
-      cls, harness: primitive_harness.Harness) -> Sequence["Jax2TfLimitation"]:
+      cls, harness: test_harnesses.Harness) -> Sequence[Jax2TfLimitation]:
     group_method = getattr(cls, harness.group_name, None)
     if harness.group_name in cls.harness_groups_no_limitations:
       assert group_method is None, (
@@ -172,19 +174,19 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         custom_assert=custom_assert)
 
   @classmethod
-  def random_seed(cls, handess: primitive_harness.Harness):
+  def random_seed(cls, handess: test_harnesses.Harness):
     return [custom_random_keys_output()]
 
   @classmethod
-  def random_split(cls, handess: primitive_harness.Harness):
+  def random_split(cls, handess: test_harnesses.Harness):
     return [custom_random_keys_output()]
 
   @classmethod
-  def random_fold_in(cls, handess: primitive_harness.Harness):
+  def random_fold_in(cls, handess: test_harnesses.Harness):
     return [custom_random_keys_output()]
 
   @classmethod
-  def acos(cls, harness: primitive_harness.Harness):
+  def acos(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(
             dtypes=[np.complex64],
@@ -196,19 +198,35 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             devices=("cpu", "gpu"),
             tol=1e-13,
             modes=("eager", "graph", "compiled")),
+        custom_numeric(
+            dtypes=[np.complex64],
+            devices=("tpu",),
+            tol=1e-3,
+            modes=("eager", "graph", "compiled"),
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
     ]
 
   @classmethod
-  def acosh(cls, harness: primitive_harness.Harness):
+  def acosh(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
                        tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
-        cls.helper_get_trig_custom_limitation(np.cosh)
+        Jax2TfLimitation(
+            "TF2XLA impl for Acosh doesn't properly handle large complex types,"
+            " native serialization more closely matches numpy numerics.",
+            dtypes=[np.complex64, np.complex128],
+            devices=("cpu", "gpu", "tpu"),
+            modes="compiled",
+            expect_tf_error=False,
+            skip_comparison=True,
+            native_serialization=Jax2TfLimitation.FOR_NON_NATIVE,
+        ),
+        cls.helper_get_trig_custom_limitation(np.cosh),
     ]
 
   @classmethod
-  def approx_top_k(cls, harness: primitive_harness.Harness):
+  def approx_top_k(cls, harness: test_harnesses.Harness):
     supported_dtypes = jtu.supported_dtypes()
     def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
       del tol, err_msg
@@ -245,7 +263,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             custom_assert=custom_assert)]
 
   @classmethod
-  def argmax(cls, harness: primitive_harness.Harness):
+  def argmax(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             "different results when the input contains NaN and enable_xla=False",
@@ -258,11 +276,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def argmin(cls, harness: primitive_harness.Harness):
+  def argmin(cls, harness: test_harnesses.Harness):
     return cls.argmax(harness)
 
   @classmethod
-  def asin(cls, harness: primitive_harness.Harness):
+  def asin(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-4,
                        modes=("eager", "graph", "compiled")),
@@ -274,16 +292,25 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def asinh(cls, harness: primitive_harness.Harness):
+  def asinh(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu", "tpu"),
                        tol=1e-3),
         custom_numeric(dtypes=[np.complex128], devices=("cpu", "gpu"), tol=1e-12),
+        custom_numeric(dtypes=[np.complex64, np.complex128],
+                       devices=("cpu", "gpu", "tpu"),
+                       modes=("compiled",),
+                       tol=1e-3,
+                       native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
+        custom_numeric(dtypes=[np.complex128], devices=("cpu",),
+                       modes=("eager", "compiled", "graph"),
+                       tol=1e-13,
+                       native_serialization=Jax2TfLimitation.FOR_NATIVE | Jax2TfLimitation.FOR_NON_NATIVE),
         cls.helper_get_trig_custom_limitation(np.sinh)
     ]
 
   @classmethod
-  def atan(cls, harness: primitive_harness.Harness):
+  def atan(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.complex64], devices=("cpu", "gpu"), tol=1e-5),
         custom_numeric(dtypes=[np.complex64], devices=("tpu"), tol=1e-3),
@@ -292,7 +319,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def atanh(cls, harness: primitive_harness.Harness):
+  def atanh(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.float64], tol=1e-14),
         custom_numeric(dtypes=[np.complex64], tol=1e-3),
@@ -301,7 +328,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def bessel_i0e(cls, harness: primitive_harness.Harness):
+  def bessel_i0e(cls, harness: test_harnesses.Harness):
     return [
         missing_tf_kernel(
             dtypes=[dtypes.bfloat16],
@@ -310,17 +337,17 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def bessel_i1e(cls, harness: primitive_harness.Harness):
+  def bessel_i1e(cls, harness: test_harnesses.Harness):
     return cls.bessel_i0e(harness)
 
   @classmethod
-  def cbrt(cls, harness: primitive_harness.Harness):
+  def cbrt(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.float32], devices=("tpu"), tol=1e-5),
     ]
 
   @classmethod
-  def cholesky(cls, harness: primitive_harness.Harness):
+  def cholesky(cls, harness: test_harnesses.Harness):
 
     def custom_assert(tst, result_jax, result_tf, *, tol, err_msg, **_):
       # cholesky_p returns garbage in the strictly upper triangular part of the
@@ -345,6 +372,13 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             tol=5e-2,
             devices=("cpu", "gpu"),
             modes=("eager", "graph", "compiled")),
+        custom_numeric(
+            dtypes=[dtypes.bfloat16],
+            tol=5e-5,
+            # Error for GL
+            devices=("tpu",),
+            modes=("eager", "graph", "compiled"),
+            native_serialization=Jax2TfLimitation.FOR_NATIVE),
         custom_numeric(
             custom_assert=custom_assert,
             description=(
@@ -356,7 +390,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def conv_general_dilated(cls, harness: primitive_harness.Harness):
+  def conv_general_dilated(cls, harness: test_harnesses.Harness):
     prefer_elem = harness.params["preferred_element_type"]
     return [
         Jax2TfLimitation(
@@ -387,39 +421,38 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   @classmethod
   def cumlogsumexp(cls, harness):
     return [
-        # JAX uses a different lowering for CPU and GPU.
         custom_numeric(
-            dtypes=(np.float16, jnp.bfloat16),
-            devices=("cpu", "gpu"),
+            dtypes=(np.float16, jnp.bfloat16, np.float32),
+            devices=("cpu", "gpu", "tpu"),
             modes=("eager", "graph", "compiled"),
-            tol=5e-1)
+            tol=5e-1,
+        )
     ]
 
-
   @classmethod
   def cumprod(cls, harness):
     return [
-        # JAX uses a different lowering for CPU and GPU.
         custom_numeric(
             dtypes=(np.float16, jnp.bfloat16),
-            devices=("cpu", "gpu"),
+            devices=("cpu", "gpu", "tpu"),
             modes=("eager", "graph", "compiled"),
-            tol=5e-1)
+            tol=5e-1,
+        )
     ]
 
   @classmethod
   def cumsum(cls, harness):
     return [
-        # JAX uses a different lowering for CPU and GPU.
         custom_numeric(
             dtypes=(np.float16, jnp.bfloat16),
-            devices=("cpu", "gpu"),
+            devices=("cpu", "gpu", "tpu"),
             modes=("eager", "graph", "compiled"),
-            tol=5e-1)
+            tol=5e-1,
+        )
     ]
 
   @classmethod
-  def custom_linear_solve(cls, harness: primitive_harness.Harness):
+  def custom_linear_solve(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             "TODO: large numerical discrepancy",
@@ -432,7 +465,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def digamma(cls, harness: primitive_harness.Harness):
+  def digamma(cls, harness: test_harnesses.Harness):
     dtype = harness.dtype
 
     # In the bfloat16 case, TF and lax both return NaN in undefined cases.
@@ -474,7 +507,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def div(cls, harness: primitive_harness.Harness):
+  def div(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             "TF integer division fails if divisor contains 0; JAX returns NaN",
@@ -487,7 +520,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def dot_general(cls, harness: primitive_harness.Harness):
+  def dot_general(cls, harness: test_harnesses.Harness):
     prefer_elem = harness.params["preferred_element_type"]
     return [
         missing_tf_kernel(dtypes=[np.bool_],),
@@ -545,10 +578,16 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         # may be more precise.
         custom_numeric(dtypes=[np.float16], devices=["cpu"], tol=1e-2,
                        modes=("eager", "graph", "compiled")),
+        # Flakiness on different_dtypes_lhs_int16_4_3_rhs_float16_3_6_dimensionnumbers_1_0_enable_xla_True
+        # Strangely, we only see the flakiness in primitives_graph_serialization_test_gpu_pjrt_c_api
+        custom_numeric(dtypes=[np.int16], devices=["gpu"], tol=1e-2,
+                       modes=("eager", "graph", "compiled"),
+                       enabled=(harness.params["enable_xla"] and
+                                harness.dtype != harness.params["rhs_dtype"])),
     ]
 
   @classmethod
-  def eig(cls, harness: primitive_harness.Harness):
+  def eig(cls, harness: test_harnesses.Harness):
     compute_left_eigenvectors = harness.params["compute_left_eigenvectors"]
     compute_right_eigenvectors = harness.params["compute_right_eigenvectors"]
     dtype = harness.dtype
@@ -615,7 +654,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def eigh(cls, harness: primitive_harness.Harness):
+  def eigh(cls, harness: test_harnesses.Harness):
     dtype = harness.dtype
 
     def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
@@ -687,15 +726,15 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def erf(cls, harness: primitive_harness.Harness):
+  def erf(cls, harness: test_harnesses.Harness):
     return []
 
   @classmethod
-  def erfc(cls, harness: primitive_harness.Harness):
+  def erfc(cls, harness: test_harnesses.Harness):
     return []
 
   @classmethod
-  def erf_inv(cls, harness: primitive_harness.Harness):
+  def erf_inv(cls, harness: test_harnesses.Harness):
     # erf_inv is not defined for arg <= -1 or arg >= 1
     def custom_assert(tst, result_jax, result_tf, *, args, tol,
                       err_msg):  # noqa: F811
@@ -726,7 +765,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def expm1(cls, harness: primitive_harness.Harness):
+  def expm1(cls, harness: test_harnesses.Harness):
     return [custom_numeric(dtypes=[np.float64], tol=1e-5)]
 
   @classmethod
@@ -745,11 +784,15 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             enabled=(str(harness.params["fft_type"]) in ["FftType.IFFT",
                                                          "FftType.IRFFT"])),
         # TODO: very high tolerance
-        custom_numeric(tol=1e-3, modes=("eager", "graph", "compiled")),
+        custom_numeric(tol=1e-3, modes=("eager", "graph", "compiled"),
+                       native_serialization=Jax2TfLimitation.FOR_NON_NATIVE),
+        custom_numeric(tol=1e-5, modes=("eager", "graph", "compiled"),
+                       native_serialization=Jax2TfLimitation.FOR_NATIVE,
+                       devices=("cpu",)),
     ]
 
   @classmethod
-  def _pow_test_util(cls, harness: primitive_harness.Harness):
+  def _pow_test_util(cls, harness: test_harnesses.Harness):
 
     def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
       # NaNs are mismatched, but assertAllClose will also behave weirdly for
@@ -776,7 +819,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def igamma(cls, harness: primitive_harness.Harness):
+  def igamma(cls, harness: test_harnesses.Harness):
     dtype = harness.dtype
 
     # igamma is not defined when the first argument is <=0
@@ -815,7 +858,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def igammac(cls, harness: primitive_harness.Harness):
+  def igammac(cls, harness: test_harnesses.Harness):
     dtype = harness.dtype
 
     # igammac is not defined when the first argument is <=0
@@ -858,7 +901,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def integer_pow(cls, harness: primitive_harness.Harness):
+  def integer_pow(cls, harness: test_harnesses.Harness):
     y = harness.params["y"]
     return [
         # TODO: on TPU, for f16, we get different results with eager mode
@@ -884,11 +927,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ] + list(cls._pow_test_util(harness))
 
   @classmethod
-  def pow(cls, harness: primitive_harness.Harness):
+  def pow(cls, harness: test_harnesses.Harness):
     return cls._pow_test_util(harness)
 
   @classmethod
-  def lgamma(cls, harness: primitive_harness.Harness):
+  def lgamma(cls, harness: test_harnesses.Harness):
     return [
         missing_tf_kernel(
             dtypes=[dtypes.bfloat16],
@@ -899,7 +942,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def log1p(cls, harness: primitive_harness.Harness):
+  def log1p(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.complex128], tol=3e-14),
         custom_numeric(dtypes=[np.float64], tol=1e-10),
@@ -907,7 +950,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def lu(cls, harness: primitive_harness.Harness):
+  def lu(cls, harness: test_harnesses.Harness):
     dtype = harness.dtype
 
     def custom_assert(tst, result_jax, result_tf, *, args, tol, err_msg):
@@ -958,7 +1001,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def max(cls, harness: primitive_harness.Harness):
+  def max(cls, harness: test_harnesses.Harness):
     # TODO(bchetioui): discrepancies between TF & JAX when comparing with NaN;
     # JAX always returns NaN, while TF returns the value NaN is compared with.
     def custom_assert(tst, result_jax, result_tf, err_msg, **_):
@@ -988,7 +1031,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def min(cls, harness: primitive_harness.Harness):
+  def min(cls, harness: test_harnesses.Harness):
     # TODO(bchetioui): discrepancies between TF & JAX when comparing with NaN;
     # JAX always returns NaN, while TF returns the value NaN is compared with.
     def custom_assert(tst, result_jax, result_tf, *, err_msg, **_):
@@ -1018,11 +1061,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def nextafter(cls, harness: primitive_harness.Harness):
+  def nextafter(cls, harness: test_harnesses.Harness):
     return [missing_tf_kernel(dtypes=[np.float16, dtypes.bfloat16])]
 
   @classmethod
-  def qr(cls, harness: primitive_harness.Harness):
+  def qr(cls, harness: test_harnesses.Harness):
     # See https://github.com/google/jax/pull/3775#issuecomment-659407824;
     #     # jit_compile=True breaks for complex types.
     # TODO: see https://github.com/google/jax/pull/3775#issuecomment-659407824.
@@ -1047,21 +1090,21 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def random_gamma(cls, harness: primitive_harness.Harness):
+  def random_gamma(cls, harness: test_harnesses.Harness):
     return [custom_numeric(devices="tpu", tol=1e-3)]
 
   @classmethod
-  def reduce_max(cls, harness: primitive_harness.Harness):
+  def reduce_max(cls, harness: test_harnesses.Harness):
     # Unlike reduce_window_max, we use a native TF op: tf.reduce_max, which
     # does not work for complex
     return [missing_tf_kernel(dtypes=[np.complex64, np.complex128])]
 
   @classmethod
-  def reduce_min(cls, harness: primitive_harness.Harness):
+  def reduce_min(cls, harness: test_harnesses.Harness):
     return cls.reduce_max(harness)
 
   @classmethod
-  def reduce_window_add(cls, harness: primitive_harness.Harness):
+  def reduce_window_add(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             "Small deviations on GPU for large inputs and enable_xla=False",
@@ -1074,7 +1117,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
             tol=3e-5),
         Jax2TfLimitation(
             "Large deviations on TPU for enable_xla=False",
-            dtypes=[np.float16, np.float32],
+            dtypes=[dtypes.bfloat16, np.float16, np.float32],
             devices="tpu",
             modes=("eager", "graph", "compiled"),
             expect_tf_error=False,
@@ -1084,17 +1127,19 @@ class Jax2TfLimitation(primitive_harness.Limitation):
                      modes=("eager", "graph", "compiled",), tol=1e-5),
       custom_numeric(devices=("cpu", "gpu"), dtypes=[np.float16],
                      modes=("eager", "graph", "compiled",), tol=5e-3),
+      custom_numeric(devices=("cpu", "gpu"), dtypes=[dtypes.bfloat16],
+                     modes=("eager", "graph", "compiled",), tol=5e-1),
     ]
 
   @classmethod
-  def regularized_incomplete_beta(cls, harness: primitive_harness.Harness):
+  def regularized_incomplete_beta(cls, harness: test_harnesses.Harness):
     return [
         custom_numeric(dtypes=[np.float64], tol=1e-14),
         missing_tf_kernel(dtypes=[np.float16, dtypes.bfloat16])
     ]
 
   @classmethod
-  def rem(cls, harness: primitive_harness.Harness):
+  def rem(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             "TF integer division fails if divisor contains 0; JAX returns NaN",
@@ -1116,11 +1161,11 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def rng_bit_generator(cls, harness: primitive_harness.Harness):
+  def rng_bit_generator(cls, harness: test_harnesses.Harness):
     return []
 
   @classmethod
-  def round(cls, harness: primitive_harness.Harness):
+  def round(cls, harness: test_harnesses.Harness):
     return [
         missing_tf_kernel(
             dtypes=[dtypes.bfloat16],
@@ -1163,7 +1208,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
   @classmethod
   def select_and_gather_add(cls, harness):
     return [
-        # This JAX primitives is not not exposed directly in the JAX API
+        # This JAX primitives is not exposed directly in the JAX API
         # but arises from JVP of `lax.reduce_window` for reducers
         # `lax.max` or `lax.min`. It also arises from second-order
         # VJP of the same. Implemented using XlaReduceWindow.
@@ -1176,7 +1221,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def sort(cls, harness: primitive_harness.Harness):
+  def sort(cls, harness: test_harnesses.Harness):
     return [
         Jax2TfLimitation(
             # I think that this is because TF is running on CPU even for GPU tests?
@@ -1189,7 +1234,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def svd(cls, harness: primitive_harness.Harness):
+  def svd(cls, harness: test_harnesses.Harness):
     # TODO: slow test
     compute_uv = harness.params["compute_uv"]
 
@@ -1273,7 +1318,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
         # values like 1.0000001 on float32, which are clipped to 1.0. It is
         # possible that anything other than `cos_angular_diff` can be outside
         # the interval [0, 1] due to roundoff.
-        cos_angular_diff = jnp.clip(cos_angular_diff, a_min=0.0, a_max=1.0)
+        cos_angular_diff = jnp.clip(cos_angular_diff, min=0.0, max=1.0)
 
         angular_diff = jnp.arccos(cos_angular_diff)
 
@@ -1429,7 +1474,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def triangular_solve(cls, harness: primitive_harness.Harness):
+  def triangular_solve(cls, harness: test_harnesses.Harness):
     return [
         missing_tf_kernel(
             dtypes=[dtypes.bfloat16],
@@ -1444,7 +1489,7 @@ class Jax2TfLimitation(primitive_harness.Limitation):
     ]
 
   @classmethod
-  def tridiagonal_solve(cls, harness: primitive_harness.Harness):
+  def tridiagonal_solve(cls, harness: test_harnesses.Harness):
     return []
 
 def custom_numeric(
