@@ -19,7 +19,6 @@ import contextlib
 import math
 import os
 import re
-from typing import Callable, Optional
 import unittest
 
 from absl import logging
@@ -28,36 +27,55 @@ from absl.testing import absltest, parameterized
 import jax
 from jax import ad_checkpoint
 from jax import dtypes
+from jax import export
 from jax import lax
 from jax import numpy as jnp
 from jax import sharding
+from jax._src import config
 from jax._src import core
 from jax._src import source_info_util
 from jax._src import test_util as jtu
-from jax._src.lib import xla_client
-from jax._src.lib.mlir.dialects import hlo
-import jax._src.xla_bridge
-from jax import config
+from jax._src import xla_bridge as xb
 from jax.experimental import jax2tf
-from jax.experimental.export import export
 from jax.experimental.jax2tf.tests import tf_test_util
-from jax.experimental.maps import xmap
 from jax.experimental.shard_map import shard_map
 from jax.experimental import pjit
-from jax.interpreters import mlir
 from jax.sharding import PartitionSpec as P
 
 import numpy as np
-import tensorflow as tf  # type: ignore[import]
+import tensorflow as tf
 # pylint: disable=g-direct-tensorflow-import
-from tensorflow.compiler.tf2xla.python import xla as tfxla  # type: ignore[import]
+from tensorflow.compiler.tf2xla.python import xla as tfxla
 # pylint: enable=g-direct-tensorflow-import
 
 config.parse_flags_with_absl()
+_exit_stack = contextlib.ExitStack()
+
+# TODO(necula): Remove once tensorflow is 2.10.0 everywhere.
+def setUpModule():
+  if not hasattr(tfxla, "optimization_barrier"):
+    _exit_stack.enter_context(jtu.global_config_context(jax_remat_opt_barrier=False))
+
+def tearDownModule():
+  _exit_stack.close()
 
 
 class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
+  @classmethod
+  def setUpClass(cls):
+    # One TF device of each device_type
+    cls.tf_devices = []
+    for tf_device in (tf.config.list_logical_devices("TPU") +
+                      tf.config.list_logical_devices("GPU") +
+                      tf.config.list_logical_devices()):
+      if tf_device.device_type == "TPU_SYSTEM":
+        continue  # A virtual device
+      if all(tf_device.device_type != d.device_type for d in cls.tf_devices):
+        cls.tf_devices.append(tf_device)
+
+    super().setUpClass()
+
   def test_empty(self):
     f_jax = lambda x, y: x
     self.ConvertAndCompare(f_jax, 0.7, 1)
@@ -233,7 +251,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     with_function=[True, False],
   )
   def test_converts_64bit(self, dtype=np.int64, with_function=False):
-    if not config.jax_enable_x64:
+    if not config.enable_x64.value:
       self.skipTest("requires x64 mode")
     big_const = np.full((5,), 2 ** 33, dtype=dtype)
     self.ConvertAndCompare(jnp.sin, big_const)
@@ -249,7 +267,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_64bit_behavior_enable_x64_readme(self):
     # Tests some of the examples from the README
-    if not config.jax_enable_x64:
+    if not config.enable_x64.value:
       self.skipTest("requires x64 mode")
 
     # JAX and TF have different default float types if JAX_ENABLE_X64=1
@@ -266,7 +284,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_64bit_behavior_not_enable_x64_readme(self):
     # Tests some of the examples from the README
-    if config.jax_enable_x64:
+    if config.enable_x64.value:
       self.skipTest("requires not x64 mode")
 
     # JAX and TF have same default float types if JAX_ENABLE_X64=0
@@ -817,7 +835,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     arg = np.array(3.)
     f_tf = jax2tf.convert(jax.grad(remat_f))
     f_tf_hlo = self.TfToHlo(f_tf, arg)
-    if jax.config.jax_remat_opt_barrier:
+    if config.remat_opt_barrier.value:
       self.assertRegex(f_tf_hlo, r"opt-barrier")
     else:
       self.assertRegex(f_tf_hlo,
@@ -841,17 +859,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       return ad_checkpoint.checkpoint_name(jnp.sin(x), "sin")
     jax2tf.convert(f_jax)(1.)  # No error.
 
-  def test_convert_nullary_func(self):
-    # Even nullary functions are converted to TF (as opposed to constant-folded
-    # in JAX prior to conversion).
-    def f_jax():
-      return jnp.sin(1.)
-    f_tf = jax2tf.convert(f_jax)
-    # for native serialization the HLO we get from TF is constant-folded, so this
-    # test fails.
-    if not config.jax2tf_default_native_serialization:
-      self.assertIn("sine(", self.TfToHlo(f_tf))
-
   def test_convert_of_nested_independent_jit(self):
     def func(x):
       def inner1(y):
@@ -954,12 +961,12 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
       out = jax2tf.convert(caller_jax, with_gradient=False)(2.)
       return out
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       self.assertIn("my_test_function_jax/mul", self.TfToHlo(run_tf))
     else:
       graph_def = str(tf.function(run_tf, autograph=False).get_concrete_function().graph.as_graph_def())
-      if "my_test_function_jax/pjit_fn_/Mul" not in graph_def:
-        self.assertIn("my_test_function_jax/jit_fn_/Mul", graph_def)
+      if "my_test_function_jax/pjit_multiply_/Mul" not in graph_def:
+        self.assertIn("my_test_function_jax/jit_multiply_/Mul", graph_def)
 
   def test_bfloat16_constant(self):
     # Re: https://github.com/google/jax/issues/3942
@@ -984,7 +991,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_shared_constants(self):
     # Check that the constants are shared properly in converted functions
     # See https://github.com/google/jax/issues/7992.
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       raise unittest.SkipTest("shared constants tests not interesting for native serialization")
     const = np.random.uniform(size=256).astype(np.float32)  # A shared constant
     def f(x):
@@ -996,7 +1003,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
   def test_shared_constants_under_cond(self):
     # Check that the constants are shared properly in converted functions
     # See https://github.com/google/jax/issues/7992.
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       raise unittest.SkipTest("shared constants tests not interesting for native serialization")
     const_size = 512
     const = np.random.uniform(size=const_size).astype(np.float32)  # A shared constant
@@ -1012,7 +1019,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_shared_constants_under_scan(self):
     # See https://github.com/google/jax/issues/7992.
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       raise unittest.SkipTest("shared constants tests not interesting for native serialization")
     const_size = 512
     const = np.random.uniform(size=const_size).astype(np.float32)  # A shared constant
@@ -1031,7 +1038,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
 
   def test_shared_constants_under_jit(self):
     # We do not share constants under jit.
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       raise unittest.SkipTest("shared constants tests not interesting for native serialization")
     const = np.random.uniform(size=(16, 16)).astype(np.float32)  # A shared constant
     @jax.jit
@@ -1047,7 +1054,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     # randint has the property that the TF lowering of the randbits_p
     # primitive generates constants that did not exist in the Jaxpr. As such
     # it has created new errors related to the sharing of the constants.
-    if config.jax2tf_default_native_serialization:
+    if config.jax2tf_default_native_serialization.value:
       raise unittest.SkipTest("shared constants tests not interesting for native serialization")
 
     key = jax.random.PRNGKey(42)
@@ -1113,31 +1120,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
                          np.full_like(x[1], fill_value=2.)),
                         (grad_tf[0].numpy(), grad_tf[1].numpy()))
 
-  @jtu.skip_on_flag("jax2tf_default_native_serialization", True)
-  def test_enable_xla(self):
-    # Tests that enable_xla flag is properly scoped to a conversion.
-    def fun(x):
-      # lax.reduce is unlikely to ever be convertible with enable_xla=False
-      return lax.reduce(x, np.float32(0), lambda v, acc: v + acc, dimensions=(0, 1))
-
-    tf_fun_with_xla = jax2tf.convert(fun, enable_xla=True)
-    tf_fun_without_xla = jax2tf.convert(fun, enable_xla=False)
-    x = np.ones((2, 3), dtype=np.float32)
-
-    self.assertAllClose(fun(x), tf_fun_with_xla(x))
-    with self.assertRaisesRegex(NotImplementedError,
-                                "Call to reduce cannot be converted with enable_xla=False"):
-      tf_fun_without_xla(x)
-
-    # Now in reverse order (we had bugs with the management of enable_xla global)
-    tf_fun2_without_xla = jax2tf.convert(lambda x: fun(x), enable_xla=False)
-    tf_fun2_with_xla = jax2tf.convert(lambda x: fun(x), enable_xla=True)
-
-    with self.assertRaisesRegex(NotImplementedError,
-                                "Call to reduce cannot be converted with enable_xla=False"):
-      tf_fun2_without_xla(x)
-    self.assertAllClose(fun(x), tf_fun2_with_xla(x))
-
   def test_device_array_arg(self):
     self.ConvertAndCompare(jnp.sin, jnp.zeros((2, 3), jnp.float32))
 
@@ -1295,8 +1277,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     shape = (3, 2)
     x = np.arange(math.prod(shape), dtype=np.float32).reshape(shape)
 
-    jax_comp = jax.xla_computation(f_while)(x)
-    backend = jax._src.xla_bridge.get_backend()
+    jax_comp = jax.jit(f_while).lower(x).compiler_ir('hlo')
+    backend = xb.get_backend()
     modules = backend.compile(jax_comp).hlo_modules()
     jax_opt_hlo = modules[0].to_string()
     print(f"JAX OPT HLO = {jax_opt_hlo}")
@@ -1342,7 +1324,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
         self.fail(f"{op.name} does not start with {scope_name}.")
 
   def test_name_scope_polymorphic(self):
-    if config.jax2tf_default_native_serialization and not config.jax_dynamic_shapes:
+    if (config.jax2tf_default_native_serialization.value and
+        not config.dynamic_shapes.value):
       self.skipTest("shape polymorphism but --jax_dynamic_shapes is not set.")
 
     def func_jax(x, y):
@@ -1443,8 +1426,7 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
           "none",
           "jit",
           "pjit", "pjit_in_shardings_None", "pjit_in_shardings_P",
-          "pjit_in_shardings_Sharding",
-          "shard_map", "xmap", "pmap"]
+          "pjit_in_shardings_Sharding", "shard_map", "pmap"]
       for transform2 in (
           ["none", "pjit_in_shardings_None", "pjit_in_shardings_P",
            "pjit_in_shardings_Sharding"]
@@ -1483,10 +1465,10 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       transformed_func = dict(
           none=func,
           jit=jax.jit(func),
-          jit_in_shardings_None=jax.jit(func, in_shardings=None),  # type: ignore
-          jit_in_shardings_P=jax.jit(func, in_shardings=(P("a"),)),  # type: ignore
+          jit_in_shardings_None=jax.jit(func, in_shardings=None),
+          jit_in_shardings_P=jax.jit(func, in_shardings=(P("a"),)),
           jit_in_shardings_Sharding=jax.jit(
-              func, in_shardings=(sharding.NamedSharding(mesh, P("a")),)),  # type: ignore
+              func, in_shardings=(sharding.NamedSharding(mesh, P("a")),)),
           pjit=pjit.pjit(func),
           pjit_in_shardings_None=pjit.pjit(func, in_shardings=None,
                                            out_shardings=None),
@@ -1499,8 +1481,6 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
           shard_map=(
               shard_map(func, mesh, in_specs=(P("a", None),),
                         out_specs=P("a", None))),
-          xmap=xmap(func, in_axes=({0: 'axis'},),
-                    out_axes={0: 'axis'}, axis_resources={'axis': 'a'}),
           pmap=jax.pmap(func, in_axes=0, out_axes=0),
       )[transform]
       return transformed_func
@@ -1508,12 +1488,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     transformed1_func = apply_transform(
         (func_shard_map if transform1 == "shard_map" else func),
         transform1)
-    assert transform2 not in ["xmap", "shard_map"]
+    assert transform2 not in ["shard_map"]
     transformed2_func = apply_transform(transformed1_func, transform2)
 
-    if transform1 == "xmap" and transform2 in ["pjit", "none"]:
-      raise unittest.SkipTest("TODO: pjit(xmap) with unspecified shardings crashes")
-
     if transform1 == "pmap":
       x = x.reshape((1, -1))  # Since we use 1 device
     if not nullary:
@@ -1539,8 +1516,8 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
       # Run the JAX native version, to check it works, and to fill caches.
       _ = func_to_convert(*args)
       exported = export.export(
-          func_to_convert,
-          lowering_platform='tpu'
+          (jax.jit(func_to_convert) if not hasattr(func_to_convert, "trace") else func_to_convert),
+          lowering_platforms=("tpu",)
       )(*(core.ShapedArray(a.shape, a.dtype) for a in args))
 
     if transform1 == "shard_map":
@@ -1664,6 +1641,39 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     f_tf_nested = jax2tf.convert(f_jax_nested, native_serialization=True)
     self.assertAllClose(res, f_tf_nested(inputs))
 
+  def test_multi_platform(self):
+    if config.enable_x64.value:
+      self.skipTest("TODO: enable when we can handle i64 platform_index_argument")
+    # Checks that we dispatch from TF to the proper JAX platform lowering.
+
+    # We add a different value to it: cpu=2., tpu=3., cuda=.4, rocm=5.
+    _testing_multi_platform_to_add = dict(cpu=2., tpu=3., cuda=4., rocm=5.)
+
+    def f_jax(x):
+      return x + lax.platform_dependent(
+        tpu=lambda: _testing_multi_platform_to_add["tpu"],
+        cuda=lambda: _testing_multi_platform_to_add["cuda"],
+        rocm=lambda: _testing_multi_platform_to_add["rocm"],
+        default=lambda: _testing_multi_platform_to_add["cpu"]
+      )
+
+    x = np.float32(.42)
+    f_tf = jax2tf.convert(
+      f_jax,
+      native_serialization=True,
+      native_serialization_platforms=("cpu", "cuda", "tpu"))
+    for tf_device in self.__class__.tf_devices:
+      logging.info(
+        f"Running on tf_device = {tf_device} of device_type = {tf_device.device_type}")
+      with tf.device(tf_device):
+        res = f_tf(x)
+      tf_device_jax_platform = dict(
+        CPU="cpu", GPU="cuda", TPU="tpu"
+      )[tf_device.device_type]
+      self.assertAllClose(
+        res,
+        x + _testing_multi_platform_to_add[tf_device_jax_platform])
+
 
 @jtu.with_config(jax_enable_custom_prng=True)
 class Jax2tfWithCustomPRNGTest(tf_test_util.JaxToTfTestCase):
@@ -1704,7 +1714,4 @@ class Jax2TfVersioningTest(tf_test_util.JaxToTfTestCase):
 
 
 if __name__ == "__main__":
-  # TODO: Remove once tensorflow is 2.10.0 everywhere.
-  if not hasattr(tfxla, "optimization_barrier"):
-    jax.config.update("jax_remat_opt_barrier", False)
   absltest.main(testLoader=jtu.JaxTestLoader())
