@@ -13,6 +13,8 @@
 # limitations under the License.
 """Utilities for synchronizing and communication across multiple hosts."""
 
+from __future__ import annotations
+
 from functools import partial, lru_cache
 from typing import Optional
 import zlib
@@ -38,10 +40,10 @@ import numpy as np
 
 
 def _psum(x: Any) -> Any:
-  return jax.tree_map(partial(jnp.sum, axis=0), x)
+  return jax.tree.map(partial(jnp.sum, axis=0), x)
 
 
-def broadcast_one_to_all(in_tree: Any, is_source: Optional[bool] = None) -> Any:
+def broadcast_one_to_all(in_tree: Any, is_source: bool | None = None) -> Any:
   """Broadcast data from a source host (host 0 by default) to all other hosts.
 
   Args:
@@ -55,6 +57,9 @@ def broadcast_one_to_all(in_tree: Any, is_source: Optional[bool] = None) -> Any:
     A pytree matching in_tree where the leaves now all contain the data from the
     first host.
   """
+  if jax.process_count() == 1:
+    return jax.tree.map(np.asarray, in_tree)
+
   if is_source is None:
     is_source = jax.process_index() == 0
 
@@ -74,10 +79,10 @@ def broadcast_one_to_all(in_tree: Any, is_source: Optional[bool] = None) -> Any:
   def post_jit(x):
     return np.asarray(x.addressable_data(0))
 
-  in_tree = jax.tree_map(pre_jit, in_tree)
+  in_tree = jax.tree.map(pre_jit, in_tree)
   out_tree = jax.jit(_psum, out_shardings=jax.sharding.NamedSharding(
       global_mesh, P()))(in_tree)
-  return jax.tree_map(post_jit, out_tree)
+  return jax.tree.map(post_jit, out_tree)
 
 
 def sync_global_devices(name: str):
@@ -146,7 +151,7 @@ def process_allgather(in_tree: Any, tiled: bool = False) -> Any:
 
   def _pjit(inp):
     return _handle_array_process_allgather(inp, tiled)
-  return jax.tree_map(_pjit, in_tree)
+  return jax.tree.map(_pjit, in_tree)
 
 
 def assert_equal(in_tree, fail_message: str = ''):
@@ -161,16 +166,16 @@ def assert_equal(in_tree, fail_message: str = ''):
 def reached_preemption_sync_point(step_id: int) -> bool:
   """Determine whether all hosts have reached a preemption sync step.
 
-  When any host receive a preemption notice, the notice will be propagated to
-  all hosts and trigger a synchronization protocol in background. The
+  When any host receives a preemption notice, the notice is propagated to all
+  hosts and triggers a synchronization protocol in the background. The
   synchronization protocol calculates the maximum step ids from all hosts, and
   uses the next step id (i.e., max + 1) as the safe step to save a checkpoint.
   All hosts should continue training more steps until this method returns True,
   indicating that the `step_id` is equal to the safe step and the hosts should
   start saving a checkpoint.
 
-  To use this API, all hosts must start training from the same step and call at
-  every training step. Example usage:
+  To use this API, all hosts must start training from the same step and call it
+  at every training step. Example usage:
 
   ```
   def should_save(step_id: int) -> bool:
@@ -243,9 +248,9 @@ def host_local_array_to_global_array_impl(
     arrays = [x.data for x in arr.addressable_shards]
   else:
     arr = xla.canonicalize_dtype(arr)
-    arrays = list(
+    arrays = [
         arr[index]
-        for d, index in local_sharding.devices_indices_map(arr.shape).items())
+        for d, index in local_sharding.devices_indices_map(arr.shape).items()]
 
   global_aval = _local_to_global_aval(
       core.ShapedArray(arr.shape, arr.dtype), global_mesh, pspec)
@@ -259,13 +264,56 @@ def host_local_array_to_global_array(
     local_inputs: Any, global_mesh: jax.sharding.Mesh, pspecs: Any):
   r"""Converts a host local value to a globally sharded jax.Array.
 
+  This function takes host-local data (which might be different
+  across hosts), and populates a global array with this data, where each
+  device on each host, get the appropriate slice of the data according to
+  sharding defined by the global_mesh/pspects.
+
+  For example:
+
+  >>> global_mesh = jax.sharding.Mesh(jax.devices(), 'x')
+  >>> pspecs = jax.sharding.PartitionSpec('x')
+  >>> host_id = jax.process_index()
+  >>> arr = host_local_array_to_global_array(np.arange(4) * host_id, mesh, pspecs)  # NB: assumes jax.local_device_count() divides 4.   # doctest: +SKIP
+
+  The resulting array will have the shape (4 * num_processes) and will
+  have distributed value of: (0, 1, 2, 3, 0, 2, 4, 6, 0, 3, 6, 9, ... ),
+  where each slice np.arange(4) * host_id will be partitioned across the
+  corresponding host's devices.
+
+  Similarly:
+
+  >>> mesh = jax.sharding.Mesh(np.array(jax.devices()).reshape(jax.process_count(), jax.local_device_count()), ['host', 'dev'])
+  >>> pspecs = jax.sharding.PartitionSpec('host')
+  >>> host_id = jax.process_index()
+  >>> arr = host_local_array_to_global_array(np.arange(4) * host_id, mesh, pspecs)  # doctest: +SKIP
+
+  will create the same distributed value (0, 1, 2, 3, 0, 2, 4, 6, ...),
+  however each slice np.arange(4) * i will be *replicated* across corresponding
+  host devices.
+
+  On the other hand, if pspecs = PartitionSpec(), which means
+  replication across all axes, then this snippet:
+
+  >>> pspecs = jax.sharding.PartitionSpec()
+  >>> arr = host_local_array_to_global_array(np.arange(4), mesh, pspecs)  # doctest: +SKIP
+
+  will have the shape (4,) and the value (0, 1, 2, 3) will be replicated
+  across all hosts and devices.
+
+  It is an undefined behavior to have not identical local_inputs with pspec
+  indicating data replication.
+
   You can use this function to transition to jax.Array. Using jax.Array with
   pjit has the same semantics of using GDA with pjit i.e. all jax.Array
   inputs to pjit should be globally shaped.
 
   If you are currently passing host local values to pjit, you can use this
   function to convert your host local values to global Arrays and then pass that
-  to pjit. Example usage.
+  to pjit.
+
+
+  Example usage.
 
   >>> from jax.experimental import multihost_utils # doctest: +SKIP
   >>>
@@ -276,10 +324,20 @@ def host_local_array_to_global_array(
   >>>
   >>> host_local_output = multihost_utils.global_array_to_host_local_array(global_out, mesh, out_pspecs) # doctest: +SKIP
 
+  Please note ths function requires global mesh to be a continuous mesh, meaning
+  that  devices that belong to each host should form a subcube in this mesh.
+  To move local data to global array with non-continuous mesh use
+  jax.make_array_from_callback or jax.make_array_from_single_device_arrays
+  instead.
+
   Args:
     local_inputs: A Pytree of host local values.
-    global_mesh: A jax.sharding.Mesh object.
+    global_mesh: A jax.sharding.Mesh object. The mesh must be a contiguous mesh,
+    that is all hosts' devices must form a subcube in this mesh.
     pspecs: A Pytree of jax.sharding.PartitionSpec's.
+
+  Returns:
+    A pytree of global arrays.
   """
   flat_inps, in_tree = tree_flatten(local_inputs)
   in_pspecs = _flatten_pspecs('input pspecs', in_tree,
@@ -311,7 +369,7 @@ def ltg_batcher(insert_axis, spmd_axis_name, axis_size,
   new_parts = None if spmd_axis_name is None else spmd_axis_name
   new_pspec = list(pspec)
   new_pspec.insert(d, new_parts)
-  new_pspec = P(*new_pspec)  # type: ignore
+  new_pspec = P(*new_pspec)
   y = host_local_array_to_global_array_p.bind(
       x, global_mesh=global_mesh, pspec=new_pspec)
   return y, d
@@ -350,9 +408,9 @@ def global_array_to_host_local_array_impl(
   else:
     # numpy array can show up here during AD.
     arr = xla.canonicalize_dtype(arr)
-    arrays = list(
+    arrays = [
         arr[index]
-        for d, index in local_sharding.devices_indices_map(arr.shape).items())
+        for d, index in local_sharding.devices_indices_map(arr.shape).items()]
     return pxla.batched_device_put(
         local_aval, local_sharding, arrays,
         list(global_mesh.local_mesh.devices.flat))
@@ -369,21 +427,27 @@ def global_array_to_host_local_array(
 
   You can use this function to convert the globally shaped `jax.Array` output
   from pjit to host local values again so that the transition to jax.Array can
-  be a mechanical change. Example usage
+  be a mechanical change.
 
-  >> from jax.experimental import multihost_utils # doctest: +SKIP
-  >>
-  >> global_inputs = multihost_utils.host_local_array_to_global_array(host_local_inputs, global_mesh, in_pspecs) # doctest: +SKIP
-  >>
-  >> with mesh: # doctest: +SKIP
-  >>   global_out = pjitted_fun(global_inputs) # doctest: +SKIP
-  >>
-  >> host_local_output = multihost_utils.global_array_to_host_local_array(global_out, mesh, out_pspecs) # doctest: +SKIP
+  Example usage:
+
+  >>> from jax.experimental import multihost_utils # doctest: +SKIP
+  >>>
+  >>> global_inputs = multihost_utils.host_local_array_to_global_array(host_local_inputs, global_mesh, in_pspecs) # doctest: +SKIP
+  >>>
+  >>> with mesh: # doctest: +SKIP
+  ...   global_out = pjitted_fun(global_inputs) # doctest: +SKIP
+  >>>
+  >>> host_local_output = multihost_utils.global_array_to_host_local_array(global_out, mesh, out_pspecs) # doctest: +SKIP
 
   Args:
     global_inputs: A Pytree of global jax.Array's.
-    global_mesh: A jax.sharding.Mesh object.
-    pspecs: A Pytree of jax.sharding.PartitionSpec's.
+    global_mesh: A :class:`jax.sharding.Mesh` object. The mesh must be contiguous
+      meaning all local devices of the host must form a subcube.
+    pspecs: A Pytree of :class:`jax.sharding.PartitionSpec` objects.
+
+  Returns:
+    A Pytree of host local arrays.
   """
   flat_inps, out_tree = tree_flatten(global_inputs)
   out_pspecs = _flatten_pspecs('output pspecs', out_tree,
