@@ -25,7 +25,6 @@ import numpy as np
 
 import jax
 import jax.numpy as jnp
-from jax import config
 from jax import lax
 from jax import tree_util
 from jax.experimental.sparse._base import JAXSparse
@@ -37,6 +36,7 @@ from jax.experimental.sparse.util import (
 from jax.util import split_list, safe_zip
 
 from jax._src import api_util
+from jax._src import config
 from jax._src import core
 from jax._src import dispatch
 from jax._src.lax.lax import DotDimensionNumbers, _dot_general_batch_dim_nums
@@ -194,9 +194,10 @@ def bcsr_fromdense(mat: ArrayLike, *, nse: int | None = None, n_batch: int = 0,
     mat_bcsr: BCSR representation of the matrix.
   """
   mat_array = jnp.asarray(mat)
-  if nse is None:
-    nse = _count_stored_elements(mat_array, n_batch, n_dense)
-  nse_int: int = core.concrete_or_error(operator.index, nse, _TRACED_NSE_ERROR)
+  nse_arr: int | Array | None = nse
+  if nse_arr is None:
+    nse_arr = _count_stored_elements(mat_array, n_batch, n_dense)
+  nse_int: int = core.concrete_or_error(operator.index, nse_arr, _TRACED_NSE_ERROR)
   return BCSR(_bcsr_fromdense(mat_array, nse=nse_int, n_batch=n_batch,
                               n_dense=n_dense, index_dtype=index_dtype),
               shape=mat_array.shape)
@@ -623,7 +624,7 @@ def _bcsr_dot_general_gpu_lowering(
     ctx, lhs_data, lhs_indices, lhs_indptr, rhs, *, dimension_numbers,
     preferred_element_type, lhs_spinfo: SparseInfo):
 
-  if not config.jax_bcoo_cusparse_lowering:
+  if not config.bcoo_cusparse_lowering.value:
     return _bcsr_dot_general_default_lowering(
       ctx, lhs_data, lhs_indices, lhs_indptr, rhs,
       dimension_numbers=dimension_numbers,
@@ -670,8 +671,8 @@ def _bcsr_dot_general_gpu_lowering(
 
   # Account for a bug in cusparse: it references indices and data beyond
   # the extent of indptr.
-  (lhs_data,), (lhs_indices,) = _bcsr_correct_out_of_bound_indices_lowered(
-      ctx, lhs_data, lhs_indices, lhs_indptr, rhs, shape=lhs_spinfo.shape)
+  lhs_data, lhs_indices = _bcsr_correct_out_of_bound_indices_lowered(
+    ctx, lhs_data, lhs_indices, lhs_indptr, rhs, shape=lhs_spinfo.shape)
 
   if rhs_aval.ndim == 1:
     dot_general_fn = csr_matvec_lowering
@@ -680,8 +681,7 @@ def _bcsr_dot_general_gpu_lowering(
     dot_general_fn = csr_matmat_lowering
     x_dtype = 'B_dtype'
     if rhs_contract[0] == 1:
-      rhs = hlo.TransposeOp(
-          rhs, permutation=mlir.dense_int_elements([1, 0])).result
+      rhs = hlo.transpose(rhs, permutation=mlir.dense_int_array([1, 0]))
   else:
     raise ValueError(f"rhs has to be 1d or 2d; get {rhs_aval.ndim}d.")
 
