@@ -14,7 +14,9 @@
 
 """Sparse linear algebra routines."""
 
-from typing import Union, Callable
+from __future__ import annotations
+
+from collections.abc import Callable
 import functools
 
 import jax
@@ -33,10 +35,10 @@ from scipy.sparse import csr_matrix, linalg
 
 
 def lobpcg_standard(
-    A: Union[jax.Array, Callable[[jax.Array], jax.Array]],
+    A: jax.Array | Callable[[jax.Array], jax.Array],
     X: jax.Array,
     m: int = 100,
-    tol: Union[jax.Array, float, None] = None):
+    tol: jax.Array | float | None = None):
   """Compute the top-k standard eigenvalues using the LOBPCG routine.
 
   LOBPCG [1] stands for Locally Optimal Block Preconditioned Conjugate Gradient.
@@ -106,7 +108,7 @@ def _lobpcg_standard_matrix(
     A: jax.Array,
     X: jax.Array,
     m: int,
-    tol: Union[jax.Array, float, None],
+    tol: jax.Array | float | None,
     debug: bool = False):
   """Computes lobpcg_standard(), possibly with debug diagnostics."""
   return _lobpcg_standard_callable(
@@ -117,7 +119,7 @@ def _lobpcg_standard_callable(
     A: Callable[[jax.Array], jax.Array],
     X: jax.Array,
     m: int,
-    tol: Union[jax.Array, float, None],
+    tol: jax.Array | float | None,
     debug: bool = False):
   """Supports generic lobpcg_standard() callable interface."""
 
@@ -343,7 +345,7 @@ def _svqb(X):
   tau = jnp.finfo(X.dtype).eps * w[0]
   padded = jnp.maximum(w, tau)
 
-  # Note the the tau == 0 edge case where X was all zeros.
+  # Note the tau == 0 edge case where X was all zeros.
   sqrted = jnp.where(tau > 0, padded, 1.0) ** (-0.5)
 
   # X^T X = V diag(w) V^T, so
@@ -389,7 +391,7 @@ def _project_out(basis, U):
   #
   # Interspersing with orthonormalization isn't directly grounded in the
   # original analysis, but taken from Algorithm 5 of [3]. In practice, due to
-  # normalization, I have noticed that that the orthonormalized basis
+  # normalization, I have noticed that the orthonormalized basis
   # does not always end up as a subspace of the starting basis in practice.
   # There may be room to refine this procedure further, but the adjustment
   # in the subsequent block handles this edge case well enough for now.
@@ -544,10 +546,9 @@ def _spsolve_cpu_lowering(ctx, data, indices, indptr, b, tol, reorder):
     A = csr_matrix((data, indices, indptr), shape=(b.size, b.size))
     return (linalg.spsolve(A, b).astype(b.dtype),)
 
-  result, _, keepalive = mlir.emit_python_callback(
+  result, _, _ = mlir.emit_python_callback(
       ctx, _callback, None, args, ctx.avals_in, ctx.avals_out,
       has_side_effect=False)
-  ctx.module_context.add_keepalive(keepalive)
   return result
 
 
