@@ -13,26 +13,52 @@
 # limitations under the License.
 """Sparse test utilities."""
 
-from collections.abc import Sequence
-import functools
-from typing import Any, Callable, Union
+from __future__ import annotations
 
-import numpy as np
+from collections.abc import Callable, Iterable, Iterator, Sequence
+import functools
+import itertools
+import math
+from typing import Any, NamedTuple
 
 import jax
 from jax import lax
-from jax._src import test_util as jtu
-from jax._src.typing import DTypeLike
 from jax import tree_util
-from jax.util import safe_zip, split_list
+from jax._src import test_util as jtu
+from jax._src.lax.lax import DotDimensionNumbers
+from jax._src.lib import gpu_sparse
+from jax._src.typing import DTypeLike
 from jax.experimental import sparse
 import jax.numpy as jnp
+from jax.util import safe_zip, split_list
+import numpy as np
+
+MATMUL_TOL = {
+    np.float32: 1e-5,
+    np.float64: 1e-10,
+    np.complex64: 1e-5,
+    np.complex128: 1e-10,
+}
 
 
 def is_sparse(x):
   return isinstance(x, sparse.JAXSparse)
 
 
+class BatchedDotGeneralProperties(NamedTuple):
+  lhs_shape: tuple[int, ...]
+  rhs_shape: tuple[int, ...]
+  n_batch: int
+  n_dense: int
+  dimension_numbers: DotDimensionNumbers
+
+
+class SparseLayout(NamedTuple):
+  n_batch: int
+  n_dense: int
+  n_sparse: int
+
+
 class SparseTestCase(jtu.JaxTestCase):
   def assertSparseArraysEquivalent(self, x, y, *, check_dtypes=True, atol=None,
                                    rtol=None, canonicalize_dtypes=True, err_msg=''):
@@ -116,8 +142,8 @@ class SparseTestCase(jtu.JaxTestCase):
 
 def _rand_sparse(shape: Sequence[int], dtype: DTypeLike, *,
                  rng: np.random.RandomState, rand_method: Callable[..., Any],
-                 nse: Union[int, float], n_batch: int, n_dense: int,
-                 sparse_format: str) -> Union[sparse.BCOO, sparse.BCSR]:
+                 nse: int | float, n_batch: int, n_dense: int,
+                 sparse_format: str) -> sparse.BCOO | sparse.BCSR:
   if sparse_format not in ['bcoo', 'bcsr']:
     raise ValueError(f"Sparse format {sparse_format} not supported.")
 
@@ -157,7 +183,7 @@ def _rand_sparse(shape: Sequence[int], dtype: DTypeLike, *,
 
 def rand_bcoo(rng: np.random.RandomState,
               rand_method: Callable[..., Any]=jtu.rand_default,
-              nse: Union[int, float]=0.5, n_batch: int=0, n_dense: int=0):
+              nse: int | float=0.5, n_batch: int=0, n_dense: int=0):
   """Generates a random BCOO array."""
   return functools.partial(_rand_sparse, rng=rng, rand_method=rand_method,
                            nse=nse, n_batch=n_batch, n_dense=n_dense,
@@ -165,8 +191,48 @@ def rand_bcoo(rng: np.random.RandomState,
 
 def rand_bcsr(rng: np.random.RandomState,
               rand_method: Callable[..., Any]=jtu.rand_default,
-              nse: Union[int, float]=0.5, n_batch: int=0, n_dense: int=0):
+              nse: int | float=0.5, n_batch: int=0, n_dense: int=0):
   """Generates a random BCSR array."""
   return functools.partial(_rand_sparse, rng=rng, rand_method=rand_method,
                            nse=nse, n_batch=n_batch, n_dense=n_dense,
                            sparse_format='bcsr')
+
+
+def iter_subsets(s: Sequence) -> Iterable[tuple]:
+  """Return an iterator over all subsets of a sequence s"""
+  return itertools.chain.from_iterable(
+      itertools.combinations(s, n) for n in range(len(s) + 1)
+  )
+
+
+def iter_sparse_layouts(
+    shape: Sequence[int], min_n_batch=0
+) -> Iterator[SparseLayout]:
+  for n_batch in range(min_n_batch, len(shape) + 1):
+    for n_dense in range(len(shape) + 1 - n_batch):
+      n_sparse = len(shape) - n_batch - n_dense
+      yield SparseLayout(n_batch=n_batch, n_sparse=n_sparse, n_dense=n_dense)
+
+
+def iter_bcsr_layouts(
+    shape: Sequence[int], min_n_batch=0
+) -> Iterator[SparseLayout]:
+  n_sparse = 2
+  for n_batch in range(min_n_batch, len(shape) - 1):
+    n_dense = len(shape) - n_sparse - n_batch
+    yield SparseLayout(n_batch=n_batch, n_sparse=n_sparse, n_dense=n_dense)
+
+
+def rand_sparse(rng, nse=0.5, post=lambda x: x, rand_method=jtu.rand_default):
+  def _rand_sparse(shape, dtype, nse=nse):
+    rand = rand_method(rng)
+    size = math.prod(shape)
+    if 0 <= nse < 1:
+      nse = nse * size
+    nse = min(size, int(nse))
+    M = rand(shape, dtype)
+    indices = rng.choice(size, size - nse, replace=False)
+    M.flat[indices] = 0
+    return post(M)
+
+  return _rand_sparse
