@@ -15,8 +15,12 @@
 # Note: import <name> as <name> is required for names to be exported.
 # See PEP 484 & https://github.com/google/jax/issues/7570
 
+from jax._src.extend.random import (
+  define_prng_impl as define_prng_impl,
+)
+
 from jax._src.prng import (
-  PRNGImpl as PRNGImpl,
+  random_seed as random_seed,
   seed_with_impl as seed_with_impl,
   threefry2x32_p as threefry2x32_p,
   threefry_2x32 as threefry_2x32,
