@@ -18,6 +18,7 @@
 from jax._src.lax.lax import (
   DotDimensionNumbers as DotDimensionNumbers,
   Precision as Precision,
+  PrecisionLike as PrecisionLike,
   RandomAlgorithm as RandomAlgorithm,
   RoundingMethod as RoundingMethod,
   abs as abs,
@@ -88,9 +89,9 @@ from jax._src.lax.lax import (
   dot_general as dot_general,
   dot_general_p as dot_general_p,
   dtype as dtype,
-  dtypes as dtypes,
   eq as eq,
   eq_p as eq_p,
+  eq_to_p as eq_to_p,
   exp as exp,
   exp_p as exp_p,
   exp2 as exp2,
@@ -116,9 +117,9 @@ from jax._src.lax.lax import (
   iota_p as iota_p,
   is_finite as is_finite,
   is_finite_p as is_finite_p,
-  itertools as itertools,
   le as le,
   le_p as le_p,
+  le_to_p as le_to_p,
   log as log,
   log1p as log1p,
   log1p_p as log1p_p,
@@ -127,14 +128,13 @@ from jax._src.lax.lax import (
   logistic_p as logistic_p,
   lt as lt,
   lt_p as lt_p,
+  lt_to_p as lt_to_p,
   max as max,
   max_p as max_p,
   min as min,
   min_p as min_p,
   mul as mul,
   mul_p as mul_p,
-  naryop as naryop,
-  naryop_dtype_rule as naryop_dtype_rule,
   ne as ne,
   ne_p as ne_p,
   neg as neg,
@@ -152,6 +152,7 @@ from jax._src.lax.lax import (
   population_count_p as population_count_p,
   pow as pow,
   pow_p as pow_p,
+  ragged_dot as ragged_dot,
   real as real,
   real_p as real_p,
   reciprocal as reciprocal,
@@ -203,10 +204,6 @@ from jax._src.lax.lax import (
   square as square,
   squeeze as squeeze,
   squeeze_p as squeeze_p,
-  standard_abstract_eval as standard_abstract_eval,
-  standard_naryop as standard_naryop,
-  standard_primitive as standard_primitive,
-  standard_unop as standard_unop,
   stop_gradient as stop_gradient,
   sub as sub,
   sub_p as sub_p,
@@ -214,13 +211,10 @@ from jax._src.lax.lax import (
   tan_p as tan_p,
   tanh as tanh,
   tanh_p as tanh_p,
-  tie_in as tie_in,
   top_k as top_k,
   top_k_p as top_k_p,
   transpose as transpose,
   transpose_p as transpose_p,
-  unop as unop,
-  unop_dtype_rule as unop_dtype_rule,
   xor_p as xor_p,
   zeros_like_array as zeros_like_array,
 )
@@ -335,6 +329,7 @@ from jax._src.lax.control_flow import (
   switch as switch,
   while_loop as while_loop,
   while_p as while_p,
+  platform_dependent as platform_dependent,
 )
 from jax._src.lax.fft import (
   fft as fft,
@@ -347,6 +342,7 @@ from jax._src.lax.parallel import (
   all_to_all_p as all_to_all_p,
   axis_index as axis_index,
   axis_index_p as axis_index_p,
+  pbroadcast as pbroadcast,
   pmax as pmax,
   pmax_p as pmax_p,
   pmean as pmean,
@@ -359,8 +355,6 @@ from jax._src.lax.parallel import (
   psum_p as psum_p,
   psum_scatter as psum_scatter,
   pswapaxes as pswapaxes,
-  pdot as pdot,
-  xeinsum as xeinsum,
 )
 from jax._src.lax.other import (
   conv_general_dilated_local as conv_general_dilated_local,
@@ -377,3 +371,16 @@ from jax.lax import linalg as linalg
 from jax._src.pjit import with_sharding_constraint as with_sharding_constraint
 from jax._src.pjit import sharding_constraint_p as sharding_constraint_p
 from jax._src.dispatch import device_put_p as device_put_p
+
+
+_deprecations = {
+  # Finalized 2024-05-13; remove after 2024-08-13
+  "tie_in": (
+    "jax.lax.tie_in is deprecated: it has been a no-op since JAX v0.2.0. "
+    "Replace z = tie_in(x, y) with z = y.", None,
+  ),
+}
+
+from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+__getattr__ = _deprecation_getattr(__name__, _deprecations)
+del _deprecation_getattr
