@@ -32,16 +32,34 @@ from jax._src.nn.functions import (
   log_sigmoid as log_sigmoid,
   log_softmax as log_softmax,
   logsumexp as logsumexp,
-  normalize as normalize,
   standardize as standardize,
   one_hot as one_hot,
   relu as relu,
   relu6 as relu6,
+  dot_product_attention as dot_product_attention,
   selu as selu,
   sigmoid as sigmoid,
   soft_sign as soft_sign,
   softmax as softmax,
   softplus as softplus,
+  sparse_plus as sparse_plus,
+  sparse_sigmoid as sparse_sigmoid,
   silu as silu,
   swish as swish,
+  squareplus as squareplus,
+  mish as mish,
 )
+
+# Deprecations
+
+_deprecations = {
+    # Finalized 2024-05-13; remove after 2024-08-13
+    "normalize": (
+        "jax.nn.normalize is deprecated. Use jax.nn.standardize instead.",
+        None,
+    ),
+}
+
+from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+__getattr__ = _deprecation_getattr(__name__, _deprecations)
+del _deprecation_getattr
