@@ -15,13 +15,15 @@
 # Note: import <name> as <name> is required for names to be exported.
 # See PEP 484 & https://github.com/google/jax/issues/7570
 
-import numpy as _numpy
-
 from jax.numpy import fft as fft
 from jax.numpy import linalg as linalg
 
 from jax._src.basearray import Array as ndarray
 
+from jax._src.dtypes import (
+    isdtype as isdtype,
+)
+
 from jax._src.numpy.lax_numpy import (
     ComplexWarning as ComplexWarning,
     allclose as allclose,
@@ -41,6 +43,7 @@ from jax._src.numpy.lax_numpy import (
     array_repr as array_repr,
     array_split as array_split,
     array_str as array_str,
+    astype as astype,
     asarray as asarray,
     atleast_1d as atleast_1d,
     atleast_2d as atleast_2d,
@@ -50,6 +53,7 @@ from jax._src.numpy.lax_numpy import (
     bincount as bincount,
     blackman as blackman,
     block as block,
+    bool_ as bool,  # Array API alias for bool_
     bool_ as bool_,
     broadcast_arrays as broadcast_arrays,
     broadcast_shapes as broadcast_shapes,
@@ -65,6 +69,7 @@ from jax._src.numpy.lax_numpy import (
     complex_ as complex_,
     complexfloating as complexfloating,
     compress as compress,
+    concat as concat,
     concatenate as concatenate,
     convolve as convolve,
     copy as copy,
@@ -96,6 +101,7 @@ from jax._src.numpy.lax_numpy import (
     expand_dims as expand_dims,
     extract as extract,
     eye as eye,
+    fill_diagonal as fill_diagonal,
     finfo as finfo,
     fix as fix,
     flatnonzero as flatnonzero,
@@ -189,6 +195,7 @@ from jax._src.numpy.lax_numpy import (
     packbits as packbits,
     pad as pad,
     partition as partition,
+    permute_dims as permute_dims,
     pi as pi,
     piecewise as piecewise,
     place as place,
@@ -226,7 +233,7 @@ from jax._src.numpy.lax_numpy import (
     tensordot as tensordot,
     tile as tile,
     trace as trace,
-    trapz as _deprecated_trapz,
+    trapezoid as trapezoid,
     transpose as transpose,
     tri as tri,
     tril as tril,
@@ -246,9 +253,11 @@ from jax._src.numpy.lax_numpy import (
     unpackbits as unpackbits,
     unravel_index as unravel_index,
     unsignedinteger as unsignedinteger,
+    unstack as unstack,
     unwrap as unwrap,
     vander as vander,
     vdot as vdot,
+    vecdot as vecdot,
     vsplit as vsplit,
     vstack as vstack,
     where as where,
@@ -256,6 +265,15 @@ from jax._src.numpy.lax_numpy import (
     zeros_like as zeros_like,
 )
 
+# TODO(slebedev): Remove the try-except once we upgrade to ml_dtypes 0.4.1.
+try:
+  from jax._src.numpy.lax_numpy import (
+    int2 as int2,
+    uint2 as uint2,
+  )
+except ImportError:
+  pass
+
 from jax._src.numpy.index_tricks import (
   c_ as c_,
   index_exp as index_exp,
@@ -287,6 +305,7 @@ from jax._src.numpy.reductions import (
     count_nonzero as count_nonzero,
     cumsum as cumsum,
     cumprod as cumprod,
+    cumulative_sum as cumulative_sum,
     max as max,
     mean as mean,
     median as median,
@@ -313,18 +332,23 @@ from jax._src.numpy.reductions import (
 )
 
 from jax._src.numpy.setops import (
-    in1d as _deprecated_in1d,
     intersect1d as intersect1d,
     isin as isin,
     setdiff1d as setdiff1d,
     setxor1d as setxor1d,
     union1d as union1d,
     unique as unique,
+    unique_all as unique_all,
+    unique_counts as unique_counts,
+    unique_inverse as unique_inverse,
+    unique_values as unique_values,
 )
 
 from jax._src.numpy.ufuncs import (
     abs as abs,
     absolute as absolute,
+    acos as acos,
+    acosh as acosh,
     add as add,
     arccos as arccos,
     arccosh as arccosh,
@@ -333,8 +357,17 @@ from jax._src.numpy.ufuncs import (
     arctan as arctan,
     arctan2 as arctan2,
     arctanh as arctanh,
+    asin as asin,
+    asinh as asinh,
+    atan as atan,
+    atanh as atanh,
+    atan2 as atan2,
     bitwise_and as bitwise_and,
+    bitwise_count as bitwise_count,
+    bitwise_invert as bitwise_invert,
+    bitwise_left_shift as bitwise_left_shift,
     bitwise_not as bitwise_not,
+    bitwise_right_shift as bitwise_right_shift,
     bitwise_or as bitwise_or,
     bitwise_xor as bitwise_xor,
     cbrt as cbrt,
@@ -392,6 +425,7 @@ from jax._src.numpy.ufuncs import (
     nextafter as nextafter,
     not_equal as not_equal,
     positive as positive,
+    pow as pow,
     power as power,
     rad2deg as rad2deg,
     radians as radians,
@@ -426,59 +460,14 @@ register_jax_array_methods()
 del register_jax_array_methods
 
 
-# Deprecations
-
 _deprecations = {
-    # Added August 10, 2023:
-    "NINF": (
-        "jax.numpy.NINF is deprecated. Use -jax.numpy.inf instead.",
-        -inf,
-    ),
-    "NZERO": (
-        "jax.numpy.NZERO is deprecated. Use -0.0 instead.",
-        -0.0,
-    ),
-    "PZERO": (
-        "jax.numpy.PZERO is deprecated. Use 0.0 instead.",
-        0.0,
-    ),
-    # Added Aug 17, 2023:
-    "issubsctype": (
-        "jax.numpy.issubsctype is deprecated. In most cases, jax.numpy.issubdtype can be used instead.",
-        _numpy.core.numerictypes.issubsctype,
-    ),
-    # Added Aug 22, 2023
-    "row_stack": (
-        "jax.numpy.row_stack is deprecated. Use jax.numpy.vstack instead.",
-        vstack,
-    ),
-    # Added Aug 23, 2023
-    "in1d": (
-        "jax.numpy.in1d is deprecated. Use jax.numpy.isin instead.",
-        _deprecated_in1d,
-    ),
-    # Added Aug 24, 2023
-    "trapz": (
-        "jax.numpy.trapz is deprecated. Use jax.scipy.integrate.trapezoid instead.",
-        _deprecated_trapz,
-    ),
+  # Deprecated 18 Sept 2023 and removed 06 Feb 2024
+  "trapz": (
+    "jnp.trapz is deprecated; use jnp.trapezoid instead.",
+    None
+  ),
 }
 
-import typing
-if typing.TYPE_CHECKING:
-  row_stack = vstack
-  NINF = -inf
-  NZERO = -0.0
-  PZERO = 0.0
-  issubsctype = _numpy.core.numerictypes.issubsctype
-  in1d = _deprecated_in1d
-  trapz = _deprecated_trapz
-else:
-  from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
-  __getattr__ = _deprecation_getattr(__name__, _deprecations)
-  del _deprecation_getattr
-del typing
-del _numpy
-
-del _deprecated_in1d
-del _deprecated_trapz
+from jax._src.deprecations import deprecation_getattr as _deprecation_getattr
+__getattr__ = _deprecation_getattr(__name__, _deprecations)
+del _deprecation_getattr
